package logrus.agents;

import jade.content.Concept;
import jade.content.OntoACLMessage;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.content.onto.basic.Result;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.Logger;
import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;
import jade.wrapper.AgentContainer;
import jade.wrapper.AgentController;
import jade.wrapper.StaleProxyException;

import java.util.logging.Level;

import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.ontology.OntoStorageOntology;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.IdentifierGenerator;
import logrus.tools.logging.InitializeLoggerFailedException;
import logrus.tools.logging.LoggerWrapper;

/**
 * Baseclass for agents used in OntoStorage system
 */
public abstract class OntoStorageAgent extends Agent
{
	// JADE-specific logger
	private Logger logger;

	// codec used for encoding/decoding messages
	private Codec codec;

	// ontology of the messages exchanged between agents
	private Ontology ontology;

	// list of currently supported services
	protected ArrayList supportedServices = new ArrayList();

	/**
	 * Default constructor
	 * 
	 * @throws InitializeLoggerFailedException
	 * @throws OntologyException
	 */
	public OntoStorageAgent () throws InitializeLoggerFailedException,
			OntologyException
	{
		super();

		// initialize logger
		this.logger = LoggerWrapper.getInstance().InitializeLogger(this);

		// initialize codec and ontology used for exchanging messages
		this.codec = new SLCodec();
		this.getContentManager().registerLanguage(this.codec);
		this.ontology = OntoStorageOntology.getInstance();
		this.getContentManager().registerOntology(this.ontology);
	}
	
	/**
	 * Store error information
	 * 
	 * @param exc
	 * @param msg
	 */
	public void onError (Exception exc,String msg)
	{
		if (exc != null)
		{
			this.logg(msg,exc);
		}
	}	

	/**
	 * Store error information
	 * 
	 * @param exc
	 */
	public void onError (Exception exc)
	{
		if (exc != null)
		{
			this.logg(exc);
		}
	}

	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger ()
	{
		return this.logger;
	}

	/**
	 * Logs strMessage with specified java.util.Logger.Level
	 * 
	 * @param strMessage
	 *           Message to be logged
	 * @param level
	 *           Message level
	 */
	public void logg (Level level, String strMessage)
	{
		if (this.logger.isLoggable(level))
		{
			this.logger.log(jade.util.Logger.INFO,strMessage);
		}
	}

	/**
	 * Logs information about exception
	 */
	public void logg (Exception exc)
	{
		if (this.logger.isLoggable(jade.util.Logger.SEVERE))
		{
			this.logger.log(jade.util.Logger.SEVERE,exc.getMessage());
		}
	}

	/**
	 * Logs information about exception
	 */
	public void logg (String strMessage, Exception exc)
	{
		this.logg(Level.SEVERE,strMessage,exc);
	}

	/**
	 * Logs information about exception
	 */
	public void logg (Level level, String strMessage, Exception exc)
	{
		if (this.logger.isLoggable(level))
		{
			String strMessageEx = strMessage + " : " + exc.toString();
			this.logger.log(jade.util.Logger.INFO,strMessageEx);
		}
	}

	/**
	 * Build service description used for registering in DF service
	 * 
	 * @param serviceType
	 *           Type of service that agent will be supporting
	 * @param ontologyName
	 *           Ontology used for communication
	 * @return Service description used for registering in DF service
	 */
	private ServiceDescription buildServiceDescription (String serviceType,
			String ontologyName)
	{
		ServiceDescription servDescr = new ServiceDescription();
		servDescr.setType(serviceType);
		servDescr.setName(serviceType + "-service");
		servDescr.addOntologies(ontologyName);
		servDescr.addLanguages(FIPANames.ContentLanguage.FIPA_SL);
		return servDescr;
	}

	/**
	 * Build agent description used for registering in DF service
	 * 
	 * @param serviceType
	 *           Type of service that agent will be supporting
	 * @param ontologyName
	 *           Ontology used for communication
	 * @return Agent description used for registering in DF service
	 */
	private DFAgentDescription buildAgentDescription (String serviceType,
			String ontologyName)
	{
		DFAgentDescription agDescr = new DFAgentDescription();
		agDescr.setName(this.getAID());
		agDescr.addServices(this
				.buildServiceDescription(serviceType,ontologyName));
		return agDescr;
	}

	/**
	 * Registers service of specified type from ontology <i>ontologyName </i>
	 * 
	 * @param serviceType
	 *           Type of service (name of appropiate AgentActionSchema from ontology <i>ontologyName
	 *           </i>)
	 * @param ontologyName
	 *           Name of used ontology
	 * @throws FIPAException
	 */
	protected void registerService (String serviceType, String ontologyName)
			throws FIPAException
	{
		// prepare description of the service
		DFAgentDescription agDescr = new DFAgentDescription();
		agDescr.setName(this.getAID());
		ServiceDescription servDesc = this.buildServiceDescription(serviceType,
				ontologyName);
		agDescr.addServices(servDesc);

		// register the service
		DFService.register(this,agDescr);
		this.supportedServices.add(servDesc);
	}

	/**
	 * Deregister specified service specified by type and name from DF
	 * 
	 * @param serviceType
	 *           Type of service to be deregistered
	 * @param ontologyName
	 *           Name of service to be deregistered
	 * @throws FIPAException
	 */
	protected void deregisterService (String serviceType, String ontologyName)
			throws FIPAException
	{
		// deregister service
		DFAgentDescription agDescr = this.buildAgentDescription(serviceType,
				ontologyName);
		DFService.deregister(this,agDescr);

		ArrayList toRemove = new ArrayList();

		// remove service description form list of supported services
		for (Iterator iter = this.supportedServices.iterator(); iter.hasNext();)
		{
			ServiceDescription servDesc = (ServiceDescription)iter.next();
			if (servDesc.getType() == serviceType
					&& servDesc.getAllOntologies().next() == ontologyName)
			{
				toRemove.add(servDesc);
			}
		}
		for (Iterator iter = toRemove.iterator(); iter.hasNext();)
		{
			this.supportedServices.remove(iter.next());
		}
	}
	
	public AID[] findAgentsPerformingService (String serviceType) throws FIPAException
	{
		return this.findAgentsPerformingService(serviceType,1024);
	}
	
	/**
	 * Finds AIDs of agents supporting specified service
	 * 
	 * @param serviceType
	 *           Type of service to be found
	 * @param maxAgents
	 *           Maximal number of agents we want to find
	 * @return Array of AIDs of agents supporting specified service or null if there are no such
	 *         agents available
	 * @throws FIPAException
	 */
	protected AID[] findAgentsPerformingService (String serviceType,
			int maxAgents) throws FIPAException
	{
		AID[] aids = null;
		DFAgentDescription templ = new DFAgentDescription();
		ServiceDescription templServDesc = new ServiceDescription();

		// prepare description of service we're looking for
		templServDesc.setType(serviceType);
		templ.addServices(templServDesc);

		// prepare search constraints
		SearchConstraints sc = new SearchConstraints();
		sc.setMaxResults(new Long(maxAgents));

		// perform search
		DFAgentDescription[] results = DFService.search(this,templ,sc);

		// if DFService found some agents then extract their AIDs from
		// service descriptions
		if (results.length > 0)
		{
			aids = new AID[results.length];
			for (int i = 0; i < results.length; i++)
			{
				aids[i] = results[i].getName();
			}
		}
		return aids;
	}

	/**
	 * Finds AID of agent supporting specified service
	 * 
	 * @param serviceType
	 *           Type of service to be found
	 * @return AID of agent supporting specified service or null if there is no such agent available
	 * @throws FIPAException
	 */
	public AID findAgentPerformingService (String serviceType)
			throws FIPAException
	{
		AID[] results = this.findAgentsPerformingService(serviceType,1);
		if (results != null)
		{
			return results[0];
		}
		else
		{
			return null;
		}
	}

	/**
	 * Launches pool of agents
	 * 
	 * @param agentClassName
	 *           Type of agents to be launched
	 * @param agentNamePrefix
	 *           Agents will get names agentNamePrefix1, agentNamePrefix2, ...
	 * @param poolsize
	 *           Number of agents to be created
	 * @throws StaleProxyException
	 */
	protected void launchAgentsPool (String agentClassName,
			String agentNamePrefix, int poolsize) throws StaleProxyException
	{
		// get agent container in which our agent is running
		AgentContainer container = this.getContainerController();

		// launch appropiate number of agents
		for (int i = 1; i < poolsize; i++)
		{
			AgentController ag = container.createNewAgent(agentNamePrefix + i,
					agentClassName,null);
			ag.start();
		}
	}

	/**
	 * Launches pool of agents
	 * 
	 * @param agentClassName
	 *           Type of agents to be launched
	 * @param poolsize
	 *           Number of agents to be created
	 * @throws StaleProxyException
	 */
	protected void launchAgentsPool (String agentClassName, int poolsize)
			throws StaleProxyException
	{
		this.launchAgentsPool(agentClassName,agentClassName,poolsize);
	}

	/**
	 * Constructs ACL message, autogenerates conversation ID
	 * 
	 * @param iFIPAPerformative
	 *           Identifier specifying type of message (according to FIPA standard)
	 * @param reciver
	 *           Reciver of this message
	 * @param content
	 *           Content of the message will be filled with this object
	 * @return Created ACL message
	 * @throws CodecException
	 * @throws OntologyException
	 * @throws IdentifierGenerationFailedException
	 */
	public OntoACLMessage buildACLMessage (int iFIPAPerformative, AID reciver,
			Concept content) throws CodecException, OntologyException,
			IdentifierGenerationFailedException
	{
		OntoACLMessage msg = new OntoACLMessage(iFIPAPerformative);
		msg.addReceiver(reciver);
		msg.setConversationId(IdentifierGenerator.getInstance().GenerateID());
		msg.setOntology(this.ontology.getName());
		msg.setLanguage(this.codec.getName());
		this.getContentManager().fillContent(msg,
				this.packToAction(reciver,content));
		return msg;
	}

	/**
	 * Constructs ACL message, conversation ID set by user
	 * 
	 * @param FIPAPerformative
	 *           Identifier specifying type of message (according to FIPA standard)
	 * @param reciver
	 *           Reciver of this message
	 * @param content
	 *           Content of the message will be filled with this object
	 * @return Created ACL message
	 * @throws CodecException
	 * @throws OntologyException
	 * @throws IdentifierGenerationFailedException
	 */
	public OntoACLMessage buildACLMessage (int FIPAPerformative, AID reciver,
			Concept content, String conversationID) throws CodecException,
			OntologyException
	{
		OntoACLMessage msg = new OntoACLMessage(FIPAPerformative);
		msg.addReceiver(reciver);
		msg.setConversationId(conversationID);
		msg.setOntology(this.ontology.getName());
		msg.setLanguage(this.codec.getName());
		this.getContentManager().fillContent(msg,
				this.packToAction(reciver,content));
		return msg;
	}

	/**
	 * Prepare reply message - set basic properties (ontology, language, reciver, conversation ID)
	 * 
	 * @param orignialMsg
	 *           Original message, reply messsage will inherit ontology, language and conversation ID
	 *           form orignial message, sender of this message will be set as an reciver of reply
	 * @return Initially prepared reply message
	 */
	public OntoACLMessage prepareReply (OntoACLMessage orignialMsg)
	{
		OntoACLMessage reply = new OntoACLMessage();
		reply.setConversationId(orignialMsg.getConversationId());
		reply.setOntology(orignialMsg.getOntology());
		reply.setLanguage(orignialMsg.getLanguage());
		reply.addReceiver(orignialMsg.getSender());
		return reply;
	}

	/**
	 * "Nearly" infinite wait for ACL message matching specified pattern
	 * 
	 * @param pattern
	 *           Message pattern we want to recive
	 * @param behaviour
	 *           Behaviour in which we want to wait for a message
	 * @param blockTime
	 *           Limit (in miliseconds) for <code>Behaviour.block()</code> method
	 * @return
	 */
	public OntoACLMessage waitForMessage (MessageTemplate pattern,
			Behaviour behaviour, int blockTime)
	{
		try
		{
			return this.waitForMessage(pattern,behaviour,blockTime,blockTime,
					Timeouts.MAX_TIMEOUT);
		}
		catch (MessageTimeoutException e)
		{
			return null;
		}
	}

	/**
	 * Wait for ACL message matching specified pattern
	 * 
	 * @param pattern
	 *           Message pattern we want to recive
	 * @param behaviour
	 *           Behaviour in which we want to wait for a message
	 * @param minBlock
	 *           Minimal limit (in miliseconds) for <code>Behaviour.block()</code> method
	 * @param maxBlock
	 *           Maximal limit (in miliseconds) for <code>Behaviour.block()</code> method
	 * @param timeout
	 *           Wait time in miliseconds after <code>MessageTimeoutException</code> is thrown
	 * @return ACL message matching the patern
	 * @throws MessageTimeoutException
	 */
	public OntoACLMessage waitForMessage (MessageTemplate pattern,
			Behaviour behaviour, int minBlock, int maxBlock, int timeout)
			throws MessageTimeoutException
	{
		int waitTime = 0;
		int currBlock = minBlock;
		int step;
		while (true)
		{
			if (waitTime > timeout)
			{
				throw new MessageTimeoutException();
			}
			ACLMessage msg = this.receive(pattern);
			if (msg != null)
			{
				return (OntoACLMessage)msg;
			}
			else
			{
				behaviour.block(currBlock);
				waitTime += currBlock;
				step = (maxBlock - minBlock) / 10;
				currBlock = currBlock + step < maxBlock ? currBlock + step
						: currBlock;
			}
		}
	}

	/**
	 * Send message of specified type <code>type</code> to agent with AID equal
	 * <code>reciver</code> and message content filled with <code>content</code>
	 * 
	 * @param type
	 *           Type of message (static code definied in ACLMessage)
	 * @param content
	 *           Content of message
	 * @param reciver
	 *           AID of reciver
	 * @return Generated conversation identifier
	 * @throws SendingMessageFailed
	 */
	public String sendMessage (int type, Concept content, AID reciver)
			throws SendingMessageFailed
	{
		try
		{
			OntoACLMessage msg = this.buildACLMessage(type,reciver,content);
			this.send(msg);
			return msg.getConversationId();
		}
		catch (CodecException exc)
		{
			throw new SendingMessageFailed(exc);
		}
		catch (OntologyException exc)
		{
			throw new SendingMessageFailed(exc);
		}
		catch (IdentifierGenerationFailedException exc)
		{
			throw new SendingMessageFailed(exc);
		}
	}

	/**
	 * Send message of specified type <code>type</code> to agent with AID equal
	 * <code>reciver</code> and message content filled with <code>content</code>
	 * 
	 * @param type
	 *           Type of message (static code definied in ACLMessage)
	 * @param content
	 *           Content of message
	 * @param reciver
	 *           AID of reciver
	 * @param conversationID
	 *           Sent message id part of message exchange process identified by
	 *           <code>conversationID</code>
	 * @throws SendingMessageFailed
	 */
	public void sendMessage (int type, Concept content, AID reciver,
			String conversationID) throws SendingMessageFailed
	{
		try
		{
			OntoACLMessage msg = this.buildACLMessage(type,reciver,content,
					conversationID);
			this.send(msg);
		}
		catch (CodecException exc)
		{
			throw new SendingMessageFailed(exc);
		}
		catch (OntologyException exc)
		{
			throw new SendingMessageFailed(exc);
		}
	}

	/**
	 * Packs concept into the Action construct (SL cocdec requirement)
	 * 
	 * @param aid
	 *           AID of message reciver
	 * @param concept
	 *           Concept to be packed
	 * @return Action wrapping specified concept
	 */
	public Action packToAction (AID aid, Concept concept)
	{
		Action action = new Action(aid,concept);
		return action;
	}

	/**
	 * Pack concept and value (result of some action) to the Result construct (operator of the FIPA
	 * SL0 action)
	 * 
	 * @param concept
	 *           Concept (action that produced some result)
	 * @param value
	 *           Object representing result of action represented by concept
	 * @return Result of action represented by <code>concept</code>, value of the result is
	 *         represented by <code>value</code>
	 */
	public Result packToResult (Concept concept, Object value)
	{
		Result result = new Result(concept,value);
		return result;
	}

	/**
	 * Extract concept from Action construct
	 * 
	 * @param action
	 *           Action from which concept is to be extracted
	 * @return Concept extracted from Action concept
	 */
	public Concept unpackFromAction (Action action)
	{
		return action.getAction();
	}

	/**
	 * Extract content from Result construct
	 * 
	 * @param result
	 *           Result from which concept is to be extracted
	 * @return Concept extracted from Result concept
	 */
	public Concept unpackFromResult (Result result)
	{
		return (Concept)result.getValue();
	}

	/**
	 * Log information about sending message failure
	 * 
	 * @param aidReceiver
	 *           <code>AID</code> of agent that has been intended as a receiver of this message
	 * @param replyCode
	 *           Code of message that has been intended to be sent
	 * @param exc
	 *           Exception that caused problems with sending the message
	 * @param msg
	 *           Content of the message (may be null)
	 */
	public void onSendMessgeFailed (AID aidReceiver, int replyCode,
			Exception exc, OSMessage msg)
	{
		// the only thing we can do is to log information about error detils
		StringBuffer buffer = new StringBuffer();
		buffer.append("Sending message ");
		buffer.append(replyCode);
		buffer.append(" from ");
		buffer.append(this.getAID());
		buffer.append(" to ");
		buffer.append(aidReceiver);
		if (msg != null)
		{
			buffer.append(" with content ");
			buffer.append(msg.toString());
		}
		buffer.append(" failed");
		this.logg(Level.SEVERE,buffer.toString(),exc);
	}

	/**
	 * Log information about unexpected message <code>msg</code> received form agent identified by
	 * AID <code>aidClient</code>
	 * 
	 * @param agentName
	 *           Name of agent
	 * @param msg
	 *           Message received form <code>aidClient</code>
	 * @param aidClient
	 *           AID of agent that has sent unexpected message to this agent
	 */
	public void onUnexpectedMessage (String agentName, OSMessage msg,
			AID aidClient)
	{
		this.logg(Level.WARNING,agentName
				+ " received unknown/unexpected message " + msg.getMessageCode()
				+ " from " + aidClient);
	}
}
