package logrus.agents.dataSupervisionAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.core.AID;
import jade.core.Agent;
import jade.domain.FIPAException;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;

import java.util.logging.Level;

import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.ServiceNames;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.dataSupervisionAgent.DataSupervisionAgent;
import logrus.agents.datasources.DataProviderID;
import logrus.agents.searchCoordinatorAgent.SearchTask;
import logrus.tools.Utils;

public class UpdateSupervisionBehaviour extends OSBehaviour
{
	// behaviour should listen for messages wuth this conversation ID
	private String conversationID;

	// DSA agent reference
	private DataSupervisionAgent agentDSA;

	// current stage of behaviour
	private int currentStage;

	// error message used in StageFinal
	private String strErrorMessage;

	// index of stage responsible for sending query to DBA
	private static final int StageSendQueryToDBA = 1;

	// index of stage responsible for waiting for response from DBA
	private static final int StageWaitForReplyFromDBA = 2;

	// index of stage responsible for sending URIs of outdated objects to SCA
	private static final int StageSendOutdatedToUpdate = 3;

	// index of stage responsible for collecting extracted object informations from
	private static final int StageCollectExtractedInfos = 4;

	// index of stage responsible for sending object infos to IA
	private static final int StageSendInfosToIA = 5;

	// index of stage responsible for collecting replies from IA
	private static final int StageCollectIAReplies = 6;

	// index of stage responsible for cleaning after error
	private static final int StageFinal = 10;

	// list of outdated objects
	private ArrayList outdatedObjects;

	// list of objects infos extracted by SCA
	private ArrayList infos;

	// AID of GUI agent
	private AID aidGUI;

	// flag determinig if behaviour should be finished
	private boolean finished;

	// number of object informations that should be collected form SCA
	private int toCollectInfos;

	// number of object informations that have been already collected form SCA
	private int collectedInfos;

	// number of objects sent to IA
	private int nObjects;

	// number of objects injected by IA
	private int nObjectsInjected;

	public UpdateSupervisionBehaviour (Agent a, String conversationID)
	{
		super(a);
		this.agentDSA = (DataSupervisionAgent)a;
		this.currentStage = StageSendQueryToDBA;
		this.finished = false;
		this.outdatedObjects = null;
		this.conversationID = conversationID;
		this.toCollectInfos = 0;
		this.collectedInfos = 0;
		this.infos = new ArrayList();
		this.nObjects = 0;
		this.nObjectsInjected = 0;
		this.aidGUI = null;
		try
		{
			this.aidGUI = this.agentDSA
					.findAgentPerformingService(ServiceNames.ServiceGUI);
		}
		catch (FIPAException exc)
		{
			this.strErrorMessage = "Can't find GUI agent: " + exc;
			this.currentStage = StageFinal;
			return;
		}

	}

	public void action ()
	{
		switch (this.currentStage)
		{
			case StageSendQueryToDBA:
				this.sendQueryToDBA();
				break;
			case StageWaitForReplyFromDBA:
				this.waitForReplyFromDBA();
				break;
			case StageSendOutdatedToUpdate:
				this.sendOutdatedToUpdate();
				break;
			case StageCollectExtractedInfos:
				this.collectExtractedInfos();
				break;
			case StageSendInfosToIA:
				this.sendInfosToIA();
				break;
			case StageCollectIAReplies:
				this.collectIAReplies();
				break;
			case StageFinal:
				break;
		}
	}

	/**
	 * Perform final actions before finishing behaviour
	 */
	private void finalizeBehaviour ()
	{
		// if there were some errors then log them
		if (this.strErrorMessage != null)
		{
			this.agentDSA.logg(Level.SEVERE,this.strErrorMessage);

			// try to send failure information to GUI
			OSMessage reply = new OSMessage();
			reply.setMessageCode(OSMessage.DSA_Update_Failed);
			reply.setAnnouncement(this.strErrorMessage);
			try
			{
				this.agentDSA.sendMessage(ACLMessage.INFORM,reply,this.aidGUI,
						this.conversationID);
			}
			catch (SendingMessageFailed exc)
			{
				this.agentDSA.onSendMessgeFailed(this.aidGUI,
						OSMessage.DSA_Update_Failed,exc,reply);
			}
		}

		// mark behaviour as finished
		this.finished = true;
	}

	private void collectIAReplies ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.IA_Mass_Index_Failed);
		matcher.addAcceptedMsgCode(OSMessage.IA_Indexed_Failed);
		matcher.addAcceptedMsgCode(OSMessage.IA_Indexed_OK);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			// extract information form message
			AID aidClient = jmsg.getSender();
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			// check which operation is requested and perform it
			switch (msg.getMessageCode())
			{
				case OSMessage.IA_Indexed_OK:
					// indexing single object ok
					++this.nObjectsInjected;
					break;
				case OSMessage.IA_Indexed_Failed:
					++this.nObjectsInjected;
					// indexing single object has failed
					break;
				case OSMessage.IA_Mass_Index_Failed:
					// big failure
					this.strErrorMessage = "Indexing crawl results failed";
					this.currentStage = StageFinal;
					return;
			}
			if (this.nObjects == this.nObjectsInjected)
			{
				// send message to GUI client
				OSMessage reply = new OSMessage();
				reply.setMessageCode(OSMessage.DSA_Update_OK);
				reply.setAnnouncement("Update complete");
				try
				{
					this.agentDSA.sendMessage(ACLMessage.INFORM,reply,this.aidGUI,
							this.conversationID);
				}
				catch (SendingMessageFailed exc)
				{
					this.strErrorMessage = "Sending message DSA_Update_OK to GUI failed";
					this.agentDSA.onSendMessgeFailed(aidClient,
							OSMessage.DSA_Update_OK,exc,reply);
				}
				finally
				{
					this.currentStage = StageFinal;
				}
			}
		}
		else
		{
			// block behaviour until next message arrives
			block();
		}
	}

	private void sendInfosToIA ()
	{
		OSMessage request = new OSMessage();
		request.setMessageCode(OSMessage.IA_Mass_Index);
		request.setObjectsCollection(this.infos);
		this.nObjects = this.infos.size();

		// find SCA
		AID aidIA = null;
		try
		{
			aidIA = this.agentDSA
					.findAgentPerformingService(ServiceNames.ServiceIA);
		}
		catch (FIPAException exc)
		{
			this.strErrorMessage = "Can't find IA: " + exc;
			this.currentStage = StageFinal;
			return;
		}

		try
		{
			this.agentDSA.sendMessage(ACLMessage.REQUEST,request,aidIA,
					this.conversationID);
			this.currentStage = StageCollectIAReplies;
		}
		catch (SendingMessageFailed exc)
		{
			this.strErrorMessage = "Sending message to IA failed: "
					+ exc.getMessage();
			this.currentStage = StageFinal;
		}
	}

	private void collectExtractedInfos ()
	{
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.SCA_Extract_Object_Info_Result);
		matcher.addAcceptedMsgCode(OSMessage.SCA_Extract_Object_Info_Failed);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			switch (msg.getMessageCode())
			{
				case OSMessage.SCA_Extract_Object_Info_Result:
					++this.collectedInfos;
					this.infos.add(msg.getObjectInfo());
					break;
				case OSMessage.SCA_Extract_Object_Info_Failed:
					++this.collectedInfos;
					break;
			}
			this.agentDSA.logg(Level.INFO,"DSA collected " + this.collectedInfos
					+ "/" + this.toCollectInfos + " results");
			if (this.collectedInfos == this.toCollectInfos)
			{
				this.currentStage = StageSendInfosToIA;
			}
		}
		else
		{
			block();
		}
	}

	/**
	 * Send requests for extracting information from data sources
	 */
	private void sendOutdatedToUpdate ()
	{
		AID aidSCA = null;
		try
		{
			aidSCA = this.osAgent
					.findAgentPerformingService(ServiceNames.ServiceSCA);
		}
		catch (FIPAException exc)
		{
			this.strErrorMessage = "Can't find SCA: " + exc;
			this.currentStage = StageFinal;
			return;
		}

		for (Iterator iter = this.outdatedObjects.iterator(); iter.hasNext();)
		{
			// prepare search task
			String url = (String)iter.next();
			String providerID = null;
			if (url.startsWith("http://chefmoz.org"))
			{
				providerID = DataProviderID.CHEFMOZ;
			}
			else if (url.startsWith("http://www.restauracje.com"))
			{
				providerID = DataProviderID.RESTCOM;
			}
			SearchTask task = new SearchTask(SearchTask.EXTRACT_INFO,providerID,
					url);
			task.setConversationID(this.conversationID);

			// prepare message for SCA
			OSMessage msg = new OSMessage();
			msg.setMessageCode(OSMessage.SCA_Extract_Object_Info);
			msg.setSearchCommand(task);

			try
			{
				this.osAgent.sendMessage(ACLMessage.REQUEST,msg,aidSCA,
						this.conversationID);
			}
			catch (SendingMessageFailed exc)
			{
				this.agentDSA.onSendMessgeFailed(aidSCA,
						OSMessage.SCA_Extract_Object_Info,exc,msg);
				this.currentStage = StageFinal;
				return;
			}
		}
		this.currentStage = StageCollectExtractedInfos;
	}

	/**
	 * Performs process of sending query for finding outdated objects
	 */
	private void sendQueryToDBA ()
	{
		AID aidDBA = null;
		OSMessage msg = null;
		try
		{
			msg = new OSMessage();
			msg.setMessageCode(OSMessage.DBA_Find_Outdated);
			aidDBA = Utils.getInstance().tossDBARead(this.osAgent);
			this.agentDSA.sendMessage(ACLMessage.REQUEST,msg,aidDBA,
					this.conversationID);
			this.currentStage = StageWaitForReplyFromDBA;
		}
		catch (FIPAException exc)
		{
			this.osAgent.onError(exc,
					"Can't find agent performing model reading service");
			this.currentStage = StageFinal;
			return;
		}
		catch (SendingMessageFailed exc)
		{
			this.agentDSA.onSendMessgeFailed(aidDBA,OSMessage.DBA_Find_Outdated,
					exc,msg);
			this.currentStage = StageFinal;
			return;
		}
	}

	/**
	 * Wait for reply from DBA
	 */
	private void waitForReplyFromDBA ()
	{
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Outdated_Result);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Outdated_Failed);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			switch (msg.getMessageCode())
			{
				case OSMessage.DBA_Find_Outdated_Result:
					this.outdatedObjects = msg.getUris();
					if (this.outdatedObjects == null)
					{
						this.osAgent.logg(Level.INFO,"No objects require update");
						this.currentStage = StageFinal;
					}
					else
					{
						this.osAgent.logg(Level.INFO,"There are "
								+ this.outdatedObjects.size()
								+ " source links to check");
						this.toCollectInfos = this.outdatedObjects.size();
						this.currentStage = StageSendOutdatedToUpdate;
					}
					break;
				case OSMessage.DBA_Find_Outdated_Failed:
					// error :(
					this.strErrorMessage = "Extracting URIs of outdated objects failed: "
							+ msg.getAnnouncement();
					this.currentStage = StageFinal;
					break;
			}
		}
		else
		{
			block();
		}
	}

	public boolean done ()
	{
		return this.finished;
	}
}
