package logrus.agents.databaseAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.leap.ArrayList;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.ParseException;
import java.util.NoSuchElementException;
import java.util.logging.Level;

import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSCyclicBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.ontmodel.IOntModelWrapper;
import logrus.ontmodel.OntUtils;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.exceptions.MatchingException;
import logrus.ontmodel.exceptions.OntPropertyNotFoundException;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.IdentifierGenerator;
import logrus.tools.Utils;
import logrus.tools.queue.Queue;

import com.hp.hpl.jena.rdql.QueryException;
import com.hp.hpl.jena.rdql.QueryResults;
import com.hp.hpl.jena.rdql.QueryResultsFormatter;
import com.hp.hpl.jena.rdql.parser.TokenMgrError;

/**
 * 
 */
public class ServiceDBABehaviour extends OSCyclicBehaviour
{
	private IOntModelWrapper model;

	/**
	 * @param a
	 */
	public ServiceDBABehaviour (Agent a, IOntModelWrapper model)
	{
		super(a);
		this.model = model;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DBA_Extract_Object_Info);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Execute_RDQL);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Inject_Data);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Matching);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Save_Model);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Build_Report);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Outdated);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			// extract information form message
			AID aidClient = jmsg.getSender();
			String conversationID = jmsg.getConversationId();
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			// check which operation is requested and perform it
			switch (msg.getMessageCode())
			{
				case OSMessage.DBA_Inject_Data:
					this.onInjectData(msg,aidClient,conversationID);
					break;
				case OSMessage.DBA_Extract_Object_Info:
					this.onExtractObjectInfo(msg,aidClient,conversationID);
					break;
				case OSMessage.DBA_Execute_RDQL:
					this.onExecuteRDQL(msg,aidClient,conversationID);
				case OSMessage.DBA_Find_Matching:
					this.onExtractMatching(msg,aidClient,conversationID);
					break;
				case OSMessage.DBA_Save_Model:
					this.onSaveModel(msg,aidClient,conversationID);
					break;
				case OSMessage.DBA_Build_Report:
					this.onBuildReport(msg,aidClient,conversationID);
					break;
				case OSMessage.DBA_Find_Outdated:
					this.onFindOutdated(msg,aidClient,conversationID);
					break;
				default:
					this.osAgent.logg(Level.WARNING,
							"DBA received unknown/unexpected message "
									+ msg.getMessageCode() + " from " + aidClient);
					return;
			}
		}
		else
		{
			// no message in queue that should be processed so wait for next one
			block();
		}
	}

	private void onBuildReport (OSMessage msg, AID aidClient,
			String conversationID)
	{
		OSMessage reply = null;
		String report = null;

		// generate report and leave critical section as fast as possible
		try
		{
			this.model.enterReadCriticalSection();
			report = this.model.buildReport(msg.getAnnouncement());
		}
		catch (OntPropertyNotFoundException exc)
		{
			this.sendErrorMessage(OSMessage.DBA_Build_Report_Failed,
					"Building report failed. ",exc,aidClient,conversationID);
		}
		finally
		{
			this.model.leaveCriticalSection();
		}

		// if report is null means some error has occured
		if (report == null)
		{
			return;
		}

		try
		{
			reply = new OSMessage();
			reply.setMessageCode(OSMessage.DBA_Build_Report_OK);
			reply.setAnnouncement(report);
			this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
					conversationID);
		}
		catch (SendingMessageFailed exc)
		{
			this.osAgent.onSendMessgeFailed(aidClient,
					OSMessage.DBA_Build_Report_OK,exc,reply);
		}
	}

	private void onSaveModel (OSMessage msg, AID aidClient, String conversationID)
	{
		String path = msg.getAnnouncement();
		OSMessage reply = null;
		boolean dumpedOK = false;

		try
		{
			this.model.enterReadCriticalSection();
			this.model.dumpToFile(path);
			dumpedOK = true;
		}
		catch (IOException exc)
		{
			this.sendErrorMessage(OSMessage.DBA_Save_Model_Failed,
					"Couldn't save model to a file",exc,aidClient,conversationID);
		}
		finally
		{
			this.model.leaveCriticalSection();
		}

		try
		{
			if (dumpedOK)
			{
				reply = new OSMessage();
				reply.setMessageCode(OSMessage.DBA_Save_Model_OK);
				this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
						conversationID);
			}
		}
		catch (SendingMessageFailed exc)
		{
			this.osAgent.onSendMessgeFailed(aidClient,OSMessage.DBA_Save_Model_OK,
					exc,reply);
		}
	}

	private void sendErrorMessage (int msgCode, String strMessage,
			Exception exc, AID receiver, String conversationID)
	{
		OSMessage reply = new OSMessage();
		reply.setMessageCode(msgCode);
		String extension = null;
		if (exc != null)
		{
			extension = exc.getMessage();
		}
		reply.setAnnouncement(strMessage + ":" + extension);
		try
		{
			this.osAgent.sendMessage(ACLMessage.FAILURE,reply,receiver,
					conversationID);
		}
		catch (SendingMessageFailed e)
		{
			this.osAgent.onSendMessgeFailed(receiver,msgCode,e,reply);
		}
	}

	private void onFindOutdated (OSMessage msg, AID aidClient,
			String conversationID)
	{
		ArrayList outdatedObjects = null;
		try
		{
			this.model.enterReadCriticalSection();
			outdatedObjects = this.model.findOutdatedObjects();
		}
		catch (OntPropertyNotFoundException exc)
		{
			this.sendErrorMessage(OSMessage.DSA_Update_Failed,
					"DSA: finding outdated objects failed: ",exc,aidClient,
					conversationID);
			return;
		}
		catch (ParseException exc)
		{
			this.sendErrorMessage(OSMessage.DSA_Update_Failed,
					"DSA: finding outdated objects failed: ",exc,aidClient,
					conversationID);
			return;
		}
		catch (NoSuchElementException exc)
		{
			this.sendErrorMessage(OSMessage.DSA_Update_Failed,
					"DSA: finding outdated objects failed: ",exc,aidClient,
					conversationID);
			return;			
		}
		finally
		{
			this.model.leaveCriticalSection();
		}

		// ok we have list outdated objects or null if no update is needed
		OSMessage reply = new OSMessage();
		reply.setMessageCode(OSMessage.DBA_Find_Outdated_Result);
		reply.setUris(outdatedObjects);

		try
		{
			this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
					conversationID);
		}
		catch (SendingMessageFailed exc)
		{
			this.osAgent.onSendMessgeFailed(aidClient,reply.getMessageCode(),exc,
					reply);
		}
	}

	private void onExtractMatching (OSMessage msg, AID aidClient,
			String conversationID)
	{
		// first prerequisite: message has to have QuadList
		if (msg.hasObjectInfo())
		{
			QuadsList objInfo = msg.getObjectInfo();
			QuadsList searchResult = null;

			try
			{
				this.model.enterReadCriticalSection();
				searchResult = this.model.findMatchingObject(objInfo);
			}
			catch (MatchingException exc)
			{
				this
						.sendErrorMessage(OSMessage.DBA_Find_Matching_Failed,
								"DBA: matching process failed: ",exc,aidClient,
								conversationID);
				return;
			}
			finally
			{
				this.model.leaveCriticalSection();
			}

			OSMessage reply;
			if (searchResult == null)
			{
				// no info found
				reply = new OSMessage();
				reply.setMessageCode(OSMessage.DBA_Find_Matching_NotFound);
			}
			else
			{
				// found info
				reply = new OSMessage();
				reply.setMessageCode(OSMessage.DBA_Find_Matching_Result);
				reply.setObjectInfo(searchResult);
			}
			try
			{
				this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
						conversationID);
			}
			catch (SendingMessageFailed exc)
			{
				this.osAgent.onSendMessgeFailed(aidClient,reply.getMessageCode(),
						exc,reply);
			}
		}
		else
		{
			// error: no quads list to match
			this.sendErrorMessage(OSMessage.DBA_Find_Matching_Failed,
					"No quads list specified for matching",null,aidClient,
					conversationID);
			return;
		}
	}

	/**
	 * Inject data sent in message (insert new data or update existing)
	 * 
	 * @param msg
	 *           Message containing sended data
	 * @param aidClient
	 *           Interlocutor's AID
	 * @param conversationID
	 *           ID of conversation with agent identified by <code>customerAID</code>
	 */
	private void onInjectData (OSMessage msg, AID aidClient,
			String conversationID)
	{
		// first prerequisite: message has to have QuadList
		if (msg.hasObjectInfo())
		{
			QuadsList newStmts = msg.getObjectInfo();

			int replyCode = 0;
			// second prerequisite: QuadList must not be empty
			String objectURI = newStmts.getObjectURI();
			if (objectURI != null)
			{
				try
				{
					boolean injectedOK = false;
					try
					{
						QuadsList oldStmts = this.model.getObjectInfo(objectURI);
						if (oldStmts == null)
						{
							oldStmts = new QuadsList();
						}

						// generate list of atomic operations that should be performed on a model...
						Queue operations = OntUtils.getInstance()
								.calculateOperations(oldStmts,newStmts);

						// ...end execute them
						this.model.enterWriteCriticalSection();
						this.model.performOperations(operations);
						injectedOK = true;
					}
					catch (OntPropertyNotFoundException exc)
					{
						// error occured during adding statements, inform interlocutor
						OSMessage reply = new OSMessage();
						replyCode = OSMessage.DBA_Inject_Data_Failed;
						reply.setMessageCode(replyCode);
						this.osAgent.sendMessage(ACLMessage.FAILURE,reply,aidClient,
								conversationID);
						this.osAgent.logg(exc);
					}
					finally
					{
						this.model.leaveCriticalSection();
					}

					if (injectedOK)
					{
						// injecting data ended successfully, inform interlocutor
						OSMessage reply = new OSMessage();
						replyCode = OSMessage.DBA_Inject_Data_OK;
						reply.setMessageCode(replyCode);
						this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
								conversationID);
					}
					else
					{
						return;
					}
				}
				catch (SendingMessageFailed exc)
				{
					// the only thing we can do is to log information about error detils
					this.osAgent.onSendMessgeFailed(aidClient,replyCode,exc,null);
				}
				catch (Exception exc)
				{
					OSMessage reply = new OSMessage();
					replyCode = OSMessage.DBA_Inject_Data_Failed;
					reply.setMessageCode(replyCode);
					reply.setAnnouncement("Inject data failed");
					try
					{
						this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
								conversationID);
					}
					catch (SendingMessageFailed e)
					{
						this.osAgent.onSendMessgeFailed(aidClient,replyCode,exc,null);
					}
				}
			}
		}
	}

	private void onExtractObjectInfo (OSMessage msg, AID aidClient,
			String conversationID)
	{
		QuadsList objectInfo = null;
		// first prerequisite: message has to have URI of object to be extracted
		if (msg.hasAnnouncement())
		{
			String objectURI = msg.getAnnouncement();
			try
			{
				this.model.enterReadCriticalSection();
				objectInfo = this.model.getObjectInfo(objectURI);
			}
			catch (OntPropertyNotFoundException exc)
			{
				// error occured during search for object info, inform interlocutor
				OSMessage reply = new OSMessage();
				reply.setMessageCode(OSMessage.DBA_Extract_Object_Failed);
				try
				{
					this.osAgent.sendMessage(ACLMessage.FAILURE,reply,aidClient,
							conversationID);
				}
				catch (SendingMessageFailed e)
				{
					// the only thing we can do is to log information about error detils
					this.osAgent.onSendMessgeFailed(aidClient,
							OSMessage.DBA_Extract_Object_Failed,exc,null);
				}
				this.osAgent.logg(exc);
			}
			finally
			{
				this.model.leaveCriticalSection();
			}
		}

		// transfer object data to interlocutor
		OSMessage reply = new OSMessage();
		reply.setMessageCode(OSMessage.DBA_Extract_Object_Result);
		reply.setObjectInfo(objectInfo);
		try
		{
			this.osAgent.sendMessage(ACLMessage.INFORM,reply,aidClient,
					conversationID);
		}
		catch (SendingMessageFailed exc)
		{
			// the only thing we can do is to log information about error detils
			this.osAgent.onSendMessgeFailed(aidClient,
					OSMessage.DBA_Extract_Object_Result,exc,null);
		}
	}

	private void onExecuteRDQL (OSMessage msg, AID customerAID,
			String conversationID)
	{
		OSMessage reply = null;
		QueryResults results = null;

		// perform query and leave critical section as fast as possible
		try
		{
			// execute query
			this.model.enterReadCriticalSection();
			results = this.model.executeRDQLQuery(msg.getAnnouncement());
			this.osAgent.logg(Level.INFO,"DBA executed RDQL query"
					+ msg.getAnnouncement());

			// write formatted text to the temporary file
			String tmpName = IdentifierGenerator.getInstance().GenerateID();
			String filePath = Utils.getInstance().getCWD() + "\\tmp\\" + tmpName
					+ ".txt";
			File file = new File(filePath);
			FileOutputStream fileStream = new FileOutputStream(file);
			PrintWriter fileWriter = new PrintWriter(fileStream);
			QueryResultsFormatter formatter = new QueryResultsFormatter(results);
			formatter.printAll(fileWriter);
			fileWriter.close();

			// read contents of the file
			FileReader fileReader = new FileReader(filePath);
			BufferedReader buffReader = new BufferedReader(fileReader);
			StringBuffer stringBuff = new StringBuffer();
			while (true)
			{
				String strLine = buffReader.readLine();
				if (strLine == null)
				{
					break;
				}
				else
				{
					stringBuff.append(strLine);
					stringBuff.append('\n');
				}
			}
			String textResult = stringBuff.toString();

			// ok, we have result in text format so let's send it back to the client
			reply = new OSMessage();
			reply.setMessageCode(OSMessage.DBA_Execute_RDQL_Result);
			reply.setAnnouncement(textResult);
			this.osAgent.sendMessage(ACLMessage.INFORM,reply,customerAID,
					conversationID);

			// delete temporary file
			new File(filePath).delete();
		}
		catch (IdentifierGenerationFailedException exc)
		{
			exc.printStackTrace();
			this.sendErrorMessage(OSMessage.DBA_Execute_RDQL_Failed,
					"Can't generate name for temporary file",exc,customerAID,
					conversationID);
		}
		catch (FileNotFoundException exc)
		{
			exc.printStackTrace();
			this.sendErrorMessage(OSMessage.DBA_Execute_RDQL_Failed,
					"Can't open temporary file",exc,customerAID,conversationID);
		}
		catch (IOException exc)
		{
			exc.printStackTrace();
			this.sendErrorMessage(OSMessage.DBA_Execute_RDQL_Failed,
					"Error while reading temporary file",exc,customerAID,
					conversationID);
		}
		catch (SendingMessageFailed exc)
		{
			exc.printStackTrace();
			this.osAgent.onSendMessgeFailed(customerAID,
					OSMessage.DBA_Execute_RDQL_Result,exc,reply);
		}
		catch (TokenMgrError exc)
		{
			exc.printStackTrace();
			this.sendErrorMessage(OSMessage.DBA_Execute_RDQL_Failed,
					"Error during parsing RDQL query string: " + exc.getMessage(),
					null,customerAID,conversationID);
		}
		catch (QueryException exc)
		{
			exc.printStackTrace();
			this.sendErrorMessage(OSMessage.DBA_Execute_RDQL_Failed,
					"Error during parsing RDQL query string: ",exc,customerAID,
					conversationID);
		}
		finally
		{
			this.model.leaveCriticalSection();
		}
	}
}
