package logrus.agents.datasources.chefmoz;

import java.io.IOException;
import java.util.Iterator;

import logrus.agents.datasources.WebpageParseException;
import logrus.agents.datasources.LinkInfo;
import logrus.agents.datasources.ILinkExtactor;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.XMLAnalysisException;
import logrus.tools.queue.Queue;
import logrus.tools.web.WebUtils;

import org.cyberneko.html.parsers.DOMParser;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Extracts links form chefmoz pages
 */
public class ChefMozLinkExtactor implements ILinkExtactor
{
	private interface LinkProcessor
	{
		public void processLink (LinkInfo info) throws WebpageParseException;

		public Object processResults ();
		
		public void processRestaurant (LinkInfo info) throws WebpageParseException;
	}

	private class ChefmozInfoProcessor implements LinkProcessor
	{
		public void processLink (LinkInfo info) throws WebpageParseException
		{
			enqueue(info);
		}
		
		public void processRestaurant (LinkInfo info) throws WebpageParseException
		{
			ChefMozPropertiesExtractor extractor;
			try
			{
				extractor = new ChefMozPropertiesExtractor();
			}
			catch (IdentifierGenerationFailedException e)
			{
				throw new WebpageParseException(e);
			}
			try
			{
				extractor.extract(info.link);
			}
			catch (XMLAnalysisException e)
			{
				throw new WebpageParseException(e);
			}
		}
		
		public Object processResults ()
		{
			return null;
		}
	}
		
	private class CollectLinksProcessor implements LinkProcessor
	{
		public void processLink (LinkInfo info)
		{
			if (info.type == ChefMozPageTypes.RestaurantPage)
			{
				ChefMozLinkExtactor.this.results.inject(info.link);
			}
			else
			{
				enqueue(info);
			}
		}
		
		public void processRestaurant (LinkInfo info)
		{
			return;
		}

		public Object processResults ()
		{
			String[] links = new String[ChefMozLinkExtactor.this.results
					.count()];
			int i = 0;
			for (Iterator iter = ChefMozLinkExtactor.this.results
					.iterator(); iter.hasNext();)
			{
				links[i++] = (String)iter.next();
			}
			return links;
		}
	}

	public static String[] ProvincesCodes = { "DS", "KP", "LD", "LU", "LB",
			"MP", "MA", "OP", "PK", "PD", "PM", "SL", "SW", "WM", "WP", "ZP" };

	public static String[] BigCitySublinksPostfixes = { "0-9", "A", "B", "C",
			"D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q",
			"R", "S", "T", "U", "V", "W", "X", "Y", "Z" };

	// queue of pages to be parsed
	private Queue toProcess;

	Queue results;

	/**
	 * Initialize set of links as links to Polish provinces
	 */
	public ChefMozLinkExtactor ()
	{
		this(ChefMozLinkExtactor.buildProvinceLinkInfos());
	}

	/**
	 * @param startLinks
	 *           set of links to start crawling
	 */
	public ChefMozLinkExtactor (LinkInfo[] startLinks)
	{
		super();
		this.toProcess = new Queue();
		this.results = new Queue();
		for (int i = 0; i < startLinks.length; i++)
		{
			this.enqueue(startLinks[i]);
		}
	}

	/**
	 * Build set of links to Polish provinces
	 * 
	 * @return Set of links to Polish provinces
	 */
	public static LinkInfo[] buildProvinceLinkInfos ()
	{
		String[] ProvCodes = ChefMozLinkExtactor.ProvincesCodes;
		LinkInfo[] infos = new LinkInfo[ProvCodes.length];
		for (int i = 0; i < ProvCodes.length; i++)
		{
			infos[i] = new LinkInfo("http://chefmoz.org/Poland/" + ProvCodes[i]
					+ "/",ChefMozPageTypes.ProvincePage);
		}
		return infos;
	}

	/* (non-Javadoc)
	 * @see logrus.agents.datasources.chefmoz.RestaurantLinkExtactor#extractRestaurantLinks()
	 */
	public String[] extractRestaurantLinks () throws WebpageParseException
	{
		return (String[])this.processLinks(new CollectLinksProcessor());
	}
	
	/* (non-Javadoc)
	 * @see logrus.agents.datasources.chefmoz.RestaurantLinkExtactor#extractRestaurantInfo()
	 */
	public String[] extractRestaurantInfo () throws WebpageParseException
	{
		return (String[])this.processLinks(new ChefmozInfoProcessor());
	}
	
	private Object processLinks (LinkProcessor processor) throws WebpageParseException
	{
		this.results.clear();
		while (!this.toProcess.empty())
		{
			LinkInfo info = this.dequeue();
			switch (info.type)
			{
				case ChefMozPageTypes.RestaurantPage:
					System.out.println("\t\tProcessing restaurant " + info.link);
					this.parseRestaurantPage(info,processor);
					break;
				case ChefMozPageTypes.ProvincePage:
					System.out.println("PROCESSING PROVIENCE " + info.link);
					this.parseProvincePage(info,processor);
					break;
				case ChefMozPageTypes.CityPage:
					System.out.println("\tProcessing city " + info.link);
					this.parseCityPage(info,processor);
					break;
				case ChefMozPageTypes.CityLetterPage:
					System.out.println("\tProcessing subcity " + info.link);
					this.parseCityLetterPage(info,processor);
					break;
			}
		}
		return processor.processResults();
	}

	private void parseRestaurantPage (LinkInfo info, LinkProcessor processor) throws WebpageParseException
	{
		processor.processRestaurant(info);
	}

	private void parseCityLetterPage (LinkInfo info, LinkProcessor processor)
			throws WebpageParseException
	{
		String link = info.link;
		// load HTML document
		DOMParser parser = new DOMParser();
		try
		{
			parser.parse(link);
		}
		catch (SAXException e)
		{
			throw new WebpageParseException(e);
		}
		catch (IOException e)
		{
			throw new WebpageParseException(e);
		}
		Node root = parser.getDocument();
		WebUtils wu = WebUtils.getInstance();

		Node tmpRoot = root;
		for (int i = 0; tmpRoot != null; tmpRoot = wu.findTagN(root,"hr",
				"center",1 + i++))
		{
			tmpRoot = wu.findTagN(tmpRoot,"a","center",1);
			if (tmpRoot != null)
			{
				String uri = wu.extractLink(tmpRoot);
				if (uri.compareTo("/") != 0)
				{
					uri = "http://chefmoz.org" + wu.extractLink(tmpRoot);
					processor
							.processLink(new LinkInfo(uri,ChefMozPageTypes.RestaurantPage));
				}
			}
			else
			{
				break;
			}
		}
	}

	private void parseCityPage (LinkInfo info, LinkProcessor processor) throws WebpageParseException
	{
		String link = info.link;
		DOMParser parser = new DOMParser();
		try
		{
			parser.parse(link);
		}
		catch (SAXException e)
		{
			throw new WebpageParseException(e);
		}
		catch (IOException e)
		{
			throw new WebpageParseException(e);
		}
		Node root = parser.getDocument();
		WebUtils wu = WebUtils.getInstance();

		// first check if we're parsing small town or big city page
		Node aux = wu.findNextTag(root,"0-9","center");
		if (aux == null)
		{
			// small town
			Node tmpRoot = root;
			for (int i = 0; tmpRoot != null; tmpRoot = wu.findTagN(root,"hr",
					"center",1 + i++))
			{
				tmpRoot = wu.findTagN(tmpRoot,"a","center",1);
				if (tmpRoot != null)
				{
					String uri = wu.extractLink(tmpRoot);
					if (uri.compareTo("/") != 0)
					{
						uri = "http://chefmoz.org" + wu.extractLink(tmpRoot);
						processor.processLink(new LinkInfo(uri,
								ChefMozPageTypes.RestaurantPage));
					}
				}
				else
				{
					break;
				}
			}
		}
		else
		{
			// big city
			String[] postfix = ChefMozLinkExtactor.BigCitySublinksPostfixes;
			for (int i = 0; i < postfix.length; i++)
			{
				String uri = "http://chefmoz.org/cgi-bin/city.pl?sort=3&sortType=3&num_results=50&start=0&cat="
						+ link.substring(19).replaceAll("/","%2F") + postfix[i];
				processor.processLink(new LinkInfo(uri,ChefMozPageTypes.CityLetterPage));
			}
		}
	}

	private void parseProvincePage (LinkInfo info, LinkProcessor processor) throws WebpageParseException
	{
		String link = info.link;
		// load HTML document
		DOMParser parser = new DOMParser();
		try
		{
			parser.parse(link);
		}
		catch (SAXException e)
		{
			throw new WebpageParseException(e);
		}
		catch (IOException e)
		{
			throw new WebpageParseException(e);
		}
		Node root = parser.getDocument();
		WebUtils wu = WebUtils.getInstance();

		// search for table with links
		root = wu.findTagN(root,"table",2);
		root = wu.findTagN(root,"tr",1);
		NodeList childern = root.getChildNodes();
		for (int i = 0; i < childern.getLength(); i++)
		{
			if (wu.compareNode(childern.item(i),"td"))
			{
				NodeList rows = wu.findTagN(childern.item(i),"ul",1)
						.getChildNodes();
				for (int j = 0; j < rows.getLength(); j++)
				{
					if (wu.compareNode(rows.item(j),"li"))
					{
						String uri = "http://chefmoz.org"
								+ wu.extractLink(wu.findTagN(rows.item(j),"a",1));
						processor.processLink(new LinkInfo(uri,ChefMozPageTypes.CityPage));
					}
				}
			}
		}
	}

	/**
	 * Enqueue link to be processed
	 * 
	 * @param info
	 *           Link info
	 */
	void enqueue (LinkInfo info)
	{
		this.toProcess.injectUnique(info);
	}

	/**
	 * Get link to be parsed from queue
	 * 
	 * @return First link in queue to be parsed or null if all links have been processed
	 */
	private LinkInfo dequeue ()
	{
		return (LinkInfo)this.toProcess.pop();
	}
}
