package logrus.agents.datasources.chefmoz;

import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import logrus.agents.datasources.WebpageParseException;
import logrus.agents.datasources.DataProviderID;
import logrus.agents.datasources.IPropertiesExtractor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.AcceptedPaymentParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.AlcoholServiceParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.DressPolicyParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.FeaturesParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.HandicappedParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.ParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.ParkingParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.ReservationsParagraphProcessor;
import logrus.agents.datasources.chefmoz.paragraphProcessors.SmokingPolicyParagraphProcessor;
import logrus.ontmodel.Prefixes;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.restaurant.CuisineTypes;
import logrus.ontmodel.restaurant.tools.RestaurantCleaner;
import logrus.ontmodel.validation.ObjectValidator;
import logrus.ontmodel.validation.ValidationStatus;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.IdentifierGenerator;
import logrus.tools.Utils;
import logrus.tools.WeekDaysMappings;
import logrus.tools.XMLAnalysisException;
import logrus.tools.web.UnexpectedHTMLStructureException;
import logrus.tools.web.WebUtils;
import logrus.tools.web.visitors.NodeVisitor;

import org.cyberneko.html.parsers.DOMParser;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Extracts restaurant properties form ChefMoz
 */
public class ChefMozPropertiesExtractor implements IPropertiesExtractor
{
	class LocatorNameLocationParagraph extends NodeVisitor
	{
		// determinies whether beginning of the review form has been found
		boolean formNodeFound;

		// header of the review
		private Node reviewNode;

		// determinies whether table header in review form has been found
		boolean tableNodeFound;

		public LocatorNameLocationParagraph ()
		{
			this.reviewNode = null;
			this.formNodeFound = false;
			this.tableNodeFound = false;
		}

		public Object getResult ()
		{
			return this.reviewNode;
		}

		public boolean processNode (Node node)
		{
			if (!this.formNodeFound)
			{
				if (Utils.strIsEqual(node.getNodeName(),"form"))
				{
					this.formNodeFound = true;
				}
				return false;
			}
			if (!this.tableNodeFound)
			{
				if (Utils.strIsEqual(node.getNodeName(),"table"))
				{
					this.tableNodeFound = true;
				}
				return false;
			}
			if (Utils.strIsEqual(node.getNodeName(),"b"))
			{
				this.reviewNode = node;
				return true;
			}
			return false;
		}

		public void setResult (Object result)
		{
			this.reviewNode = (Node)result;
		}
	}

	// ID of processed restaurant
	private String uriRestaurant;

	// extracted data transformed into RDF statements
	private QuadsList result;

	// flag determinig if cuisine paragraph existed in processed document
	boolean wasCuisineParagraph;

	/**
	 * Default constructor
	 * 
	 * @throws IdentifierGenerationFailedException
	 */
	public ChefMozPropertiesExtractor ()
			throws IdentifierGenerationFailedException
	{
		super();
		this.setResult(new QuadsList());
		this.uriRestaurant = Prefixes.BaseRestaurantURI + "/"
				+ IdentifierGenerator.getInstance().GenerateID();
		this.wasCuisineParagraph = true;
	}

	/**
	 * @param uriRestaurant
	 */
	public ChefMozPropertiesExtractor (String uriRestaurant)
	{
		super();
		this.setResult(new QuadsList());
		this.uriRestaurant = Prefixes.BaseRestaurantURI + "/" + uriRestaurant;
		this.wasCuisineParagraph = true;
	}

	/**
	 * Check if <code>node</code> is not null and is node of type <code>expectedNode</code>
	 * 
	 * @param node
	 *           Node to be checked
	 * @param expectedType
	 *           Expected node type
	 * @throws UnexpectedHTMLStructureException
	 */
	public void assertNode (Node node, String expectedType)
			throws UnexpectedHTMLStructureException
	{
		if (node == null)
		{
			throw new UnexpectedHTMLStructureException("Node is null");
		}
		else if (node.getNodeName().toLowerCase().compareTo(expectedType) != 0)
		{
			throw new UnexpectedHTMLStructureException("Node type is "
					+ node.getNodeName() + " and expected was " + expectedType);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.agents.datasources.IPropertiesExtractor#extract(java.lang.String)
	 */	
	public QuadsList extract (String url) throws WebpageParseException, XMLAnalysisException
	{
		this.extract_inner(url);
		QuadsList quads = this.getResult();
		RestaurantCleaner.getInstance().cleanInfo(quads);
		switch (ObjectValidator.getInstance().validate(quads))
		{
			case ValidationStatus.Complete:
			case ValidationStatus.Incomplete:
				return quads;
			case ValidationStatus.Incorrect:
			case ValidationStatus.Unknown:
				return null;
		}
		return null;
	}	

	private QuadsList extract_inner (String url) throws WebpageParseException
	{
		try
		{
			// load HTML document
			DOMParser parser = new DOMParser();
			parser.parse(url);
			Node root = parser.getDocument();
			WebUtils utils = WebUtils.getInstance();

			// add metadata about information source
			this.result.add(this.uriRestaurant,Prefixes.SourceLink,url,
					DataProviderID.CHEFMOZ);

			// add RDFS class binding
			this.result.add(this.uriRestaurant,Prefixes.RDF_TYPE,
					Prefixes.CLASS_RES,DataProviderID.CHEFMOZ);

			// locate <body> node
			Node bodyTagNode = utils.findNextTag(root,"body");
			this.assertNode(bodyTagNode,"body");

			// locate left panel (second table)
			Node leftPanelRoot = utils.findTagN(root,"table",2);
			assertNode(leftPanelRoot,"table");

			// extract cuisine information
			this.parseCuisineParagraph(leftPanelRoot);

			// extract address information
			this.parseAddressParagraph(leftPanelRoot);

			// extract web information paragraph
			this.parseWebInformationParagraph(leftPanelRoot);

			// extract name/location information
			this.parseNameLocationParagraph(leftPanelRoot);

			// extract price information
			this.parsePriceParagraph(leftPanelRoot);

			// extract features information
			this.parseFeaturesParagraph(leftPanelRoot);

			// extract accepted payment methods information
			this.parseAcceptedPaymentParagraph(leftPanelRoot);

			// extract smoking policy information
			this.parseSmokingPolicyParagraph(leftPanelRoot);

			// extract dress policy information
			this.parseDressPolicyParagraph(leftPanelRoot);

			// extract reservations information
			this.parseReservationsParagraph(leftPanelRoot);

			// extract alcohol service information
			this.parseAlcoholServiceParagraph(leftPanelRoot);

			// extract parking information
			this.parseParkingParagraph(leftPanelRoot);

			// extract handicapped accessiblity information
			this.parseHandicappedParagraph(leftPanelRoot);

			// extract open hours information
			this.parseOpenHoursParagraph(leftPanelRoot);

			return this.result;
		}
		catch (SAXException e)
		{
			throw new WebpageParseException(e);
		}
		catch (UnexpectedHTMLStructureException e)
		{
			throw new WebpageParseException(e);
		}
		catch (IdentifierGenerationFailedException e)
		{
			throw new WebpageParseException(e);
		}
		catch (IOException e)
		{
			throw new WebpageParseException(e);
		}
	}

	/**
	 * Locate node with cuisuine information and extract it
	 * 
	 * @param node
	 *           Root of subtree from which searching starts
	 * @return
	 * @throws UnexpectedHTMLStructureException
	 * @throws IdentifierGenerationFailedException
	 */
	private Node extractCuisine (Node node)
			throws UnexpectedHTMLStructureException,
			IdentifierGenerationFailedException
	{
		Node tdNode = WebUtils.getInstance().findNextTag(node,"td","br");
		if (tdNode == null)
		{
			return null;
		}
		else
		{
			Node cuisineNode = tdNode.getFirstChild();
			if (cuisineNode == null)
			{
				throw new UnexpectedHTMLStructureException("No cuisine text");
			}
			else
			{
				String[] splited = cuisineNode.getNodeValue().split("/");
				for (int i = 0; i < splited.length; i++)
				{
					String strCuisine = Utils.strNormString(splited[i]);
					String[] parts = strCuisine.split(" ");
					StringBuffer buffer = new StringBuffer();
					for (int j = 0; j < parts.length; j++)
					{
						buffer.append(Utils.strToUpperAt(parts[j],0));
					}
					String cuisineName = buffer.toString() + "Cuisine";
					if (CuisineTypes.getInstance().isOnList(cuisineName))
					{
						String strPredicate = Prefixes.RES_CUISINE;
						String strObject = Prefixes.BaseRestaurantURI + "#"
								+ cuisineName;
						this.getResult().add(this.uriRestaurant,strPredicate,
								strObject,DataProviderID.CHEFMOZ);
					}
				}
			}
			return tdNode;
		}
	}

	/**
	 * @return Returns the result.
	 */
	public QuadsList getResult ()
	{
		return this.result;
	}

	private Node parseAcceptedPaymentParagraph (Node root)
	{
		AcceptedPaymentParagraphProcessor processor = new AcceptedPaymentParagraphProcessor();
		return this.parseParagraph(root,"Accepts",processor,false);
	}

	private Node parseAddressParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		Node tmpRoot = root;
		WebUtils utils = WebUtils.getInstance();
		// tmpRoot = utils.findTagN(tmpRoot,"br",2);
		tmpRoot = utils.findTagN(tmpRoot,"h2","form",1);
		if (tmpRoot == null)
		{
			return null;
		}
		tmpRoot = tmpRoot.getPreviousSibling();
		ArrayList lines = new ArrayList();
		while (tmpRoot.getNodeName().toLowerCase().compareTo("b") != 0
				&& tmpRoot.getNodeName().toLowerCase().compareTo("table") != 0)
		{
			if (tmpRoot.getNodeName().toLowerCase().compareTo("br") != 0)
			{
				tmpRoot = tmpRoot.getNextSibling();
				continue;
			}
			else
			{
				Node aux = tmpRoot.getNextSibling();
				if (utils.compareNode(aux,"br"))
				{
					break;
				}
				if (aux.getNodeName().compareTo("#text") == 0)
				{
					String content = WebUtils.getInstance().extractTextNodeContent(
							aux);
					if (content != null && content.length() > 0)
					{
						lines.add(content);
					}
				}
			}
			tmpRoot = tmpRoot.getNextSibling();
		}
		Object[] strLines = lines.toArray();
		String strNeighborhood = null, strStreetAddr = null, strZip = null;
		ArrayList strPhones = new ArrayList();
		int idx = strLines.length - 1;
		// first identify number of phones
		while (true)
		{
			if (((String)strLines[idx]).startsWith("+48"))
			{
				strPhones.add(strLines[idx--]);
			}
			else
			{
				break;
			}
		}
		String aux = (String)strLines[idx--];
		if (aux.length() >= 6 && Character.isDigit(aux.charAt(0))
				&& Character.isDigit(aux.charAt(1)) && aux.charAt(2) == '-'
				&& Character.isDigit(aux.charAt(3))
				&& Character.isDigit(aux.charAt(4))
				&& Character.isDigit(aux.charAt(5)))
		{
			strZip = aux.substring(0,6);
		}
		else
		{
			strZip = null;
		}
		strStreetAddr = (String)strLines[idx--];
		if (idx == 0)
		{
			strNeighborhood = (String)strLines[idx--];
		}

		// add statements
		// street address
		this.result.add(this.uriRestaurant,Prefixes.LOC_STREET_ADDR,
				strStreetAddr,DataProviderID.CHEFMOZ);
		// zip code
		this.result.add(this.uriRestaurant,Prefixes.LOC_ZIP,strZip,
				DataProviderID.CHEFMOZ);
		// phone
		for (Iterator iter = strPhones.iterator(); iter.hasNext();)
		{
			this.result.add(this.uriRestaurant,Prefixes.LOC_PHONE,(String)iter
					.next(),DataProviderID.CHEFMOZ);
		}
		// neighborhood
		if (strNeighborhood != null)
		{
			this.result.add(this.uriRestaurant,Prefixes.LOC_NEIGHBORHOOD,
					strNeighborhood,DataProviderID.CHEFMOZ);
		}
		return tmpRoot;
	}

	private Node parseAlcoholServiceParagraph (Node root)
	{
		AlcoholServiceParagraphProcessor processor = new AlcoholServiceParagraphProcessor();
		return this.parseParagraph(root,"Alcohol",processor,false);
	}

	private Node parseCuisineParagraph (Node root)
			throws UnexpectedHTMLStructureException,
			IdentifierGenerationFailedException
	{
		Node tmpRoot = root;
		Node cuisineTable = WebUtils.getInstance().findTagN(tmpRoot,"table",
				"form",2);
		if (cuisineTable == null
				|| !WebUtils.getInstance().matchesValue(cuisineTable,"cellpadding",
						"4"))
		{
			this.wasCuisineParagraph = false;
			return null;
		}
		tmpRoot = cuisineTable;
		/*
		 * // locate cuisine paragraph if
		 * (!WebUtils.getInstance().traverseHTMLDocument(tmpRoot,locator)) { throw new
		 * UnexpectedHTMLStructureException( "Can't locate cuisine paragraph"); } // extract cuisine
		 * information Node tableHeader = (Node)locator.getResult(); tmpRoot = tableHeader;
		 */
		boolean firstRun = true;
		while (tmpRoot != null)
		{
			if (firstRun)
			{
				tmpRoot = WebUtils.getInstance().findNextTag(tmpRoot,"tr","br");
				firstRun = false;
			}
			else
			{
				tmpRoot = WebUtils.getInstance().findTagN(tmpRoot,"tr","br",2);
			}
			if (tmpRoot != null)
			{
				this.extractCuisine(tmpRoot);
				tmpRoot = tmpRoot.getNextSibling();
			}
		}
		return cuisineTable;
		// return tableHeader.getNextSibling();
	}

	private Node parseDressPolicyParagraph (Node root)
	{
		DressPolicyParagraphProcessor processor = new DressPolicyParagraphProcessor();
		return this.parseParagraph(root,"Dress",processor,false);
	}

	private Node parseFeaturesParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		FeaturesParagraphProcessor processor = new FeaturesParagraphProcessor();
		return this.parseParagraph(root,"Features",processor,false);
	}

	// extract parking information
	private Node parseHandicappedParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		HandicappedParagraphProcessor processor = new HandicappedParagraphProcessor();
		return this.parseParagraph(root,"Handicapped",processor,true);
	}

	/**
	 * Parse open hours
	 * 
	 * @param lines
	 *           Lines describing open hours
	 * @return Array of seven strings with open hours, one for each day of week
	 */
	public String[] parseHours (ArrayList lines)
	{
		String[] hours = new String[7];
		for (Iterator iter = lines.iterator(); iter.hasNext();)
		{
			String strLine = (String)iter.next();
			// daily prefix => fill all days with the same period
			if (strLine.startsWith("daily"))
			{
				String aux = strLine.substring(5).trim();
				for (int j = 0; j < 7; j++)
				{
					hours[j] = aux;
				}
			}

			// opened 24h => fill all days with the same period (0-24)
			if (strLine.compareTo("24h") == 0)
			{
				for (int j = 0; j < 7; j++)
				{
					hours[j] = "24h";
				}
			}

			// Day-Day OpenHours pattern or Day OpenHours
			String regex = "Mon|Tue|Wed|Thu|Fri|Sat|Sun";
			Pattern pattern = Pattern.compile(regex);
			Matcher matcher = pattern.matcher(strLine.substring(0,3));
			int iStart = -1, iEnd = -1;
			if (matcher.find())
			{
				iStart = iEnd = WeekDaysMappings.getInstance().getIndex(
						strLine.substring(0,3));
				if (strLine.length() > 7)
				{
					matcher = pattern.matcher(strLine.substring(4,7));
					if (matcher.find())
					{
						iEnd = WeekDaysMappings.getInstance().getIndex(
								strLine.substring(4,7));
					}
					String aux = null;
					if (iEnd != iStart)
					{
						aux = strLine.substring(7).trim();
					}
					else
					{
						aux = strLine.substring(3).trim();
					}
					for (int j = iStart; j <= iEnd; j++)
					{
						hours[j] = aux;
					}
				}
			}

			int start, end;
		}
		return hours;
	}

	private Node parseNameLocationParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		LocatorNameLocationParagraph locator = new LocatorNameLocationParagraph();
		Node tmpRoot = root;
		if (!WebUtils.getInstance().traverseHTMLDocument(tmpRoot,locator))
		{
			throw new UnexpectedHTMLStructureException(
					"Can't locate name/location paragraph");
		}
		tmpRoot = (Node)locator.getResult();
		this.assertNode(tmpRoot,"b");
		Node textNode = tmpRoot.getFirstChild();
		if (textNode == null)
		{
			throw new UnexpectedHTMLStructureException(
					"No text in name/location paragraph");
		}
		String infoString = textNode.toString().substring(15,
				textNode.toString().length() - 1);
		String[] tmpArr = infoString.split(" in ");

		String strRestName = null, strCountry = null, strRegion = null, strCity = null;
		// parse restaurant name
		strRestName = tmpArr[0].trim().split(",")[0].trim();

		// parse location part
		if (tmpArr.length > 1)
		{
			tmpArr = tmpArr[1].split("/");
			strCountry = tmpArr[0].trim();
			// we can have 2 parts (Country/City)
			if (tmpArr.length == 2)
			{
				strRegion = null;
				strCity = strCity = tmpArr[1].trim();
			}
			// or we can have 3 parts (Country/Region/City)
			if (tmpArr.length == 3)
			{
				strRegion = tmpArr[1].trim();
				strCity = tmpArr[2].trim();
			}
		}
		// add statements to the result set
		// restaurant name
		this.result.add(this.uriRestaurant,Prefixes.RES_NAME,strRestName,
				DataProviderID.CHEFMOZ);
		// location: country
		if (strCountry != null)
		{
			this.result.add(this.uriRestaurant,Prefixes.LOC_COUNTRY,strCountry,
					DataProviderID.CHEFMOZ);
		}
		// location: city
		if (strCity != null)
		{
			this.result.add(this.uriRestaurant,Prefixes.LOC_CITY,strCity,
					DataProviderID.CHEFMOZ);
		}
		// location: region (if extracted)
		if (strRegion != null)
		{
			this.result.add(this.uriRestaurant,Prefixes.LOC_REGION,strRegion,
					DataProviderID.CHEFMOZ);
		}

		return tmpRoot;
	}

	private Node parseOpenHoursParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		ArrayList lines = new ArrayList();
		Node tmpRoot = WebUtils.getInstance().findTextTagPrefixN(root,"Hours",
				"form",1);
		if (tmpRoot == null)
		{
			return null;
		}
		tmpRoot = tmpRoot.getParentNode().getNextSibling();
		if (tmpRoot != null)
		{
			for (Node auxRoot = tmpRoot; auxRoot != null; auxRoot = auxRoot
					.getNextSibling())
			{
				if (WebUtils.getInstance().compareNode(auxRoot,"b"))
				{
					break;
				}
				else if (auxRoot.getNodeName().compareTo("#text") == 0)
				{
					String strLine = WebUtils.getInstance().extractTextNodeContent(
							auxRoot);
					if (strLine.length() > 0)
					{
						lines.add(strLine);
					}
				}
			}
		}
		String[] daysHr = this.parseHours(lines);
		StringBuffer buffer = new StringBuffer();
		for (int i = 0; i < daysHr.length; i++)
		{
			if (daysHr[i] != null)
			{
				buffer.append(daysHr[i]);
			}
			if (i != 6)
			{
				buffer.append('|');
			}
		}
		String parsedHours = buffer.toString();
		if (parsedHours.compareTo("||||||") != 0)
		{
			this.result.add(this.uriRestaurant,Prefixes.RES_PAR_HR,parsedHours,
					DataProviderID.CHEFMOZ);
		}
		return root;
	}

	/**
	 * Process chefmoz page paragraph
	 * 
	 * @param root
	 *           Start searching for specified paragraph form this node
	 * @param headerName
	 *           Name of paragraph header
	 * @param processor
	 *           Processor used for parsing textual content
	 * @return Last processed node
	 */
	private Node parseParagraph (Node root, String headerName,
			ParagraphProcessor processor, boolean searchTextPrefix)
	{
		WebUtils webUtils = WebUtils.getInstance();
		Node tmpRoot = null;
		if (!searchTextPrefix)
		{
			tmpRoot = webUtils.findNextTag(root,headerName,"form");
		}
		else
		{
			tmpRoot = webUtils.findTextTagPrefixN(root,headerName,"form",1);
		}
		if (tmpRoot == null)
		{
			return root;
		}
		tmpRoot = tmpRoot.getParentNode().getParentNode();
		boolean prevBR = false;
		while (true)
		{
			Node brNode = tmpRoot.getNextSibling();
			if (webUtils.compareNode(brNode,""))
			{
				tmpRoot = brNode;
				continue;
			}
			if (!webUtils.compareNode(brNode,"br"))
			{
				break;
			}
			Node contentNode = brNode.getNextSibling();
			if (contentNode != null
					&& contentNode.getNodeName().compareTo("#text") == 0)
			{
				String content = webUtils.extractTextNodeContent(contentNode);
				if (content == null || content.length() == 0)
				{
					break;
				}
				processor.process(content,this.result,this.uriRestaurant);
				tmpRoot = contentNode;
			}
			else if (webUtils.compareNode(brNode,"br"))
			{
				break;
			}
		}
		return root;
	}

	// extract parking information
	private Node parseParkingParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		ParkingParagraphProcessor processor = new ParkingParagraphProcessor();
		return this.parseParagraph(root,"Parking",processor,false);
	}

	private Node parsePriceParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		WebUtils webUtils = WebUtils.getInstance();
		Node tableNode = webUtils.findTagN(root,"table","form",3);
		if (tableNode != null)
		{
			NodeList children = tableNode.getChildNodes();
			for (int i = 0; i < children.getLength(); i++)
			{
				Node row = children.item(i);
				if (webUtils.compareNode(row,"tr"))
				{
					Node b1 = webUtils.findTagN(row,"b",1);
					String mealName = webUtils.extractTextNodeContent(b1
							.getNextSibling());
					mealName = mealName.substring(1,mealName.length() - 1);
					mealName = Prefixes.BaseRestaurantURI + "#" + mealName + "Price";
					Node f2 = webUtils.findTagN(row,"font",2);
					String priceStr = webUtils.extractTextNodeContent(f2
							.getFirstChild());
					String fuzzyPrice = Prefixes.BaseMoneyURI + "#";
					if (priceStr.compareTo("gr") == 0)
					{
						fuzzyPrice += "VeryCheap";
					}
					if (priceStr.compareTo("z") == 0)
					{
						fuzzyPrice += "Inexpensive";
					}
					else if (priceStr.compareTo("zz") == 0)
					{
						fuzzyPrice += "Moderate";
					}
					else if (priceStr.compareTo("zzz") == 0)
					{
						fuzzyPrice += "Expensive";
					}
					else if (priceStr.compareTo("zzzz") == 0)
					{
						fuzzyPrice += "VeryExpensive";
					}
					this.result.add(this.uriRestaurant,mealName,fuzzyPrice,
							DataProviderID.CHEFMOZ);
				}
			}
		}
		return tableNode;
	}

	// extract reservations information
	private Node parseReservationsParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		ReservationsParagraphProcessor processor = new ReservationsParagraphProcessor();
		return this.parseParagraph(root,"Reservations",processor,false);
	}

	private Node parseSmokingPolicyParagraph (Node root)
	{
		SmokingPolicyParagraphProcessor processor = new SmokingPolicyParagraphProcessor();
		return this.parseParagraph(root,"Smoking",processor,false);
	}

	private Node parseWebInformationParagraph (Node root)
			throws UnexpectedHTMLStructureException
	{
		Node tmpRoot = WebUtils.getInstance().findNextTag(root,"Web Information",
				"form");
		if (tmpRoot == null)
		{
			return root;
		}
		tmpRoot = tmpRoot.getParentNode();
		boolean wasWWW = false, wasEmail = false;
		if (tmpRoot != null)
		{
			for (Node auxRoot = tmpRoot; auxRoot != null
					&& auxRoot.getNodeName().compareTo("table") != 0; auxRoot = auxRoot
					.getNextSibling())
			{

				if (WebUtils.getInstance().compareNode(auxRoot,"a"))
				{
					String parsedURI = WebUtils.getInstance().extractLink(auxRoot);
					if (parsedURI.startsWith("http://") && !wasWWW)
					{
						wasWWW = true;
						this.result.add(this.uriRestaurant,Prefixes.RES_WWW,
								parsedURI,DataProviderID.CHEFMOZ);
					}
					else if (parsedURI.startsWith("mailto:") && !wasEmail)
					{
						this.result.add(this.uriRestaurant,Prefixes.RES_EMAIL,
								parsedURI.substring(7,parsedURI.length()),
								DataProviderID.CHEFMOZ);
						break;
					}
				}
			}
		}
		return tmpRoot;
	}

	/**
	 * @param result
	 *           The result to set.
	 */
	public void setResult (QuadsList result)
	{
		this.result = result;
	}
}
