package logrus.agents.datasources.wrapperAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

import java.util.logging.Level;

import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSCyclicBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.datasources.WebpageParseException;
import logrus.agents.datasources.DataProviderID;
import logrus.agents.datasources.LinkInfo;
import logrus.agents.datasources.chefmoz.ChefMozLinkExtactor;
import logrus.agents.datasources.chefmoz.ChefMozPropertiesExtractor;
import logrus.agents.datasources.restauracjecom.RestComLinkExtractor;
import logrus.agents.datasources.restauracjecom.RestComPageTypes;
import logrus.agents.datasources.restauracjecom.RestComPropertiesExtractor;
import logrus.agents.datasources.wrapperAgent.WrapperAgent;
import logrus.agents.searchCoordinatorAgent.SearchTask;
import logrus.ontmodel.Prefixes;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.randomization.DataRandomizerFactory;
import logrus.ontmodel.randomization.IDataRandomizer;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.XMLAnalysisException;
import logrus.tools.settings.SettingsManager;

public class ServiceWABehaviour extends OSCyclicBehaviour
{
	WrapperAgent agentWA;

	public ServiceWABehaviour (Agent a)
	{
		super(a);
		this.agentWA = (WrapperAgent)a;
	}

	public void action ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.WA_Extract_Object_Info);
		matcher.addAcceptedMsgCode(OSMessage.WA_Crawl);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// receive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			SearchTask task = msg.getSearchCommand();
			String dataSourceID = task.getDataSource();
			switch (msg.getMessageCode())
			{
				case OSMessage.WA_Extract_Object_Info:
					this.agentWA.logg(Level.INFO,"WA " + this.agentWA.getWANumber()
							+ " parsing webpage " + task.getUrlToParse());
					if (dataSourceID.compareTo(DataProviderID.CHEFMOZ) == 0)
					{
						// run ChefMoz parser
						this.onExtractInfoChefmoz(task);
					}
					else if (dataSourceID.compareTo(DataProviderID.RESTCOM) == 0)
					{
						// run Restauracje.com parser
						this.onExtractInfoRC(task);
					}
					this.agentWA.logg(Level.INFO,"WA " + this.agentWA.getWANumber()
							+ " parsed webpage " + task.getUrlToParse());
					break;
				case OSMessage.WA_Crawl:
					this.agentWA.logg(Level.INFO,"WA " + this.agentWA.getWANumber()
							+ " crawling webpage " + task.getUrlToParse());
					if (dataSourceID.compareTo(DataProviderID.CHEFMOZ) == 0)
					{
						// run ChefMoz crawler
						this.onCrawlChefmoz(task);
					}
					else if (dataSourceID.compareTo(DataProviderID.RESTCOM) == 0)
					{
						// run restauracje.com crawler
						this.onCrawlRC(task);
					}
					break;
				default:
					this.osAgent.logg(Level.WARNING,
							"WA received unknown message type: "
									+ msg.getMessageCode());
					break;
			}
		}
		else
		{
			this.block(1000);
		}
	}

	private void onCrawlChefmoz (SearchTask task)
	{
		// run ChefMoz crawler
		try
		{
			LinkInfo[] linkInfos = new LinkInfo[] { new LinkInfo(task
					.getUrlToParse(),Integer.parseInt(task.getCommandOptions())) };
			ChefMozLinkExtactor extractorCM = new ChefMozLinkExtactor(linkInfos);

			// ok, extractor created - now run extraction process
			String[] linksRest = extractorCM.extractRestaurantLinks();
			this.agentWA.logg(Level.INFO,"WA " + this.agentWA.getWANumber()
					+ " crawled webpage " + task.getUrlToParse() + " and extracted "
					+ linksRest.length + " restaurant links");
			// we have some list of links, send results back to the SCA using specialized behaviour
			this.agentWA.addBehaviour(new SendLinksSetWABehaviour(this.agentWA,
					this.agentWA.getAidSCA(),task.getConversationID(),linksRest));
		}
		catch (WebpageParseException exc)
		{
			// error: internal ChefMoz parser error
			this.onError(OSMessage.WA_Crawl_Failed,
					"Internal ChefMoz parser error",exc,task.getConversationID());
		}
		catch (Exception exc)
		{
			// error: internal ChefMoz parser error
			this.onError(OSMessage.WA_Crawl_Failed,
					"Internal ChefMoz parser error",exc,task.getConversationID());
		}
	}

	private void onCrawlRC (SearchTask task)
	{
		// run RC crawler
		try
		{
			// check if restaurant ID has been specified
			LinkInfo[] linkInfos = new LinkInfo[] { new LinkInfo(task
					.getUrlToParse(),RestComPageTypes.ListPage) };
			RestComLinkExtractor extractorRC = new RestComLinkExtractor(linkInfos);

			// ok, extractor created - now run extraction process
			String[] linksRest = extractorRC.extractRestaurantLinks();

			// we have some list of links, send results back to the SCA using specialized behaviour
			this.agentWA.addBehaviour(new SendLinksSetWABehaviour(this.agentWA,
					this.agentWA.getAidSCA(),task.getConversationID(),linksRest));
		}
		catch (WebpageParseException exc)
		{
			// error: internal restauracje.com parser error
			this.onError(OSMessage.WA_Crawl_Failed,
					"Internal restauracje.com parser error",exc,task
							.getConversationID());
		}
		catch (Exception exc)
		{
			// error: internal restauracje.com parser error
			this.onError(OSMessage.WA_Crawl_Failed,
					"Internal restauracje.com parser error",exc,task
							.getConversationID());
		}
	}

	/**
	 * Executes parsing single restaurant webpage from restauracje.com
	 * 
	 * @param task
	 *           Search task to be performed
	 */
	private void onExtractInfoRC (SearchTask task)
	{
		// run RC parser
		try
		{
			// check if restaurant ID has been specified
			String restaurantID = task.getCommandOptions();
			RestComPropertiesExtractor extractorRC = null;
			if (restaurantID == null)
			{
				extractorRC = new RestComPropertiesExtractor();
			}
			else
			{
				extractorRC = new RestComPropertiesExtractor(restaurantID);
			}

			// ok, extractor created - now run extraction process
			QuadsList result = extractorRC.extract(task.getUrlToParse());
			
			// randomize 
			if (SettingsManager.getSettings().isUseDataRandomization())
			{
				IDataRandomizer dataRandomizer = DataRandomizerFactory
						.getInstance().createDataRandomizer(Prefixes.CLASS_RES);
				dataRandomizer.randomize(result,DataProviderID.RESTCOM);
			}

			// parsing URL has been succesful, send results back to the SCA using specialized behaviour
			this.agentWA.addBehaviour(new SendObjectInfoWABehaviour(this.agentWA,
					result,this.agentWA.getAidSCA(),task.getConversationID()));
		}
		catch (IdentifierGenerationFailedException exc)
		{
			// error: initializing restauracje.com parser has failed, inform SCA
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Can't generate restaurant ID",exc,task.getConversationID());
		}
		catch (WebpageParseException exc)
		{
			// error: internal restauracje.com parser error
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Internal restauracje.com parser error",exc,task
							.getConversationID());
		}
		catch (XMLAnalysisException exc)
		{
			// error: internal restauracje.com parser error
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Internal restauracje.com parser error",exc,task
							.getConversationID());
		}
	}

	/**
	 * Executes parsing single restaurant webpage from ChefMoz
	 * 
	 * @param task
	 *           Search task to be performed
	 */
	private void onExtractInfoChefmoz (SearchTask task)
	{
		// run ChefMoz parser
		try
		{
			// check if restaurant ID has been specified
			String restaurantID = task.getCommandOptions();
			ChefMozPropertiesExtractor extractorCM = null;
			if (restaurantID == null)
			{
				extractorCM = new ChefMozPropertiesExtractor();
			}
			else
			{
				extractorCM = new ChefMozPropertiesExtractor(restaurantID);
			}

			// ok, extractor created - now run extraction process
			QuadsList result = extractorCM.extract(task.getUrlToParse());

			// randomize 
			if (SettingsManager.getSettings().isUseDataRandomization())
			{
				IDataRandomizer dataRandomizer = DataRandomizerFactory
						.getInstance().createDataRandomizer(Prefixes.CLASS_RES);
				dataRandomizer.randomize(result,DataProviderID.CHEFMOZ);
			}

			// parsing URL has been succesful, send results back to the SCA using specialized behaviour
			this.agentWA.addBehaviour(new SendObjectInfoWABehaviour(this.agentWA,
					result,this.agentWA.getAidSCA(),task.getConversationID()));
		}
		catch (IdentifierGenerationFailedException exc)
		{
			// error: initializing ChefMoz parser has failed, inform SCA
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Can't generate restaurant ID",exc,task.getConversationID());
		}
		catch (WebpageParseException exc)
		{
			// error: internal ChefMoz parser error
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Internal ChefMoz parser error",exc,task.getConversationID());
		}
		catch (XMLAnalysisException exc)
		{
			// error: internal ChefMoz parser error
			this.onError(OSMessage.WA_Extract_Object_Info_Failed,
					"Internal ChefMoz parser error",exc,task.getConversationID());
		}
	}

	private void onError (int errorMessageCode, String announcement,
			Exception exc, String conversationID)
	{
		OSMessage msg = new OSMessage();
		msg.setMessageCode(errorMessageCode);
		msg.setAnnouncement(announcement + " : " + exc.getMessage());
		try
		{
			this.agentWA.sendMessage(ACLMessage.FAILURE,msg,this.agentWA
					.getAidSCA(),conversationID);
		}
		catch (SendingMessageFailed excSend)
		{
			// the only thing we can do is to log error :(
			this.agentWA.onSendMessgeFailed(this.agentWA.getAidSCA(),
					errorMessageCode,excSend,msg);
		}
	}
}
