package logrus.agents.gui.mainDialogAgent.behaviours;

import java.util.logging.Level;

import jade.content.OntoACLMessage;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.gui.mainDialogAgent.MainDialogAgent;
import logrus.agents.searchCoordinatorAgent.SearchTask;
import logrus.tools.IdentifierGenerationFailedException;
import logrus.tools.IdentifierGenerator;

/**
 * 
 */
public class CrawlRequestBehaviour extends OSBehaviour
{
	// GUI agent reference
	private MainDialogAgent agentGUI;

	// ID of provider that should be used for extracting information form webpageURI
	private String providerID;

	// URL of start page of crawl
	private String webpageURL;

	// conversation ID for whole process
	private String conversationID;

	// flag determining if this behaviour is finished
	private boolean finished;

	// current stage of behaviour
	private int stage;

	// index of stage responsible for sending crawl request to MA
	private static final int StageSendCrawlRequest = 0;

	// index of stage responsible for waiting for reply from MA
	private static final int StageWaitForReply = 1;

	// index of final stage of this behaviour
	private static final int StageFinal = 10;

	// error message
	private String strErrorMessage;
	
	// search options
	private int options;

	/**
	 * @param a
	 *           Reference to the GUI agent
	 * @param provider
	 *           ID of provider that should be used for extracting information form
	 *           <code>webpageURL</code>
	 * @param webpageURL
	 *           URL of start page of crawl
	 */
	public CrawlRequestBehaviour (Agent a, String providerID, String webpageURL,int options)
	{
		super(a);
		this.agentGUI = (MainDialogAgent)a;
		this.providerID = providerID;
		this.webpageURL = webpageURL;
		this.stage = StageSendCrawlRequest;
		this.strErrorMessage = null;
		this.conversationID = null;
		this.options = options;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		switch (this.stage)
		{
			case StageSendCrawlRequest:
				this.sendCrawlRequest();
				break;
			case StageWaitForReply:
				this.waitForReply();
				break;
			case StageFinal:
				this.finalizeBehaviour();
				break;
		}
	}

	/**
	 * Wait for reply from MA
	 */
	private void waitForReply ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.MA_Crawl_Request_OK);
		matcher.addAcceptedMsgCode(OSMessage.MA_Crawl_Request_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			switch (msg.getMessageCode())
			{
				case OSMessage.MA_Crawl_Request_OK:
					this.agentGUI.displayMessage("Crawl OK: "
							+ msg.getAnnouncement());
					break;
				case OSMessage.MA_Crawl_Request_Failed:
					this.agentGUI.displayMessage("Crawl Failed: "
							+ msg.getAnnouncement());
					break;
			}
		}
		else
		{
			// block behaviour until message arrives
			block();
		}
	}

	/**
	 * Send MA_Crawl_Request to MA
	 */
	private void sendCrawlRequest ()
	{
		// prepare conversationID
		try
		{
			this.conversationID = IdentifierGenerator.getInstance().GenerateID();
		}
		catch (IdentifierGenerationFailedException exc)
		{
			this.strErrorMessage = "Couldn't generate ID for conversation "
					+ exc.getMessage();
			this.stage = StageFinal;
		}

		// prepare crawl task
		SearchTask task = new SearchTask();
		task.setSearchType(SearchTask.EXTRACT_LINKS);
		task.setDataSource(this.providerID);
		task.setUrlToParse(this.webpageURL);
		task.setConversationID(this.conversationID);
		task.setCommandOptions(new Integer(this.options).toString());

		// prepare message
		OSMessage msg = new OSMessage();
		msg.setMessageCode(OSMessage.MA_Crawl_Request);
		msg.setSearchCommand(task);

		// send message
		try
		{
			this.agentGUI.sendMessage(ACLMessage.REQUEST,msg,this.agentGUI
					.getAidMA(),this.conversationID);
			this.stage = StageWaitForReply;
		}
		catch (SendingMessageFailed exc)
		{
			this.agentGUI.onSendMessgeFailed(this.agentGUI.getAidMA(),
					OSMessage.MA_Crawl_Request,exc,msg);
			this.strErrorMessage = "Couldn't send message MA_Crawl_Request to MA"
					+ exc.getMessage();
			this.stage = StageFinal;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#done()
	 */
	public boolean done ()
	{
		return this.finished;
	}

	/**
	 * Perform final actions before finishing behaviour
	 */
	private void finalizeBehaviour ()
	{
		// if there were some errors then log them
		if (this.strErrorMessage != null)
		{
			this.agentGUI.logg(Level.SEVERE,this.strErrorMessage);
			this.agentGUI.displayMessage(this.strErrorMessage);
		}

		// mark behaviour as finished
		this.finished = true;
	}
}
