package logrus.agents.indexingAgent;

import jade.content.onto.OntologyException;
import jade.core.AID;
import jade.domain.FIPAException;
import logrus.agents.OntoStorageAgent;
import logrus.agents.ServiceNames;
import logrus.agents.indexingAgent.behaviours.BootIndexingAgentBehaviour;
import logrus.agents.indexingAgent.behaviours.SchedulerBehaviour;
import logrus.agents.ontology.OntoStorageOntology;
import logrus.tools.logging.InitializeLoggerFailedException;
import logrus.tools.queue.Queue;

/**
 * 
 */
public class IndexingAgent extends OntoStorageAgent
{
	// AID of manager agent
	private AID aidMA;

	// AID of database agent
	private AID aidDBA;

	// Queue of URLs of indexing tasks waiting to be performed
	private Queue tasksQueue;

	// number of supervision behaviours that can be concurrently run in this moment
	private int freeSlots;

	// tasksScheduler for indexing tasks
	private SchedulerBehaviour tasksScheduler;

	/**
	 * @throws InitializeLoggerFailedException
	 * @throws OntologyException
	 */
	public IndexingAgent (AID aidManagerAg, AID aidDatabaseAg)
			throws InitializeLoggerFailedException, OntologyException
	{
		super();
		this.setAidDBA(aidDatabaseAg);
		this.setAidMA(aidManagerAg);
		this.freeSlots = 1;
		this.tasksScheduler = null;
		this.tasksQueue = new Queue();
	}

	/**
	 * Decrease number of free slots for behaviour by one
	 */
	public void decreaseFreeSlots ()
	{
		--this.freeSlots;
	}

	/**
	 * Increase number of free slots for behaviour by one
	 */
	public void increaseFreeSlots ()
	{
		++this.freeSlots;
	}

	/**
	 * Check if there are some tasks waiting to be performed
	 * 
	 * @return
	 */
	public boolean isTaskToPerform ()
	{
		return this.tasksQueue.count() > 0;
	}

	/**
	 * Get number of tasks awaiting in queue
	 * 
	 * @return Number of tasks awaiting in queue
	 */
	public int getNumberAwaitingTasks ()
	{
		return this.tasksQueue.count();
	}

	/**
	 * Check if there are some free slots for supervision behaviours right now
	 * 
	 * @return
	 */
	public boolean isFreeSubBehaviour ()
	{
		return this.freeSlots > 0;
	}

	/**
	 * Enqueue indexing task to be performed
	 * 
	 * @param task
	 *           Task to be performed
	 */
	public void enqueueTask (IndexingTask task)
	{
		this.tasksQueue.inject(task);
	}

	/**
	 * Dequeue oldest indexing task
	 * 
	 * @return Oldest indexing task waiting in queue or null if queue is empty
	 */
	public IndexingTask dequeueTask ()
	{
		if (this.tasksQueue.count() > 0)
		{
			return (IndexingTask)this.tasksQueue.pop();
		}
		else
		{
			return null;
		}
	}

	public void setup ()
	{
		this.addBehaviour(new BootIndexingAgentBehaviour(this));
		try
		{
			this.registerService(ServiceNames.ServiceIA,
					OntoStorageOntology.ONTOLOGY_NAME);
		}
		catch (FIPAException exc)
		{
			this.logg("IA couldn't register service",exc);
		}
	}

	/**
	 * @param aidMA
	 *           The aidMA to set.
	 */
	public void setAidMA (AID aidManagerAg)
	{
		this.aidMA = aidManagerAg;
	}

	/**
	 * @return Returns the aidMA.
	 */
	public AID getAidMA ()
	{
		return this.aidMA;
	}

	/**
	 * @param aidDBA
	 *           The aidDBA to set.
	 */
	public void setAidDBA (AID aidDatabaseAg)
	{
		this.aidDBA = aidDatabaseAg;
	}

	/**
	 * @return Returns the aidDBA.
	 */
	public AID getAidDBA ()
	{
		return this.aidDBA;
	}

	/**
	 * @param tasksQueue
	 *           The tasksQueue to set.
	 */
	public void setTasksQueue (Queue tasksQueue)
	{
		this.tasksQueue = tasksQueue;
	}

	/**
	 * @return Returns the tasksQueue.
	 */
	public Queue getTasksQueue ()
	{
		return this.tasksQueue;
	}

	/**
	 * @param freeSlots
	 *           The freeSlots to set.
	 */
	public void setFreeSlots (int freeBehaviours)
	{
		this.freeSlots = freeBehaviours;
	}

	/**
	 * @return Returns the freeSlots.
	 */
	public int getFreeSlots ()
	{
		return this.freeSlots;
	}

	/**
	 * @param tasksScheduler
	 *           The tasksScheduler to set.
	 */
	public void setTasksScheduler (SchedulerBehaviour scheduler)
	{
		this.tasksScheduler = scheduler;
	}

	/**
	 * @return Returns the tasksScheduler.
	 */
	public SchedulerBehaviour getTasksScheduler ()
	{
		return this.tasksScheduler;
	}

	/**
	 * Wakes behavior responsible for scheduling indexing tasks
	 */
	public void wakeTasksScheduler ()
	{
		if (this.tasksScheduler != null)
		{
			this.tasksScheduler.restart();
		}
	}
}
