package logrus.agents.indexingAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.core.AID;
import jade.core.Agent;
import jade.domain.FIPAException;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

import java.text.ParseException;
import java.util.logging.Level;

import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.indexingAgent.IndexingAgent;
import logrus.agents.indexingAgent.IndexingTask;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.merging.MergingReport;
import logrus.ontmodel.merging.Unificator;
import logrus.ontmodel.merging.UnionException;
import logrus.ontmodel.validation.ObjectValidator;
import logrus.ontmodel.validation.ValidationStatus;
import logrus.tools.Utils;
import logrus.tools.XMLAnalysisException;

/**
 * 
 */
public class IndexingBehaviour extends OSBehaviour
{
	// flag determinig if behaviour should be finished
	private boolean finished;

	// reference to indexing agent
	private IndexingAgent agentIA;

	// current stage of this behaviour
	private int stage;

	// index of stage responsible for validating task (indexed object)
	private static final int StageTaskValidation = 0;

	// index of stage responsible for sending query to DBA
	private static final int StageSendQueryToDBA = 1;

	// index of stage responsible for waiting for response from DBA
	private static final int StageWaitForReplyFromDBA = 2;

	// index of stage responsible for unification old information with new one
	private static final int StageUnification = 3;

	// index of stage responsible for sending merged object to DBA
	private static final int StageSendInjectDBA = 4;

	// index of stage responsible for waiting for DBA's reply on injecting data
	private static final int StageWaitInjectReply = 5;

	// index of stage responsible for cleaning after error
	private static final int StageFinal = 10;

	// indexing task to be performed
	private IndexingTask task;

	// object beeing indexed, after successful validation
	private QuadsList objIncl;

	// object extracted form model and describing the same physical object as the one that is
	// currently beeing indexed
	private QuadsList objRef;

	// merged object, ready to be sent to DBA
	private QuadsList objMerged;

	// error message used in StageFinal
	private String strErrorMessage;

	// ID of whole conversation related to performing indexing task
	private String conversationID;

	// old timestamps
	private String[] oldStamps;

	/**
	 * @param a
	 *           Indexing Agent reference
	 * @param task
	 *           Task to be performed
	 */
	public IndexingBehaviour (Agent a, IndexingTask task)
	{
		super(a);
		this.finished = false;
		this.agentIA = (IndexingAgent)a;
		this.stage = StageTaskValidation;
		this.task = task;
		this.objIncl = null;
		this.objRef = null;
		this.objMerged = null;
		this.strErrorMessage = null;
		this.conversationID = task.getConversationID();
		this.oldStamps = null;
	}

	/**
	 * Perform validation of task
	 */
	private void validate ()
	{
		QuadsList quads = this.task.getObjIncl();
		if (quads == null || quads.count() == 0)
		{
			this.agentIA.logg(Level.WARNING,"Empty quads list can't be indexed");
			this.stage = IndexingBehaviour.StageFinal;
			return;
		}
		try
		{
			switch (ObjectValidator.getInstance().validate(quads))
			{
				case ValidationStatus.Complete:
				case ValidationStatus.Incomplete:
					// ok, validation successful
					this.objIncl = quads;
					this.stage = StageSendQueryToDBA;
					break;
				case ValidationStatus.Incorrect:
					this.strErrorMessage = "Indexed object is incorrect: lacking essential properties"
							+ " or some properties marked as singular are in fact multiple";
					this.stage = StageFinal;
					break;
			}
		}
		catch (XMLAnalysisException exc)
		{
			this.osAgent.logg("Internal ObjectValidator error",exc);
			this.finished = true;
			return;
		}
	}

	/**
	 * Performs process of sending query for matching object to DBA
	 */
	private void sendQueryToDBA ()
	{
		AID aidDBA = null;
		OSMessage msg = null;
		try
		{
			msg = new OSMessage();
			msg.setMessageCode(OSMessage.DBA_Find_Matching);
			msg.setObjectInfo(this.objIncl);
			// aidDBA = this.agentIA.getAidDBA();
			aidDBA = Utils.getInstance().tossDBARead(this.osAgent);
			this.agentIA.sendMessage(ACLMessage.REQUEST,msg,aidDBA,
					this.conversationID);
			this.stage = StageWaitForReplyFromDBA;
		}
		catch (FIPAException exc)
		{
			this.osAgent.onError(exc,
					"Can't find agent performing model reading service");
			this.stage = StageFinal;
			return;
		}
		catch (SendingMessageFailed exc)
		{
			this.agentIA.onSendMessgeFailed(aidDBA,OSMessage.DBA_Find_Matching,
					exc,msg);
			this.stage = StageFinal;
			return;
		}
	}

	/**
	 * Wait for reply from DBA
	 */
	private void waitForReplyFromDBA ()
	{
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Matching_Result);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Matching_NotFound);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Find_Matching_Failed);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			switch (msg.getMessageCode())
			{
				case OSMessage.DBA_Find_Matching_NotFound:
					// no matching object found so just send indexed object to the DBA
					this.objMerged = this.objIncl;
					this.stage = StageSendInjectDBA;
					break;
				case OSMessage.DBA_Find_Matching_Result:
					// there is some object to be merged, proceed to the unification phase
					this.objRef = msg.getObjectInfo();
					this.oldStamps = this.objRef.removeTimestamps();
					this.stage = StageUnification;
					break;
				case OSMessage.DBA_Find_Matching_Failed:
					// error :(
					this.strErrorMessage = "Extracting information about existing object failed: "
							+ msg.getAnnouncement();
					this.stage = StageFinal;
					break;
			}
		}
		else
		{
			block();
		}
	}

	/**
	 * Send DBA_Inject_Data message with merged object
	 */
	private void sendInjectDBA ()
	{
		this.objMerged.cleanQuadsWithNull();

		// apply timestamps
		if (this.oldStamps == null)
		{
			// new object
			this.objMerged.stampNewObject();
		}
		else
		{
			// check if update has been performed
			boolean wasUpdate = !this.objRef.isEqual(this.objMerged);
			try
			{
				this.objMerged.updateStamps(this.oldStamps,wasUpdate);
			}
			catch (ParseException exc)
			{
				this.strErrorMessage = "Error during calculating timestamps "
						+ exc.getMessage();
				this.stage = StageFinal;
				return;
			}
		}
 
		OSMessage msg = new OSMessage();
		msg.setMessageCode(OSMessage.DBA_Inject_Data);
		msg.setObjectInfo(this.objMerged);
		AID aidDBA = this.agentIA.getAidDBA();
		try
		{
			this.agentIA.sendMessage(ACLMessage.REQUEST,msg,aidDBA,
					this.conversationID);
			this.stage = StageWaitInjectReply;
		}
		catch (SendingMessageFailed exc)
		{
			this.strErrorMessage = "Sending DBA_Inject_Data request to DBA has failed "
					+ exc.getMessage();
			this.stage = StageFinal;
		}

	}

	/**
	 * Perform unification of objects
	 */
	private void unification ()
	{
		try
		{
			this.objRef.cleanQuadsWithNull();
			this.objIncl.cleanQuadsWithNull(); // added
			MergingReport report = Unificator.getInstance().union(this.objRef,
					this.objIncl);
			// check if merge was successful
			if (!report.isCorrectMerge())
			{
				// ok, send result to DBA
				this.strErrorMessage = "Unification errors: "
						+ report.getErrorsReport();
				this.stage = StageFinal;
			}
			else
			{
				// no, there were errors
				this.objMerged = report.getResult();
				this.stage = StageSendInjectDBA;
			}
		}
		catch (UnionException exc)
		{
			this.strErrorMessage = "Internal error in unification processs: "
					+ exc.getMessage();
			this.stage = StageFinal;
		}
		catch (XMLAnalysisException exc)
		{
			this.strErrorMessage = "Rules set analysis error: " + exc.getMessage();
			this.stage = StageFinal;
		}
	}

	/**
	 * Wait for result of injecting data
	 */
	private void waitInjectReply ()
	{
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DBA_Inject_Data_OK);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Inject_Data_Failed);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			switch (msg.getMessageCode())
			{
				case OSMessage.DBA_Inject_Data_OK:
					// OK, we've successfully finished process of indexing new object so send
					// confirmation to client and end behaviour
					OSMessage reply = new OSMessage();
					reply.setMessageCode(OSMessage.IA_Indexed_OK);
					try
					{
						this.agentIA.sendMessage(ACLMessage.INFORM,reply,this.task
								.getReplyTo(),this.conversationID);
					}
					catch (SendingMessageFailed exc)
					{
						this.strErrorMessage = "Couldn't send IA_Indexed_OK to client "
								+ exc.getMessage();
					}
					this.stage = StageFinal;
					this.osAgent.logg(Level.INFO,"Indexed information about object "
							+ this.objMerged.getObjectURI());
					break;
				case OSMessage.DBA_Inject_Data_Failed:
					// some error during injecting data occured
					this.strErrorMessage = "Error while injecting data";
					this.stage = StageFinal;
					break;
			}
		}
		else
		{
			block();
		}
	}

	/**
	 * Perform final actions before finishing behaviour
	 */
	private void finalizeBehaviour ()
	{
		// if there were some errors then log them
		if (this.strErrorMessage != null)
		{
			this.agentIA.logg(Level.SEVERE,this.strErrorMessage);

			// try to inform client about failure
			OSMessage reply = new OSMessage();
			reply.setMessageCode(OSMessage.IA_Indexed_Failed);
			try
			{
				this.agentIA.sendMessage(ACLMessage.INFORM,reply,this.task
						.getReplyTo(),this.conversationID);
			}
			catch (SendingMessageFailed exc)
			{
				this.agentIA.onSendMessgeFailed(this.task.getReplyTo(),
						OSMessage.IA_Indexed_Failed,exc,reply);
			}
		}

		// mark slot as free and wake scheduler
		this.agentIA.increaseFreeSlots();
		this.agentIA.wakeTasksScheduler();

		// mark behaviour as finished
		this.finished = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		switch (this.stage)
		{
			case StageTaskValidation:
				// validate indexed object, if successful perform sending query to DBA immediately
				this.validate();
			case StageSendQueryToDBA:
				// send request to the DBA for checking if object already exists in the model
				this.sendQueryToDBA();
				break;
			case StageWaitForReplyFromDBA:
				// wait for result form DBA
				this.waitForReplyFromDBA();
				break;
			case StageUnification:
				// if object already exists in the database then perforom unification with the
				// additional provided information
				this.unification();
				break;
			case StageSendInjectDBA:
				// so send result to the DBA to be included in the model
				this.sendInjectDBA();
				break;
			case StageWaitInjectReply:
				this.waitInjectReply();
				break;
			case StageFinal:
				this.finalizeBehaviour();
				break;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#done()
	 */
	public boolean done ()
	{
		return this.finished;
	}
}
