package logrus.agents.managerAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.content.onto.OntologyException;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.wrapper.StaleProxyException;

import java.util.logging.Level;

import logrus.AgentsPlatform;
import logrus.agents.AgentNames;
import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.dataSupervisionAgent.DataSupervisionAgent;
import logrus.agents.databaseAgent.DatabaseAgent;
import logrus.agents.indexingAgent.IndexingAgent;
import logrus.agents.managerAgent.ManagerAgent;
import logrus.agents.searchCoordinatorAgent.SearchCoordinatorAgent;
import logrus.ontmodel.IOntModelWrapper;
import logrus.ontmodel.OntModelWrapperFactory;
import logrus.tools.Utils;
import logrus.tools.logging.InitializeLoggerFailedException;
import logrus.tools.settings.Settings;
import logrus.tools.settings.SettingsManager;

import com.hp.hpl.jena.ontology.OntModelSpec;

/**
 * 
 */
public class BootMainSequenceBehaviour extends OSBehaviour
{
	private String databaseName = null;

	private ManagerAgent agentManager;

	private AID aidDBA;

	private AID aidMA;

	private AID aidIA;

	private AID aidDSA;

	private AID aidSCA;

	private int currentState;

	private static final int START_STATE = 0;

	private static final int BOOT_DBA_STATE = 1;

	private static final int BOOT_DBA_STATE_WAIT = 2;

	private static final int BOOT_DSA_STATE = 3;

	private static final int BOOT_DSA_STATE_WAIT = 4;

	private static final int BOOT_IA_STATE = 5;

	private static final int BOOT_IA_STATE_WAIT = 6;

	private static final int BOOT_SCA_STATE = 7;

	private static final int BOOT_SCA_STATE_WAIT = 8;

	private static final int FINAL_STATE = 9;

	private static final int DONE_STATE = 10;

	private boolean wasError;

	private int numberDBA;

	private int collectedDBA;

	/**
	 * @param a
	 */
	public BootMainSequenceBehaviour (Agent a, String databaseName)
	{
		super(a);
		this.databaseName = databaseName;
		this.agentManager = (ManagerAgent)a;
		this.aidMA = this.agentManager.getAID();
		this.currentState = START_STATE;
		this.wasError = false;
		this.numberDBA = 0;
		this.collectedDBA = 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		switch (this.currentState)
		{
			case START_STATE:
				this.osAgent.logg(Level.INFO,"Starting booting sequence");
				this.currentState = BOOT_DBA_STATE;
				break;
			case BOOT_DBA_STATE:
				this.bootDBA();
				break;
			case BOOT_DBA_STATE_WAIT:
				this.collectConfirmationDBA();
				break;
			case BOOT_IA_STATE:
				this.bootIA();
				break;
			case BOOT_IA_STATE_WAIT:
				this.collectConfirmationIA();
				break;
			case BOOT_DSA_STATE:
				this.bootDSA();
				break;
			case BOOT_DSA_STATE_WAIT:
				this.collectConfirmationDSA();
				break;
			case BOOT_SCA_STATE:
				this.bootSCA();
				break;
			case BOOT_SCA_STATE_WAIT:
				this.collectConfirmationSCA();
				break;
			case FINAL_STATE:
				// all agents booted successfully so send confirmation to GUI agent
				try
				{
					OSMessage confMsg = new OSMessage(OSMessage.MA_Boot_OK);
					this.agentManager.addBehaviour(new ServiceMABehaviour(
							this.agentManager));
					this.osAgent.sendMessage(ACLMessage.INFORM,confMsg,
							this.agentManager.getAidGUIAgent());
				}
				catch (SendingMessageFailed e)
				{
					this.osAgent.logg(Level.SEVERE,
							"Sending boot confirmation to GA failed: " + e);
					this.osAgent.doDelete();
				}
				this.currentState = DONE_STATE;
				break;
		}
	}

	private void onBootFailure (String announcement)
	{
		try
		{
			this.osAgent.sendMessage(ACLMessage.FAILURE,new OSMessage(
					OSMessage.MA_Boot_Failed,announcement),this.agentManager
					.getAidGUIAgent());
			this.osAgent.logg(Level.SEVERE,announcement);
			this.osAgent.doDelete();
			return;
		}
		catch (SendingMessageFailed e)
		{
			this.osAgent.logg(Level.SEVERE,announcement);
			this.osAgent.doDelete();
			return;
		}
	}

	private void bootSCA ()
	{
		// init SCA agent
		SearchCoordinatorAgent agentToBoot;
		try
		{
			agentToBoot = new SearchCoordinatorAgent(this.aidMA,this.aidIA,
					this.aidDBA,this.aidDSA);
			AgentsPlatform.getInstance().addAgent(agentToBoot,
					AgentNames.SearchCoordinatorAgent);
			this.aidSCA = agentToBoot.getAID();
			this.agentManager.setAidSCA(this.aidSCA);
			this.currentState = BOOT_SCA_STATE_WAIT;
		}
		catch (OntologyException exc)
		{
			this.onBootFailure("Creating SCA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
		catch (InitializeLoggerFailedException exc)
		{
			this.onBootFailure("Creating SCA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
		catch (StaleProxyException exc)
		{
			this.onBootFailure("Creating SCA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
	}

	private void collectConfirmationSCA ()
	{
		// now wait for message from DSA
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.SCA_Boot_OK);
		matcher.addAcceptedMsgCode(OSMessage.SCA_Boot_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage osmsg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			if (osmsg.getMessageCode() == OSMessage.SCA_Boot_Failed)
			{
				// DSA boot failed: internal SCA error - send message to GUI agent and terminate
				this.onBootFailure(osmsg.getAnnouncement());
				this.wasError = true;
				this.currentState = FINAL_STATE;
				return;
			}
			else
			{
				// OK
				this.currentState = FINAL_STATE;
			}
		}
		else
		{
			block();
		}

	}

	private void bootDSA ()
	{
		// init DSA agent
		DataSupervisionAgent agentToBoot;
		try
		{
			agentToBoot = new DataSupervisionAgent(this.aidMA,this.aidIA,
					this.aidDBA);
			AgentsPlatform.getInstance().addAgent(agentToBoot,
					AgentNames.DataSupervisionAgent);
			this.aidDSA = agentToBoot.getAID();
			this.agentManager.setAidDSA(this.aidDSA);
			this.currentState = BOOT_DSA_STATE_WAIT;
		}
		catch (OntologyException exc)
		{
			this.onBootFailure("Creating DSA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
			return;
		}
		catch (InitializeLoggerFailedException exc)
		{
			this.onBootFailure("Creating DSA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
			return;
		}
		catch (StaleProxyException exc)
		{
			this.onBootFailure("Creating DSA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
			return;
		}
	}

	private void collectConfirmationDSA ()
	{
		// now wait for message from DSA
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DSA_Boot_OK);
		matcher.addAcceptedMsgCode(OSMessage.DSA_Boot_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage osmsg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			if (osmsg.getMessageCode() == OSMessage.DSA_Boot_Failed)
			{
				// DSA boot failed: internal DSA error - send message to GUI agent and terminate
				this.onBootFailure(osmsg.getAnnouncement());
				this.wasError = true;
				this.currentState = FINAL_STATE;
				return;
			}
			else
			{
				// OK
				this.currentState = BOOT_SCA_STATE;
			}
		}
		else
		{
			block();
		}
	}

	private void bootDBA ()
	{
		// init DBA agent
		DatabaseAgent agentToBoot;
		try
		{
			// load database from files
			// load model from file
			this.osAgent.logg(Level.INFO,"Attempting to load model from "
					+ this.databaseName);
			IOntModelWrapper wrapper = OntModelWrapperFactory.getInstance()
					.createModelFromFile("file:///" + this.databaseName,
							OntModelSpec.RDFS_MEM);
			if (wrapper != null)
			{
				Utils tools = Utils.getInstance();
				Settings settings = SettingsManager.getSettings();

				// add DBA agent running in MODE_WRITE mode
				agentToBoot = new DatabaseAgent(this.databaseName,this.osAgent
						.getAID(),DatabaseAgent.MODE_WRITE,wrapper);
				AgentsPlatform.getInstance().addAgent(agentToBoot,
						tools.getDBAName(DatabaseAgent.MODE_WRITE));
				this.aidDBA = agentToBoot.getAID();
				this.agentManager.setAidDBA(this.aidDBA);

				// add DBA agents running in MODE_READ mode
				for (int i = 0; i < settings.getNumberReadDBA(); i++)
				{
					agentToBoot = new DatabaseAgent(this.databaseName,this.osAgent
							.getAID(),DatabaseAgent.MODE_READ,wrapper);
					AgentsPlatform.getInstance().addAgent(agentToBoot,
							tools.getDBAName(DatabaseAgent.MODE_READ)+"_"+i);
				}

				this.numberDBA = settings.getNumberReadDBA() + 1;

				this.currentState = BOOT_DBA_STATE_WAIT;
			}
			else
			{
				this.onBootFailure("Creating DBA failed - can't load model");
			}
		}
		catch (OntologyException exc)
		{
			this.onBootFailure("Creating DBA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
			return;
		}
		catch (InitializeLoggerFailedException exc)
		{
			this.onBootFailure("Creating DBA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
			return;
		}
		catch (StaleProxyException exc)
		{
			this.onBootFailure("Creating DBA failed " + exc.getMessage());
			return;
		}
	}

	private void collectConfirmationDBA ()
	{
		// now wait for message from DBA
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.DBA_Boot_OK);
		matcher.addAcceptedMsgCode(OSMessage.DBA_Boot_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage osmsg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			if (osmsg.getMessageCode() == OSMessage.DBA_Boot_Failed)
			{
				// DBA boot failed: internal DBA error - send message to GUI agent and terminate
				this.onBootFailure(osmsg.getAnnouncement());
				this.wasError = true;
				this.currentState = FINAL_STATE;
				return;
			}
			else
			{
				// OK
				++this.collectedDBA;
				if (this.collectedDBA == this.numberDBA)
				{
					this.currentState = BOOT_IA_STATE;
				}
				return;
			}
		}
		else
		{
			block();
		}
	}

	private void bootIA ()
	{
		// init IA agent
		IndexingAgent agentToBoot;
		try
		{
			agentToBoot = new IndexingAgent(this.aidMA,this.aidDBA);
			AgentsPlatform.getInstance().addAgent(agentToBoot,
					AgentNames.IndexingAgent);
			this.aidIA = agentToBoot.getAID();
			this.agentManager.setAidIA(this.aidIA);
			this.currentState = BOOT_IA_STATE_WAIT;
		}
		catch (OntologyException exc)
		{
			this.onBootFailure("Creating IA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
		catch (InitializeLoggerFailedException exc)
		{
			this.onBootFailure("Creating IA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
		catch (StaleProxyException exc)
		{
			this.onBootFailure("Creating IA failed " + exc.getMessage());
			this.wasError = true;
			this.currentState = FINAL_STATE;
		}
	}

	private void collectConfirmationIA ()
	{
		// now wait for message from IA
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.IA_Boot_OK);
		matcher.addAcceptedMsgCode(OSMessage.IA_Boot_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage osmsg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			if (osmsg.getMessageCode() == OSMessage.IA_Boot_Failed)
			{
				// IA boot failed: internal IA error - send message to GUI agent and terminate
				this.onBootFailure(osmsg.getAnnouncement());
				this.wasError = true;
				this.currentState = FINAL_STATE;
				return;
			}
			else
			{
				// OK
				this.currentState = BOOT_DSA_STATE;
			}
		}
		else
		{
			block();
		}
	}

	public boolean done ()
	{
		return this.currentState == DONE_STATE;
	}
}
