package logrus.agents.managerAgent.behaviours;

import java.util.logging.Level;

import jade.content.OntoACLMessage;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.leap.ArrayList;
import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.managerAgent.ManagerAgent;
import logrus.agents.searchCoordinatorAgent.SearchTask;

/**
 * 
 */
public class CrawlSupervisionBehaviour extends OSBehaviour
{
	// MA reference
	private ManagerAgent agentMA;

	// flag determining if this behaviour is finished
	private boolean finished;

	// current stage of behaviour
	private int stage;

	// index of stage responsible for sending crawl request to MA
	private static final int StageSendCrawlRequest = 0;

	// index of stage responsible for waiting for reply from MA
	private static final int StageWaitForCrawlReply = 1;

	// index of stage responsible for collecting replies from IA
	private static final int StageCollectIAReplies = 2;

	// index of final stage of this behaviour
	private static final int StageFinal = 10;

	// error message
	private String strErrorMessage;

	// crawl task description
	private SearchTask task;

	// ID for whole conversation process
	private String conversationID;

	// AID of agent operating with GUI client
	private AID aidClientGUI = null;

	// List of object descriptions extracted during crawl process
	private ArrayList listObjects;

	// total number of objects to be sent to IA
	private int nObjects;

	// number of objects sent to IA and IA has responded
	private int nObjectsInjected;

	// number of objects sent
	private int nObjectsInjectedCorrect;

	/**
	 * @param a
	 */
	public CrawlSupervisionBehaviour (Agent a, SearchTask task,
			String conversationID, AID aidClientGUI)
	{
		super(a);
		this.agentMA = (ManagerAgent)a;
		this.stage = StageSendCrawlRequest;
		this.task = task;
		this.conversationID = conversationID;
		this.aidClientGUI = aidClientGUI;
		this.listObjects = null;
		this.nObjects = 0;
		this.nObjectsInjected = 0;
		this.nObjectsInjectedCorrect = 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		switch (this.stage)
		{
			case StageSendCrawlRequest:
				this.sendCrawlRequest();
				break;
			case StageWaitForCrawlReply:
				this.waitForCrawlReply();
				break;
			case StageCollectIAReplies:
				this.collectIAReplies();
				break;
		}
	}

	private void collectIAReplies ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.IA_Mass_Index_Failed);
		matcher.addAcceptedMsgCode(OSMessage.IA_Indexed_Failed);
		matcher.addAcceptedMsgCode(OSMessage.IA_Indexed_OK);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);
		
		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			// extract information form message
			AID aidClient = jmsg.getSender();
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			// check which operation is requested and perform it
			switch (msg.getMessageCode())
			{
				case OSMessage.IA_Indexed_OK:
					// indexing single object ok
					++this.nObjectsInjected;
					++this.nObjectsInjectedCorrect;
					break;
				case OSMessage.IA_Indexed_Failed:
					++this.nObjectsInjected;
					// indexing single object has failed
					break;
				case OSMessage.IA_Mass_Index_Failed:
					// big failure
					this.strErrorMessage = "Indexing crawl results failed";
					this.stage = StageFinal;
					return;
			}
			if (this.nObjects == this.nObjectsInjected)
			{
				// send message to GUI client
				OSMessage reply = new OSMessage();
				reply.setMessageCode(OSMessage.MA_Crawl_Request_OK);
				reply.setAnnouncement("Crawled information about " + this.nObjects
						+ " restaurants, " + this.nObjectsInjectedCorrect
						+ " have been successfuly added to the model");
				try
				{
					this.agentMA.sendMessage(ACLMessage.INFORM,reply,
							this.aidClientGUI,this.conversationID);
				}
				catch (SendingMessageFailed exc)
				{
					this.strErrorMessage = "Sending message MA_Crawl_Request_OK to GUI failed";
					this.agentMA.onSendMessgeFailed(aidClient,
							OSMessage.MA_Crawl_Request_OK,exc,reply);
				}
				finally
				{
					this.stage = StageFinal;
				}
			}
		}
		else
		{
			// block behaviour until next message arrives
			block();
		}
	}

	/**
	 * Wait for reply from SCA with crawl results
	 */
	private void waitForCrawlReply ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.SCA_Crawl_Result);
		matcher.addAcceptedMsgCode(OSMessage.SCA_Crawl_Failed);
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			// extract information form message
			AID aidClient = jmsg.getSender();
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			// check which operation is requested and perform it
			switch (msg.getMessageCode())
			{
				case OSMessage.SCA_Crawl_Result:
					this.onSCA_Crawl_Result(msg);
					break;
				case OSMessage.SCA_Crawl_Failed:
					// send error message to GUI
					this.onErrorInProcess(msg.getAnnouncement() != null ? msg
							.getAnnouncement() : "");
					break;
			}
		}
		else
		{
			// block behaviour
			block();
		}
	}

	private void onSCA_Crawl_Result (OSMessage msg)
	{
		this.listObjects = msg.getObjectsCollection();
		this.nObjects = this.listObjects.size();

		OSMessage request = new OSMessage();
		request.setMessageCode(OSMessage.IA_Mass_Index);
		request.setObjectsCollection(this.listObjects);
		try
		{
			this.agentMA.sendMessage(ACLMessage.REQUEST,request,this.agentMA
					.getAidIA(),this.conversationID);
			this.stage = StageCollectIAReplies;
		}
		catch (SendingMessageFailed exc)
		{
			this.strErrorMessage = "Sending message to IA failed: "
					+ exc.getMessage();
			this.stage = StageFinal;
		}
	}

	/**
	 * Some error in process, inform GUI about this
	 * 
	 * @param message
	 *           Message to GUI client
	 */
	private void onErrorInProcess (String message)
	{
		OSMessage msg = new OSMessage();
		msg.setMessageCode(OSMessage.MA_Crawl_Request_Failed);
		this.strErrorMessage = "MA reports error in webparsing stage: " + message;
		msg.setAnnouncement(this.strErrorMessage);
		try
		{
			this.agentMA.sendMessage(ACLMessage.REQUEST,msg,this.aidClientGUI,
					this.conversationID);
		}
		catch (SendingMessageFailed exc)
		{
			this.strErrorMessage = "MA Couldn't send MA_Crawl_Request_Failed message to GUIA: "
					+ exc.getMessage();
			this.agentMA.onSendMessgeFailed(this.aidClientGUI,
					OSMessage.MA_Crawl_Request_Failed,exc,msg);
		}
		finally
		{
			this.stage = StageFinal;
		}
	}

	/**
	 * Send SCA_Crawl message to SCA
	 */
	private void sendCrawlRequest ()
	{
		OSMessage msg = new OSMessage();
		msg.setMessageCode(OSMessage.SCA_Crawl);
		msg.setSearchCommand(this.task);
		try
		{
			this.agentMA.sendMessage(ACLMessage.REQUEST,msg,this.agentMA
					.getAidSCA(),this.conversationID);
			this.stage = StageWaitForCrawlReply;
		}
		catch (SendingMessageFailed exc)
		{
			this.strErrorMessage = "MA Couldn't send SCA_Crawl message to SCA: "
					+ exc.getMessage();
			this.agentMA.onSendMessgeFailed(this.agentMA.getAidSCA(),
					OSMessage.SCA_Crawl,exc,msg);
			this.stage = StageFinal;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#done()
	 */
	public boolean done ()
	{
		return this.finished;
	}

	/**
	 * Perform final actions before finishing behaviour
	 */
	private void finalizeBehaviour ()
	{
		// if there were some errors then log them
		if (this.strErrorMessage != null)
		{
			this.agentMA.logg(Level.SEVERE,this.strErrorMessage);

			// try to send failure information to GUI
			OSMessage reply = new OSMessage();
			reply.setMessageCode(OSMessage.MA_Crawl_Request_Failed);
			reply.setAnnouncement(this.strErrorMessage);
			try
			{
				this.agentMA.sendMessage(ACLMessage.INFORM,reply,this.aidClientGUI,
						this.conversationID);
			}
			catch (SendingMessageFailed exc)
			{
				this.agentMA.onSendMessgeFailed(this.aidClientGUI,
						OSMessage.MA_Crawl_Request_Failed,exc,reply);
			}
		}

		// mark behaviour as finished
		this.finished = true;
	}
}
