package logrus.agents.ontology;

import jade.content.onto.BasicOntology;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.schema.AgentActionSchema;
import jade.content.schema.ConceptSchema;
import jade.content.schema.ObjectSchema;
import jade.content.schema.PrimitiveSchema;
import logrus.agents.OSMessage;
import logrus.agents.searchCoordinatorAgent.SearchTask;
import logrus.ontmodel.Quad;
import logrus.ontmodel.QuadsList;

/**
 * Represents ontology used by OntoStorage system for exchanging information using ACL messages.
 * This class is implemented using singleton design pattern.
 */
public class OntoStorageOntology extends Ontology implements
		OntoStorageVocabulary
{
	// singleton instance
	private static OntoStorageOntology instance = null;

	// ontology name
	public static final String ONTOLOGY_NAME = "OntoStorage-ontology";

	/**
	 * @return Instance of OntoStorageOntology class
	 */
	public static OntoStorageOntology getInstance () throws OntologyException
	{
		if (OntoStorageOntology.instance == null)
		{
			OntoStorageOntology.instance = new OntoStorageOntology();
		}
		return OntoStorageOntology.instance;
	}

	/**
	 * Constructor kept private (following singleton design pattern)
	 */
	private OntoStorageOntology () throws OntologyException
	{
		super(ONTOLOGY_NAME,BasicOntology.getInstance());

		// add concepts
		this.addQuadConcept();
		this.addQuadsConcept();
		this.addSearchTaskConcept();
		this.addOSMessageConcept();
	}

	private void addSearchTaskConcept () throws OntologyException
	{
		ConceptSchema cs = new ConceptSchema(SEARCH_TASK);
		cs.add(SEARCH_TASK_TYPE,
				(PrimitiveSchema)getSchema(BasicOntology.INTEGER),
				ObjectSchema.MANDATORY);
		cs.add(SEARCH_TASK_URL,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.MANDATORY);		
		cs.add(SEARCH_TASK_DATA_SOURCE,
				(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.MANDATORY);
		cs.add(SEARCH_TASK_COMMAND_OPTIONS,
				(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		cs.add(SEARCH_TASK_CONVERSATION_ID,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);		
		this.add(cs,SearchTask.class);
	}

	private void addQuadsConcept () throws OntologyException
	{
		ConceptSchema cs = new ConceptSchema(QUADS);
		cs.add(QUADS_COLLECTION,(ConceptSchema)getSchema(QUAD),0,
				ObjectSchema.UNLIMITED);
		this.add(cs,QuadsList.class);
	}

	private void addQuadConcept () throws OntologyException
	{
		ConceptSchema cs = new ConceptSchema(QUAD);
		cs.add(QUAD_SUBJECT,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		cs.add(QUAD_PREDICATE,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		cs.add(QUAD_OBJECT,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		cs.add(QUAD_PROVIDER,(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		this.add(cs,Quad.class);
	}

	private void addOSMessageConcept () throws OntologyException
	{
		AgentActionSchema as = new AgentActionSchema(OS_MESSAGE);
		as.add(OS_MESSAGE_CODE,(PrimitiveSchema)getSchema(BasicOntology.INTEGER),
				ObjectSchema.MANDATORY);
		as.add(OS_MESSAGE_ANNOUNCEMENT,
				(PrimitiveSchema)getSchema(BasicOntology.STRING),
				ObjectSchema.OPTIONAL);
		as.add(OS_MESSAGE_QUAD_STATEMENT,(ConceptSchema)getSchema(QUAD),
				ObjectSchema.OPTIONAL);
		as.add(OS_MESSAGE_MULTIPLE_QUADS,(ConceptSchema)getSchema(QUADS),
				ObjectSchema.OPTIONAL);
		as.add(OS_MESSAGE_SEARCH_TASK,(ConceptSchema)getSchema(SEARCH_TASK),
				ObjectSchema.OPTIONAL);
		as.add(OS_MESSAGE_URI_COLLECTION,
				(PrimitiveSchema)getSchema(BasicOntology.STRING),0,
				ObjectSchema.UNLIMITED);
		as.add(OS_MESSAGE_OBJECTS_COLLECTION,
				(ConceptSchema)getSchema(QUADS),0,
				ObjectSchema.UNLIMITED);		
		this.add(as,OSMessage.class);
	}
}
