package logrus.agents.searchCoordinatorAgent;

import java.util.logging.Level;

import jade.content.onto.OntologyException;
import jade.core.AID;
import jade.domain.FIPAException;
import logrus.agents.OntoStorageAgent;
import logrus.agents.ServiceNames;
import logrus.agents.ontology.OntoStorageOntology;
import logrus.agents.searchCoordinatorAgent.behaviours.BootSearchCoordinatorAgentBehaviour;
import logrus.agents.searchCoordinatorAgent.behaviours.TasksSchedulerBehaviour;
import logrus.tools.logging.InitializeLoggerFailedException;
import logrus.tools.queue.Queue;
import logrus.tools.settings.SettingsManager;

/**
 * 
 */
public class SearchCoordinatorAgent extends OntoStorageAgent
{
	private AID aidMA;

	private AID aidIA;

	private AID aidDBA;

	private AID aidDSA;

	// Queue of URLs of webpages to be parsed
	private Queue tasksQueue;

	// maximal number of simultaneously running search agents
	private int maxAgents;

	// queue containing IDs of currently free wrapper agents
	private Queue freeAgents;

	// behaviour responsible for scheduling tasks
	private TasksSchedulerBehaviour tasksScheduler;

	/**
	 * @throws InitializeLoggerFailedException
	 * @throws OntologyException
	 */
	public SearchCoordinatorAgent (AID aidMA, AID aidIA, AID aidDBA, AID aidDSA)
			throws InitializeLoggerFailedException, OntologyException
	{
		super();
		this.aidMA = aidMA;
		this.aidIA = aidIA;
		this.aidDBA = aidDBA;
		this.aidDSA = aidDSA;
		this.tasksQueue = new Queue();
		this.freeAgents = new Queue();
		this.maxAgents = SettingsManager.getSettings().getNumberWA();
	}

	/**
	 * Wakes scheduler if it has been previously suspeneded
	 */
	public void wakeScheduler ()
	{
		this.tasksScheduler.restart();
	}

	/**
	 * Add search task to the end of queue
	 * 
	 * @param task
	 *           Search task to be performed some time in future
	 */
	public void enqueueSearchTask (SearchTask task)
	{
		this.tasksQueue.inject(task);
	}

	/**
	 * Get oldest task from queue of tasks to be performed
	 * 
	 * @return oldest task from queue of tasks to be performed
	 */
	public SearchTask dequeueSearchTask ()
	{
		Object obj = this.tasksQueue.pop();
		if (obj == null)
		{
			return null;
		}
		else
		{
			return (SearchTask)obj;
		}
	}

	/**
	 * Check if there are some search tasks that should be launched
	 * 
	 * @return Boolean value indicating if there are some search tasks that should be launched
	 */
	public boolean isSearchToPerform ()
	{
		return this.tasksQueue.count() > 0;
	}

	/**
	 * Check if there are some free agents that can perform search task
	 * 
	 * @return Boolean value indicating if there are some free agents that can perform search task
	 */
	public boolean isFreeAgent ()
	{
		return this.freeAgents.count() > 0;
	}

	/**
	 * Try to register free WA agent with AID equal <code>aidWA</code>
	 * 
	 * @param aidWA
	 *           AID of free WA
	 * @return Boolean value indicating if specified agent has been really registerd this time.
	 *         Returning true means that agent hasn't been present in the queue, false means that it
	 *         already has been considered as a free agent.
	 */
	public boolean registerFreeAgent (AID aidWA)
	{
		int oldSize = this.freeAgents.count();
		this.freeAgents.injectUnique(aidWA);
		return oldSize < this.freeAgents.count();
	}

	public void setup ()
	{
		this.addBehaviour(new BootSearchCoordinatorAgentBehaviour(this));
		try
		{
			this.registerService(ServiceNames.ServiceSCA,
					OntoStorageOntology.ONTOLOGY_NAME);
		}
		catch (FIPAException exc)
		{
			this.logg("SCA couldn't register service",exc);
		}
	}

	public void takeDown ()
	{
		try
		{
			this.logg(Level.INFO,"SCA going offline");
			this.deregisterService(ServiceNames.ServiceSCA,
					OntoStorageOntology.ONTOLOGY_NAME);
		}
		catch (FIPAException exc)
		{
			this.logg("SCA couldn't deregister service",exc);
		}
	}

	/**
	 * Get number of free agents (agents that are not performing any search and wait for commands)
	 * 
	 * @return Number of currently free wrapper agents
	 */
	public int getNumberOfFreeAgents ()
	{
		return this.freeAgents.count();
	}

	/**
	 * Get number of tasks waiting to be performed
	 * 
	 * @return Number of tasks waiting to be performed
	 */
	public int getNumberOfTasks ()
	{
		return this.tasksQueue.count();
	}

	/**
	 * Get first free agent form list
	 * 
	 * @return AID of first free agent or null if there are no free agents
	 */
	public AID getFirstFreeAgent ()
	{
		if (this.isFreeAgent())
		{
			return (AID)this.freeAgents.pop();
		}
		return null;
	}

	/**
	 * Mark WA agent as free = add it's AID to the list of free agents
	 * 
	 * @param agentWA
	 *           <code>AID</code> of WA agent that should be marked as free
	 */
	public void markAgentAsFree (AID agentWA)
	{
		this.freeAgents.injectUnique(agentWA);
	}

	/**
	 * @param aidMA
	 *           The aidMA to set.
	 */
	public void setAidMA (AID aidMA)
	{
		this.aidMA = aidMA;
	}

	/**
	 * @return Returns the aidMA.
	 */
	public AID getAidMA ()
	{
		return this.aidMA;
	}

	/**
	 * @param aidIA
	 *           The aidIA to set.
	 */
	public void setAidIA (AID aidIA)
	{
		this.aidIA = aidIA;
	}

	/**
	 * @return Returns the aidIA.
	 */
	public AID getAidIA ()
	{
		return this.aidIA;
	}

	/**
	 * @param aidDBA
	 *           The aidDBA to set.
	 */
	public void setAidDBA (AID aidDBA)
	{
		this.aidDBA = aidDBA;
	}

	/**
	 * @return Returns the aidDBA.
	 */
	public AID getAidDBA ()
	{
		return this.aidDBA;
	}

	/**
	 * @param aidDSA
	 *           The aidDSA to set.
	 */
	public void setAidDSA (AID aidDSA)
	{
		this.aidDSA = aidDSA;
	}

	/**
	 * @return Returns the aidDSA.
	 */
	public AID getAidDSA ()
	{
		return this.aidDSA;
	}

	/**
	 * @param tasksQueue
	 *           The tasksQueue to set.
	 */
	public void setTasksQueue (Queue toParseQueue)
	{
		this.tasksQueue = toParseQueue;
	}

	/**
	 * @return Returns the tasksQueue.
	 */
	public Queue getTasksQueue ()
	{
		return this.tasksQueue;
	}

	/**
	 * @param maxAgents
	 *           The maxAgents to set.
	 */
	public void setMaxAgents (int maxAgents)
	{
		this.maxAgents = maxAgents;
	}

	/**
	 * @return Returns the maxAgents.
	 */
	public int getMaxAgents ()
	{
		return this.maxAgents;
	}

	/**
	 * @param freeAgents
	 *           The freeAgents to set.
	 */
	public void setFreeAgents (Queue freeAgents)
	{
		this.freeAgents = freeAgents;
	}

	/**
	 * @return Returns the freeAgents.
	 */
	public Queue getFreeAgents ()
	{
		return this.freeAgents;
	}

	/**
	 * @param tasksScheduler
	 *           The tasksScheduler to set.
	 */
	public void setTasksScheduler (TasksSchedulerBehaviour tasksScheduler)
	{
		this.tasksScheduler = tasksScheduler;
	}

	/**
	 * @return Returns the tasksScheduler.
	 */
	public TasksSchedulerBehaviour getTasksScheduler ()
	{
		return this.tasksScheduler;
	}
}
