package logrus.agents.searchCoordinatorAgent.behaviours;

import jade.content.OntoACLMessage;
import jade.content.onto.OntologyException;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.wrapper.StaleProxyException;

import java.util.logging.Level;

import logrus.AgentsPlatform;
import logrus.agents.AgentNames;
import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.datasources.wrapperAgent.WrapperAgent;
import logrus.agents.searchCoordinatorAgent.SearchCoordinatorAgent;
import logrus.tools.logging.InitializeLoggerFailedException;

/**
 * 
 */
public class BootSearchCoordinatorAgentBehaviour extends OSBehaviour
{
	private SearchCoordinatorAgent agentSCA;

	// AID of SCA
	private AID aidSCA;

	// current behaviour stage number
	private int currentStage;

	// number of last stage (behaviour should be terminated when currentStage reaches this value)
	private static final int LAST_STAGE = 3;

	/**
	 * @param a
	 */
	public BootSearchCoordinatorAgentBehaviour (Agent a)
	{
		super(a);
		this.agentSCA = (SearchCoordinatorAgent)a;
		this.aidSCA = this.agentSCA.getAID();
		this.currentStage = 0;
	}

	private void bootWA ()
	{
		// try to boot maxAgents
		int maxAgents = this.agentSCA.getMaxAgents();
		for (int i = 0; i < maxAgents; i++)
		{
			WrapperAgent agentToBoot;
			try
			{
				// run WA
				agentToBoot = new WrapperAgent(this.aidSCA);
				AgentsPlatform.getInstance().addAgent(agentToBoot,
						AgentNames.WrapperAgent + "_" + i);
			}
			catch (OntologyException e)
			{
				// failed to boot: take down
				this.onWABootFailure("Failed to boot WA number " + i + " because: "
						+ e.toString());
				return;
			}
			catch (InitializeLoggerFailedException e)
			{
				// failed to boot: take down
				this.onWABootFailure("Failed to boot WA number " + i + " because: "
						+ e.toString());
				return;
			}
			catch (StaleProxyException e)
			{
				// failed to boot: take down
				this.onWABootFailure("Failed to boot WA number " + i + " because: "
						+ e.toString());
				return;
			}
		}
		// ok, all agents booted successfully so proceed to the next stage
		++this.currentStage;
	}

	private void waitWABootConfirmations ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		matcher.addAcceptedMsgCode(OSMessage.WA_Boot_OK);
		matcher.addAcceptedMsgCode(OSMessage.WA_Boot_Failed);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// recive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);

		if (jmsg != null)
		{
			// extract information form message
			AID aidClient = jmsg.getSender();
			String conversationID = jmsg.getConversationId();
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());

			// check which operation is requested and perform it
			switch (msg.getMessageCode())
			{
				case OSMessage.WA_Boot_OK:
					// boot OK, add WA AID to free agents list
					this.agentSCA.registerFreeAgent(aidClient);
					// and check if all WA has reported to be ready
					if (this.agentSCA.getNumberOfFreeAgents() == this.agentSCA
							.getMaxAgents())
					{
						// if so, then proceed to the next stage
						++this.currentStage;
					}
					break;

				case OSMessage.WA_Boot_Failed:
					// inform MA about failure and take down
					this.onWABootFailure("WA" + aidClient
							+ " reported unsuccessful boot");
					break;
				default:
					this.osAgent.logg(Level.WARNING,
							"SCA received unknown/unexpected message "
									+ msg.getMessageCode() + " from " + aidClient);
					return;
			}
		}
		else
		{
			// no message in queue that should be processed so wait for next one
			block();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		switch (this.currentStage)
		{
			case 0:
				// first stage : try to boot wrapper agents
				this.bootWA();
				break;
			case 1:
				// second stage: wait for boot confirmations form all WA
				this.waitWABootConfirmations();
				break;
			case 2:
				// third stage: run messages listener and tasks scheduler behaviours,
				// all WA agents booted successfully so inform MA that SCA is ready
				this.finalizeBoot();
				break;
		}
	}

	private void finalizeBoot ()
	{
		OSMessage confMsg = new OSMessage(OSMessage.SCA_Boot_OK);

		// first start messages listener and tasks scheduler behaviours
		TasksSchedulerBehaviour scheduler = new TasksSchedulerBehaviour(
				this.agentSCA);
		this.agentSCA.setTasksScheduler(scheduler);
		this.agentSCA.addBehaviour(scheduler);
		
		ServiceSCABehaviour listener = new ServiceSCABehaviour(this.agentSCA);
		this.agentSCA.addBehaviour(listener);
		
		// and try to send successful boot confirmation to MA
		try
		{
			this.osAgent.sendMessage(ACLMessage.INFORM,confMsg,this.agentSCA
					.getAidMA());
			this.osAgent.addBehaviour(new ServiceSCABehaviour(this.osAgent));
			++this.currentStage;
			this.osAgent.logg(Level.INFO,"Booted SCA");
		}
		catch (SendingMessageFailed e)
		{
			this.osAgent.logg(Level.SEVERE,
					"Sending boot confirmation to MA failed: " + e);
			this.osAgent.doDelete();
		}
	}

	private void onWABootFailure (String announcement)
	{
		try
		{
			this.osAgent
					.sendMessage(ACLMessage.FAILURE,new OSMessage(
							OSMessage.SCA_Boot_Failed,announcement),this.agentSCA
							.getAidMA());
			this.osAgent.logg(Level.SEVERE,announcement);
			this.osAgent.doDelete();
			return;
		}
		catch (SendingMessageFailed e)
		{
			this.osAgent.logg(Level.SEVERE,announcement);
			this.osAgent.doDelete();
			return;
		}
	}

	public boolean done ()
	{
		return this.currentStage == BootSearchCoordinatorAgentBehaviour.LAST_STAGE;
	}

}
