package logrus.agents.searchCoordinatorAgent.behaviours;

import java.util.logging.Level;

import jade.content.OntoACLMessage;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;
import logrus.agents.OSMessage;
import logrus.agents.OSMessageMatcher;
import logrus.agents.base.behaviours.OSBehaviour;
import logrus.agents.base.exceptions.SendingMessageFailed;
import logrus.agents.searchCoordinatorAgent.SearchCoordinatorAgent;
import logrus.agents.searchCoordinatorAgent.SearchTask;
import logrus.ontmodel.QuadsList;

/**
 * 
 */
public class CrawlSupervisionBehaviour extends OSBehaviour
{
	// behaviour should listen for messages wuth this conversation ID
	private String conversationID;

	// AID of agent that commisoned supervised search
	private AID aidClient;

	// SCA agent reference
	private SearchCoordinatorAgent agentSCA;

	// flag determining if behaviour should be considered as finished
	private boolean endFlag;

	// current stage of behaviour
	private int currentStage;

	// number of links to be parsed (initialized after receiving first message from WA)
	private int linksToReceive;

	// number of restaurant links already parsed (whatever successfully or not)
	private int linksParsed;

	//
	private String providerID;

	// list of correct results (QuadLists containing information about restaurants)
	private ArrayList correctResults;

	/**
	 * @param a
	 */
	public CrawlSupervisionBehaviour (Agent a, AID aidClient,
			String conversationID, String providerID)
	{
		super(a);
		this.conversationID = conversationID;
		this.aidClient = aidClient;
		this.agentSCA = (SearchCoordinatorAgent)a;
		this.endFlag = false;
		this.currentStage = 0;
		this.linksToReceive = 0;
		this.providerID = providerID;
		this.linksParsed = 0;
		this.correctResults = new ArrayList();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action ()
	{
		// prepare accepted message templates
		OSMessageMatcher matcher = new OSMessageMatcher(this.osAgent
				.getContentManager());
		if (this.currentStage == 0)
		{
			// in first stage we're waiting only for crawl result
			matcher.addAcceptedMsgCode(OSMessage.WA_Crawl_Result);
			matcher.addAcceptedMsgCode(OSMessage.WA_Crawl_Failed);
		}
		else
		{
			// in second stage we're waiting only for individual pages results
			matcher.addAcceptedMsgCode(OSMessage.WA_Extract_Object_Info_Result);
			matcher.addAcceptedMsgCode(OSMessage.WA_Extract_Object_Info_Failed);
		}
		matcher.setAcceptedConversationID(this.conversationID);
		MessageTemplate pattern = new MessageTemplate(matcher);

		// receive message
		OntoACLMessage jmsg = (OntoACLMessage)this.osAgent.receive(pattern);
		if (jmsg != null)
		{
			OSMessage msg = OSMessageMatcher.extractMessage(jmsg,this.osAgent
					.getContentManager());
			switch (this.currentStage)
			{
				case 0:
					this.onCrawlResult(msg);
					break;
				case 1:
					this.onExtractResult(msg);
					break;
			}

			// whatever has happend register WA as a free one
			this.agentSCA.registerFreeAgent(jmsg.getSender());

			// wake scheduler
			this.agentSCA.wakeScheduler();
		}
		else
		{
			// no message from appropriate WA so switch to inactive mode
			this.block(1000);
		}
	}

	private void onExtractResult (OSMessage msg)
	{
		switch (msg.getMessageCode())
		{
			case OSMessage.WA_Extract_Object_Info_Result:
				// add result to list of correctly parsed restaurants
				QuadsList quads = msg.getObjectInfo();
				if (quads != null)
				{
					quads.cleanQuadsWithNull();
					this.correctResults.add(quads);
				}
				++this.linksParsed;
				break;
			case OSMessage.WA_Extract_Object_Info_Failed:
				++this.linksParsed;
				break;
		}
		double percentage = ((double)this.linksParsed / (double)this.linksToReceive) * 100.0;
		this.agentSCA.logg(Level.INFO,"Crawl " + percentage
				+ "% finished (parsed " + this.linksParsed + "/"
				+ this.linksToReceive + " including " + this.correctResults.size()
				+ " correctly parsed");
		// now chceck if we've received replies from all WA
		if (this.linksParsed == this.linksToReceive)
		{
			// send results back to the client
			OSMessage replyMsg = new OSMessage();
			replyMsg.setMessageCode(OSMessage.SCA_Crawl_Result);
			replyMsg.setObjectsCollection(this.correctResults);
			try
			{
				this.agentSCA.sendMessage(ACLMessage.INFORM,replyMsg,
						this.aidClient,this.conversationID);
			}
			catch (SendingMessageFailed e)
			{
				this.agentSCA.onSendMessgeFailed(this.aidClient,replyMsg
						.getMessageCode(),e,replyMsg);
			}
			// end finish behaviour
			this.endFlag = true;
		}
	}

	private void onCrawlResult (OSMessage msg)
	{
		switch (msg.getMessageCode())
		{
			case OSMessage.WA_Crawl_Result:
				ArrayList links = msg.getUris();

				// remember number of links that should be parsed
				this.linksToReceive = links.size();

				// process received links
				for (Iterator iter = links.iterator(); iter.hasNext();)
				{
					// prepare search tasks for individual pages and place them in global queue
					String restLink = (String)iter.next();
					SearchTask task = new SearchTask(SearchTask.EXTRACT_INFO,
							this.providerID,restLink);
					task.setConversationID(this.conversationID);
					this.agentSCA.enqueueSearchTask(task);
				}
				++this.currentStage;

				// and wake
				this.agentSCA.wakeScheduler();
				break;
			case OSMessage.WA_Crawl_Failed:
				// crawl failed so inform client that SCA couldn't perform this task
				OSMessage replyFailure = new OSMessage();
				replyFailure.setMessageCode(OSMessage.SCA_Crawl_Failed);
				try
				{
					this.agentSCA.sendMessage(ACLMessage.INFORM,replyFailure,
							this.aidClient,this.conversationID);
				}
				catch (SendingMessageFailed exc)
				{
					this.agentSCA.onSendMessgeFailed(this.aidClient,
							OSMessage.SCA_Crawl_Failed,exc,replyFailure);
				}
				this.endFlag = true;
				break;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see jade.core.behaviours.Behaviour#done()
	 */
	public boolean done ()
	{
		return this.endFlag;
	}

}
