package logrus.ontmodel;

import com.hp.hpl.jena.db.DBConnection;
import com.hp.hpl.jena.db.IDBConnection;

import java.sql.SQLException;

import logrus.tools.settings.Settings;
import logrus.tools.settings.SettingsManager;

/**
 * Class wrapping database-related operations
 */
public class DatabaseWrapper
{
	// path of JDBC driver class
	private String driverJDBC = null;

	// URL of the database
	private String dbURL = null;

	// Type of the database (MySQL, Oracle, etc.)
	private String dbType = null;

	// Database user
	private String dbUser = null;

	// Database user password
	private String dbPassword = null;

	// Connection to the database
	private DBConnection dbConnection = null;

	/**
	 * Default constructor initializes database connection for logrus's computer
	 */
	public DatabaseWrapper ()
	{
		super();
		Settings settings = SettingsManager.getSettings();

		this.setDbPassword(settings.getDbPassword());
		this.setDbUser(settings.getDbUser());

		if (settings.getDbProvider().compareTo("MySQL") == 0)
		{
			// this.setDbURL("jdbc:mysql://puter/OntoStorage");
			this.setDbURL("jdbc:mysql://" + settings.getDbServer()
					+ "/OntoStorage");
			this.setDbType("MySQL");
			this.setDriverJDBC("com.mysql.jdbc.Driver");
		}
	}

	/**
	 * Loads JDBC Driver
	 * 
	 * @throws ClassNotFoundException
	 *            Can't find the driver class
	 */
	private void loadDriver () throws ClassNotFoundException
	{
		Class.forName(this.driverJDBC);
	}

	/**
	 * Disconnect from the database
	 * 
	 * @throws SQLException
	 */
	public void disconnect () throws SQLException
	{
		if (this.getDbConnection() != null)
		{
			this.getDbConnection().close();
		}
	}

	/**
	 * Connect to the database
	 * 
	 * @return Connection to the database
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public IDBConnection connect () throws SQLException, ClassNotFoundException
	{
		this.loadDriver();
		// disconnect if we've been previously connected
		if (this.getDbConnection() != null
				&& this.getDbConnection().getConnection() != null)
		{
			this.disconnect();
		}
		// connect to the database
		this.setDbConnection(new DBConnection(this.dbURL,this.dbUser,
				this.dbPassword,this.dbType));
		return this.getDbConnection();
	}

	/**
	 * Check if we're connected to the database and this connection is active
	 * 
	 * @return Boolean value indicating if connection is active
	 * @throws SQLException
	 */
	public boolean isConnected () throws SQLException
	{
		if (this.getDbConnection() != null
				&& this.getDbConnection().getConnection() != null)
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * @param dbPassword
	 *           The dbPassword to set.
	 */
	public void setDbPassword (String databasePassword)
	{
		this.dbPassword = databasePassword;
	}

	/**
	 * @return Returns the dbPassword.
	 */
	public String getDbPassword ()
	{
		return this.dbPassword;
	}

	/**
	 * @param dbType
	 *           The dbType to set.
	 */
	public void setDbType (String databaseType)
	{
		this.dbType = databaseType;
	}

	/**
	 * @return Returns the dbType.
	 */
	public String getDbType ()
	{
		return this.dbType;
	}

	/**
	 * @param dbURL
	 *           The dbURL to set.
	 */
	public void setDbURL (String databaseURL)
	{
		this.dbURL = databaseURL;
	}

	/**
	 * @return Returns the dbURL.
	 */
	public String getDbURL ()
	{
		return this.dbURL;
	}

	/**
	 * @param dbUser
	 *           The dbUser to set.
	 */
	public void setDbUser (String databaseUser)
	{
		this.dbUser = databaseUser;
	}

	/**
	 * @return Returns the dbUser.
	 */
	public String getDbUser ()
	{
		return this.dbUser;
	}

	/**
	 * @param driverJDBC
	 *           The driverJDBC to set.
	 */
	public void setDriverJDBC (String driverJDBC)
	{
		this.driverJDBC = driverJDBC;
	}

	/**
	 * @return Returns the driverJDBC.
	 */
	public String getDriverJDBC ()
	{
		return this.driverJDBC;
	}

	/**
	 * @param dbConnection
	 *           The dbConnection to set.
	 */
	private void setDbConnection (DBConnection dbConnection)
	{
		this.dbConnection = dbConnection;
	}

	/**
	 * @return Returns the dbConnection.
	 */
	public DBConnection getDbConnection ()
	{
		return this.dbConnection;
	}

}
