package logrus.ontmodel;

import jade.util.leap.ArrayList;

import java.io.IOException;
import java.text.ParseException;

import logrus.ontmodel.exceptions.MatchingException;
import logrus.ontmodel.exceptions.OntPropertyNotFoundException;
import logrus.ontmodel.exceptions.ResourceNotFoundException;
import logrus.tools.queue.Queue;

import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdql.QueryResults;
import com.sun.corba.se.internal.io.TypeMismatchException;

/**
 * Just a wrapper for JENA persistent ontological model
 */
public interface IOntModelWrapper
{

	/**
	 * Extract object information matching information included in <code>objInfo</code>
	 * 
	 * @param objInfo
	 *           Object information used as pattern
	 * @return Object information matching information included in <code>objInfo</code> if matching
	 *         object exists in model, otherwise returns false
	 */
	public QuadsList findMatchingObject (QuadsList objInfo) throws MatchingException;

	/**
	 * @param subject
	 * @param predicate
	 * @param object
	 * @param providerID
	 */
	public void addDescriptiveStatement (String providerID, String subject,
			String predicate, String predValue)
			throws OntPropertyNotFoundException;

	/**
	 * Insert information about ontological object described by statements included in
	 * <code>info</code>
	 * 
	 * @param info
	 *           Collection of statements describing ontological object
	 * @throws TypeMismatchException
	 * @throws OntPropertyNotFoundException
	 */
	public void addObjectInfo (QuadsList info)
			throws OntPropertyNotFoundException;

	/**
	 * Add resource with URI <code>uriRes</code> if it does not exist in model
	 * 
	 * @param uriRes
	 *           URI of new resource
	 * @return Created (or already existing) resource
	 */
	public Resource addResource (String uriRes);

	/**
	 * @param subject
	 * @param predicate
	 * @param predRange
	 * @param predValue
	 */
	public Statement addStatement (String subject, String predicate,
			String predValue) throws OntPropertyNotFoundException;

	/**
	 * Add statement described by <code>elem</code>
	 * 
	 * @param elem
	 *           Statement description including provider metadata
	 * @throws TypeMismatchException
	 * @throws OntPropertyNotFoundException
	 */
	public void addStatementElements (Quad elem)
			throws OntPropertyNotFoundException;

	/**
	 * Get total number of statements in model
	 * 
	 * @return Total number of statements in model
	 */
	public long countStatements ();

	/**
	 * Dump model to specified file
	 * 
	 * @param filename
	 *           Name of file to which model data should be written
	 * @throws IOException
	 *            File operation error
	 */
	public void dumpToFile (String filename) throws IOException;

	/**
	 * @return Returns the dbWrapper.
	 */
	public DatabaseWrapper getDbWrapper ();

	/**
	 * @return Returns the model.
	 */
	public OntModel getModel ();

	/**
	 * Get information about ontological object specified by <code>uri</code>
	 * 
	 * @param uri
	 *           Object URI
	 * @return Collection of statements describing object including provider metadata
	 * @throws OntPropertyNotFoundException
	 */
	public QuadsList getObjectInfo (String uri)
			throws OntPropertyNotFoundException;

	/**
	 * Get property of specified uri
	 * 
	 * @param uri
	 *           URI of the property
	 * @return property of specified uri
	 * @throws OntPropertyNotFoundException
	 */
	public Property getProperty (String uri) throws OntPropertyNotFoundException;

	/**
	 * Get range of specified property
	 * 
	 * @param prop
	 *           Property to be checked
	 * @return String describing range of specified property or null if property has no range set
	 */
	public String getPropertyRange (Property prop);

	/**
	 * Get range of specified property
	 * 
	 * @param uri
	 *           URI of property to be checked
	 * @return String describing range of specified property or null if property has no range set
	 */
	public String getPropertyRange (String uri);

	/**
	 * Get code of information provider of statement <code>statement</code>
	 * 
	 * @param statement
	 *           Statement to be exemined
	 * @return Code of infomation provider or <code>null</code> when no reifcation has been
	 *         specified
	 * @throws OntPropertyNotFoundException
	 */
	public String getProvider (Statement statement)
			throws OntPropertyNotFoundException;

	/**
	 * Get code of information provider for property <code>propertyURI</code> of resource with URI
	 * <code>resourceURI</code>
	 * 
	 * @param resourceURI
	 *           URI of examined resource
	 * @param propertyURI
	 *           URI of desired property
	 * @return Code of infomation provider or <code>null</code> when no reifcation has been
	 *         specified
	 * @throws OntPropertyNotFoundException
	 *            Thrown when specified resource <code>resourceURI</code> hasn't had property
	 *            <code>propertyURI</code> specified
	 * @throws ResourceNotFoundException
	 *            Thrown when resource with specified URI <code>resourceURI</code> does not exist
	 *            in the model
	 */
	public String getProvider (String resourceURI, String propertyURI)
			throws ResourceNotFoundException, OntPropertyNotFoundException;

	/**
	 * Get resource describied by URI <code>uriRes</code>
	 * 
	 * @param uriRes
	 *           URI describing resource
	 * @return Existing resource or null if no such resource has been found
	 */
	public Resource getResource (String uriRes);

	/**
	 * Get iterator enumerating properties
	 * 
	 * @param uri
	 *           URI of resource
	 * @return iterator enumerating properties
	 */
	public StmtIterator getStatementIter (String uri);

	/**
	 * @return Returns the persistentModel.
	 */
	public boolean isPersistentModel ();

	/**
	 * Perform single operation on model describied by <code>oper</code>
	 * 
	 * @param oper
	 *           Description of operation to be performed
	 * @throws OntPropertyNotFoundException
	 */
	public void performOperation (ModelOperation oper)
			throws OntPropertyNotFoundException;

	/**
	 * Perform operations in order specifed by queue
	 * 
	 * @param queue
	 *           Queue of operations to be performed on wrapped model
	 * @throws OntPropertyNotFoundException
	 */
	public void performOperations (Queue queue)
			throws OntPropertyNotFoundException;

	/**
	 * Perform specified RDQL query on wrapped model
	 * 
	 * @param queryString
	 *           Query to be performed
	 * @return Results of specified query
	 */
	public QueryResults performRDQLQuery (String queryString);

	/**
	 * Print properties of resource with specified URI
	 * 
	 * @param uri
	 *           URI of resource
	 */
	public void printProperties (String uri);

	/**
	 * Remove statement described by <code>elem</code>
	 * 
	 * @param elem
	 *           Statement description including provider metadata
	 * @throws OntPropertyNotFoundException
	 */
	public void removeStatementElements (Quad elem)
			throws OntPropertyNotFoundException;

	/**
	 * @return Returns the dbWrapper.
	 */
	public void setDbWrapper (DatabaseWrapper wrapper);

	/**
	 * @param model
	 */
	public void setModel (OntModel model);

	/**
	 * @return Sets the persistentModel.
	 */
	public void setPersistentModel (boolean isPersistent);
	
	/**
	 * Perform RDQL query described by <code>queryString</code>
	 * @param queryString Expression in RDQL to be executed
	 * @return Query results
	 */
	public QueryResults executeRDQLQuery (String queryString);
	
	
	/**
	 * Build report about contents in database
	 * @param ontClass String describing URI of ontological class we're interested in
	 * @return String containing report about contents in database
	 * @throws OntPropertyNotFoundException 
	 */
	public String buildReport (String ontClass) throws OntPropertyNotFoundException;
	
	/**
	 * Enter critical section, request access for reading
	 */
	public void enterReadCriticalSection();
	
	/**
	 * Enter critical section, request access for writing
	 */	
	public void enterWriteCriticalSection();
	
	/**
	 * Leave critical section
	 */	
	public void leaveCriticalSection();
	
	/**
	 * Find objects that are outdated (current time is later than appointed update time) 
	 * @return List of URI of outdated objects or null if no objects need update
	 * @throws OntPropertyNotFoundException 
	 * @throws ParseException 
	 */
	public ArrayList findOutdatedObjects () throws OntPropertyNotFoundException, ParseException;
}
