package logrus.ontmodel;

import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ModelMaker;

import java.sql.SQLException;

import logrus.ontmodel.exceptions.ModelNotFoundException;

/**
 * Factory for IOntModelWrapper, singleton
 */
public class OntModelWrapperFactory
{
	// singleton instance
	private static OntModelWrapperFactory instance = null;
	
	/**
	 * Constructor kept private (following singleton design pattern) 
	 */
	private OntModelWrapperFactory ()
	{
		super();
	}
	
	/**
	 * Get instance of OntModelWrapperFactory
	 * @return Instance of OntModelWrapperFactory
	 */
	public static OntModelWrapperFactory getInstance ()
	{
		if (OntModelWrapperFactory.instance == null)
		{
			OntModelWrapperFactory.instance = new OntModelWrapperFactory();
		}
		return OntModelWrapperFactory.instance;
	}
	
	/**
	 * Loads ontological model from database
	 * @param dbName Database name of desired model
	 * @param dbWrapper Wrapper used for access to the database
	 * @return Created model 
	 */
	public IOntModelWrapper createModelFromDatabase (String dbName,
			DatabaseWrapper dbWrapper) throws ModelNotFoundException
	{
		IOntModelWrapper wrapper = new OntModelWrapperImpl();
		wrapper.setDbWrapper(dbWrapper);
		//this.model = this.retrieveModelFromDatabase(name);
		IDBConnection conn = null;
		try
		{
			conn = dbWrapper.connect();
		}
		catch (SQLException exc)
		{
			exc.printStackTrace();
			return null;
		}
		catch (ClassNotFoundException exc)
		{
			exc.printStackTrace();
			return null;
		}
		if (!conn.containsModel(dbName))
		{
			throw new ModelNotFoundException("Can't retrive model \""+dbName
					+"\" - model doesn't exist in the current database");
		}
		else
		{
			ModelMaker mMaker = ModelFactory.createModelRDBMaker(conn);
			OntModelSpec spec = new OntModelSpec(OntModelSpec.RDFS_MEM);
			spec.setModelMaker(mMaker);
			Model base = mMaker.openModel(dbName);
			wrapper.setModel(ModelFactory.createOntologyModel(spec,base));
			wrapper.setPersistentModel(true);
			return wrapper;
		}		
	}	
	
	/**
	 * Loads ontological model from database using default database wrapper 
	 * @param dbName Database name of desired model
	 * @return Created model 
	 */
	public IOntModelWrapper createModelFromDatabase (String dbName)
		throws ModelNotFoundException
	{
		return this.createModelFromDatabase(dbName, new DatabaseWrapper());
	}
	
	/**
	 * Loads onotological model from specified file and 
	 * @param filename File containing description of the model
	 * @return
	 */
	public IOntModelWrapper createModelFromFile (String filename)
	{
		return this.createModelFromFile(filename,OntModelSpec.RDFS_MEM);
	}
	
	/**
	 * Loads onotological model from specified file
	 * @param filename File containing description of the model
	 * @param spec Specification of the model (memory mamagement and inference type)
	 * @return Created model
	 */
	public IOntModelWrapper createModelFromFile (String filename, OntModelSpec spec) 
	{
		String[] arr = new String[1];
		arr[0] = filename;
		return this.createModelFromFiles(arr,spec);
	}	
	
	/**
	 * Creates onotological model form and loads content form list of files
	 * @param filenames List of files from which content will be loaded
	 * @param spec Specification of the model (memory mamagement and inference type)
	 * @return Created model
	 */
	public IOntModelWrapper createModelFromFiles (String[] filenames, OntModelSpec spec)
	{
		IOntModelWrapper wrapper = new OntModelWrapperImpl();
		Model base = ModelFactory.createDefaultModel();
		for (int i=0; i<filenames.length; i++)
		{
			base.read(filenames[i]);			
		}
		wrapper.setModel(ModelFactory.createOntologyModel(spec,base));
		wrapper.setPersistentModel(false);	
		return wrapper;		
	}
}
