package logrus.ontmodel;

import jade.util.leap.ArrayList;

import java.io.FileOutputStream;
import java.io.IOException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.TimeZone;

import logrus.ontmodel.exceptions.MatchingException;
import logrus.ontmodel.exceptions.OntPropertyNotFoundException;
import logrus.ontmodel.exceptions.ResourceNotFoundException;
import logrus.tools.queue.Queue;

import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ModelLock;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.RSIterator;
import com.hp.hpl.jena.rdf.model.ReifiedStatement;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdql.Query;
import com.hp.hpl.jena.rdql.QueryEngine;
import com.hp.hpl.jena.rdql.QueryExecution;
import com.hp.hpl.jena.rdql.QueryResults;
import com.hp.hpl.jena.shared.PropertyNotFoundException;
import com.hp.hpl.jena.vocabulary.RDFS;

/**
 */
public class OntModelWrapperImpl implements IOntModelWrapper
{
	private static HashMap mapTypes = null;

	private interface TypeXSD
	{
		public static int UnknownType = 0;

		public static int BooleanType = 1;

		public static int CharType = 2;

		public static int DoubleType = 3;

		public static int FloatType = 4;

		public static int LongType = 5;

		public static int StringType = 6;

		public static int AnyURIType = 7;

		public static int NodeType = 8;
	};

	public Integer getTypeMarker (String strType)
	{
		Integer iType = new Integer(TypeXSD.UnknownType);
		if (mapTypes == null)
		{
			mapTypes = new HashMap();
			mapTypes.put(Prefixes.XSD_BOOLEAN,new Integer(TypeXSD.BooleanType));
			mapTypes.put(Prefixes.XSD_CHAR,new Integer(TypeXSD.CharType));
			mapTypes.put(Prefixes.XSD_DOUBLE,new Integer(TypeXSD.DoubleType));
			mapTypes.put(Prefixes.XSD_FLOAT,new Integer(TypeXSD.FloatType));
			mapTypes.put(Prefixes.XSD_LONG,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_INTEGER,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_POSITIVE_INTEGER,new Integer(
					TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_NEGATIVE_INTEGER,new Integer(
					TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_NON_NEGATIVE_INTEGER,new Integer(
					TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_NON_POSITIVE_INTEGER,new Integer(
					TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_UNSINED_LONG,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_INT,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_UNSIGNED_INT,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_SHORT,new Integer(TypeXSD.LongType));
			mapTypes
					.put(Prefixes.XSD_UNSIGNED_SHORT,new Integer(TypeXSD.LongType));
			mapTypes.put(Prefixes.XSD_STRING,new Integer(TypeXSD.StringType));
			mapTypes.put(Prefixes.XSD_NORMALIZED_STRING,new Integer(
					TypeXSD.StringType));
			mapTypes.put(Prefixes.XSD_TOKEN,new Integer(TypeXSD.StringType));
			mapTypes.put(Prefixes.XSD_URI,new Integer(TypeXSD.AnyURIType));
		}
		Object obj = mapTypes.get(strType);
		if (obj == null)
		{
			return new Integer(TypeXSD.UnknownType);
		}
		else
		{
			return (Integer)obj;
		}
	}

	/**
	 * Make a copy of specified model
	 * 
	 * @param source
	 *           Model to be copied
	 * @return Copy of <code>source</code> model
	 */
	public static Model cloneModel (Model source)
	{
		Model copy = ModelFactory.createDefaultModel();
		copy.add(source);
		return copy;
	}

	/**
	 * Make a copy of specified model
	 * 
	 * @param source
	 *           Model to be copied
	 * @return Copy of <code>source</code> model
	 */
	public static OntModel cloneModel (OntModel source)
	{
		OntModel copy = ModelFactory.createOntologyModel(OntModelSpec.RDFS_MEM,
				null);
		copy.add(source);
		return copy;
	}

	/**
	 * Access to the database
	 */
	private DatabaseWrapper dbWrapper;

	/**
	 * Wrapped ontological model
	 */
	private OntModel model;

	/**
	 * Determines if object is wrapping persistent model
	 */
	private boolean persistentModel;

	/**
	 * 
	 */
	public OntModelWrapperImpl ()
	{
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#addDescriptiveStatement(String,String,String,String,String)
	 */
	public void addDescriptiveStatement (String providerID, String subject,
			String predicate, String predValue)
			throws OntPropertyNotFoundException
	{
		// first add proper statement...
		Statement stmt = this.addStatement(subject,predicate,predValue);
		// ...then add statement describing information provider (using reification)
		Resource refStmt = this.model.getAnyReifiedStatement(stmt);
		refStmt.addProperty(this.getProperty(Prefixes.ProviderProperty),
				providerID);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#insertObjectInfo(QuadsList)
	 */
	public void addObjectInfo (QuadsList info)
			throws OntPropertyNotFoundException
	{
		for (jade.util.leap.Iterator iter = info.getList().iterator(); iter
				.hasNext();)
		{
			this.addStatementElements((Quad)iter.next());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#addResource(String)
	 */
	public Resource addResource (String uriRes)
	{
		Resource res = this.model.getResource(uriRes);
		if (res == null)
		{
			res = this.model.createResource(uriRes);
		}
		return res;
	}

	private Statement addBooleanProperty (Resource resource, Property property,
			String object)
	{
		boolean booleanValue = false;
		if (object.toLowerCase().compareTo("true") == 0)
		{
			booleanValue = true;
		}
		resource.addProperty(property,booleanValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				booleanValue);
		return iterator.nextStatement();
	}

	private void checkRemove (Statement stmt, String provider, Queue toRemove)
			throws OntPropertyNotFoundException
	{
		for (RSIterator rsIter = stmt.listReifiedStatements(); rsIter.hasNext();)
		{
			Statement provStmt = rsIter.nextRS().getProperty(
					this.getProperty(Prefixes.ProviderProperty));
			if (provStmt.getObject().toString().compareTo(provider) == 0)
			{
				toRemove.inject(provStmt);
				toRemove.inject(stmt);
			}
		}
	}

	private void prepareRemoveBooleanProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		boolean booleanValue = false;
		if (object.toLowerCase().compareTo("true") == 0)
		{
			booleanValue = true;
		}
		for (StmtIterator iter = this.model.listStatements(resource,property,
				booleanValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addCharProperty (Resource resource, Property property,
			String object)
	{
		char charValue = object.charAt(0);
		resource.addProperty(property,charValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				charValue);
		return iterator.nextStatement();
	}

	private void prepareRemoveCharProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		char charValue = object.charAt(0);
		for (StmtIterator iter = this.model.listStatements(resource,property,
				charValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addDoubleProperty (Resource resource, Property property,
			String object)
	{
		double doubleValue = Double.parseDouble(object);
		resource.addProperty(property,doubleValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				doubleValue);
		return iterator.nextStatement();
	}

	private void prepareRemoveDoubleProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		double doubleValue = Double.parseDouble(object);
		for (StmtIterator iter = this.model.listStatements(resource,property,
				doubleValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addFloatProperty (Resource resource, Property property,
			String object)
	{
		float floatValue = Float.parseFloat(object);
		resource.addProperty(property,floatValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				floatValue);
		return iterator.nextStatement();
	}

	private void prepareRemoveFloatProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		float floatValue = Float.parseFloat(object);
		for (StmtIterator iter = this.model.listStatements(resource,property,
				floatValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addLongProperty (Resource resource, Property property,
			String object)
	{
		long longValue = Long.parseLong(object);
		resource.addProperty(property,longValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				longValue);
		return iterator.nextStatement();
	}

	private void prepareRemoveLongProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		long longValue = Long.parseLong(object);
		for (StmtIterator iter = this.model.listStatements(resource,property,
				longValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addStringProperty (Resource resource, Property property,
			String object)
	{
		resource.addProperty(property,object);
		StmtIterator iterator = this.model.listStatements(resource,property,
				object);
		return iterator.nextStatement();
	}

	private void prepareRemoveStringProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		for (StmtIterator iter = this.model.listStatements(resource,property,
				object); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	private Statement addNodeProperty (Resource resource, Property property,
			String object)
	{
		Resource nodeValue = this.getResource(object);
		resource.addProperty(property,nodeValue);
		StmtIterator iterator = this.model.listStatements(resource,property,
				nodeValue);
		return iterator.nextStatement();
	}

	private void prepareRemoveNodeProperty (Resource resource,
			Property property, String object, String provider, Queue toRemove)
			throws OntPropertyNotFoundException, NoSuchElementException
	{
		Resource nodeValue = this.getResource(object);
		for (StmtIterator iter = this.model.listStatements(resource,property,
				nodeValue); iter.hasNext();)
		{
			this.checkRemove(iter.nextStatement(),provider,toRemove);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#addDescriptiveStatement(String,String,String,String)
	 */
	public Statement addStatement (String subject, String predicate,
			String predValue) throws OntPropertyNotFoundException
	{
		// add resource if it not exists in model
		Resource resource = this.getResource(subject);

		// first get the property object
		Property property = this.model.getProperty(predicate);

		// if predicate is rdf:type use string
		if (predicate.compareTo(Prefixes.RDF_TYPE) == 0)
		{
			resource.addProperty(property,predValue);
			StmtIterator iterator = this.model.listStatements(resource,property,
					predValue);
			return iterator.nextStatement();

		}
		else
		{
			// get property type
			String strRange = this.getPropertyRange(property);
			switch (this.getTypeMarker(strRange).intValue())
			{
				case TypeXSD.BooleanType:
					return this.addBooleanProperty(resource,property,predValue);
				case TypeXSD.CharType:
					return this.addCharProperty(resource,property,predValue);
				case TypeXSD.DoubleType:
					return this.addDoubleProperty(resource,property,predValue);
				case TypeXSD.FloatType:
					return this.addFloatProperty(resource,property,predValue);
				case TypeXSD.LongType:
					return this.addLongProperty(resource,property,predValue);
				case TypeXSD.StringType:
				case TypeXSD.AnyURIType:
					return this.addStringProperty(resource,property,predValue);
				case TypeXSD.UnknownType:
				default:
					// assume that it's a resource denotated by URI
					return this.addNodeProperty(resource,property,predValue);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#addStatementElements(Quad)
	 */
	public void addStatementElements (Quad elem)
			throws OntPropertyNotFoundException
	{
		this.addDescriptiveStatement(elem.getProvider(),elem.getSubject(),elem
				.getPredicate(),elem.getObject());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#countStatements()
	 */
	public long countStatements ()
	{
		long num = 0;
		for (StmtIterator iter = this.model.listStatements(); iter.hasNext();)
		{
			iter.next();
			num++;
		}
		return num;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#dumpToFile(String)
	 */
	public void dumpToFile (String filename) throws IOException
	{
		FileOutputStream outStream = new FileOutputStream(filename);
		this.model.write(outStream);
		outStream.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getDbWrapper()
	 */
	public DatabaseWrapper getDbWrapper ()
	{
		return this.dbWrapper;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getModel()
	 */
	public OntModel getModel ()
	{
		return this.model;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getObjectInfo(String)
	 */
	public QuadsList getObjectInfo (String uri)
			throws OntPropertyNotFoundException
	{
		QuadsList info = new QuadsList();
		// first get object
		Resource resource = this.model.getResource(uri);
		if (resource == null)
		{
			return null;
		}

		// new implementation
		Property prop = null;
		RDFNode node = null;
		for (StmtIterator iter = this.model.listStatements(resource,prop,node); iter
				.hasNext();)
		{
			Statement stmt = (Statement)iter.next();
			String predicate = stmt.getPredicate().getURI();
			String object = stmt.getObject().toString();
			String provider = null;
			// get provider ID
			boolean wasReif = false;
			for (RSIterator rsIter = stmt.listReifiedStatements(); rsIter
					.hasNext();)
			{
				ReifiedStatement refStmt = null;
				Statement provStmt = null;
				try
				{
					refStmt = rsIter.nextRS();
					provStmt = refStmt.getRequiredProperty(this
							.getProperty(Prefixes.ProviderProperty));
				}
				catch (PropertyNotFoundException exc)
				{
					System.out.println("No reified property for " + stmt);
					System.out.println(refStmt);
				}
				if (provStmt != null)
				{
					provider = provStmt.getObject().toString();
					wasReif = true;
					break;
				}
			}
			if (!wasReif)
			{
				System.out.println("No reification for " + stmt);
			}
			info.add(uri,predicate,object,provider);
		}
		return info;

		/*
		 * // get it's properties for (StmtIterator iter = resource.listProperties(); iter.hasNext();) {
		 * Statement stmt = (Statement)iter.next(); String predicate = stmt.getPredicate().getURI();
		 * String object = stmt.getObject().toString(); String provider = null; boolean wasRef =
		 * false; // get provider ID for (RSIterator rsIter = stmt.listReifiedStatements(); rsIter
		 * .hasNext();) { wasRef = true; Statement provStmt = rsIter.nextRS().getRequiredProperty(
		 * this.getProperty(Prefixes.ProviderProperty)); if (provStmt != null) { provider =
		 * provStmt.getObject().toString(); break; } else { System.out.println("Provider
		 * NULL!!!!!!!!"); } } if (!wasRef) { System.out.println("No reification!!!!"); }
		 * info.add(uri,predicate,object,provider); } return info;
		 */
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getProperty(String)
	 */
	public Property getProperty (String uri) throws OntPropertyNotFoundException
	{
		Property prop = this.model.getProperty(uri);
		if (prop == null)
		{
			throw new OntPropertyNotFoundException(uri);
		}
		return prop;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getPropertyRange(Property)
	 */
	public String getPropertyRange (Property prop)
	{
		Statement stmtProp = prop.getProperty(RDFS.range);
		if (stmtProp != null)
		{
			return stmtProp.getObject().toString();
		}
		else
		{
			return "";
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getPropertyRange(String)
	 */
	public String getPropertyRange (String uri)
	{
		return this.getPropertyRange(this.model.getProperty(uri));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getProvider (Statement statement)
	 */
	public String getProvider (Statement statement)
			throws OntPropertyNotFoundException
	{
		Property propertyProvider = this.getProperty(Prefixes.ProviderProperty);
		for (RSIterator iter = statement.listReifiedStatements(); iter.hasNext();)
		{
			Statement stmtRef = iter.nextRS().getProperty(propertyProvider);
			if (stmtRef != null)
			{
				return stmtRef.getObject().toString();
			}
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getProvider (String,String)
	 */
	public String getProvider (String resourceURI, String propertyURI)
			throws ResourceNotFoundException, OntPropertyNotFoundException
	{
		Resource resource = this.getResource(resourceURI);
		if (resource == null)
		{
			throw new ResourceNotFoundException(resourceURI);
		}
		Property propertyObj = this.getProperty(propertyURI);
		Statement stmtProp = resource.getProperty(propertyObj);
		return this.getProvider(stmtProp);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getResource(String)
	 */
	public Resource getResource (String uriRes)
	{
		return this.model.getResource(uriRes);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#getStatementIter(String)
	 */
	public StmtIterator getStatementIter (String uri)
	{
		return this.getResource(uri).listProperties();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#isPersistentModel()
	 */
	public boolean isPersistentModel ()
	{
		return this.persistentModel;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#performOperation(ModelOperation)
	 */
	public void performOperation (ModelOperation oper)
			throws OntPropertyNotFoundException
	{
		switch (oper.getAction())
		{
			case ModelOperation.ADD:
				this.addStatementElements(oper.getData());
				break;
			case ModelOperation.REMOVE:
				this.removeStatementElements(oper.getData());
				break;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#performOperations(Queue)
	 */
	public void performOperations (Queue queue)
			throws OntPropertyNotFoundException
	{
		for (Iterator iter = queue.iterator(); iter.hasNext();)
		{
			this.performOperation((ModelOperation)iter.next());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#performRDQLQuery(java.lang.String)
	 */
	public QueryResults performRDQLQuery (String queryString)
	{
		// write lock - to avoid adding/removing statements
		// by other agents while query is performed
		this.model.enterCriticalSection(false);
		// perform query
		QueryResults results = Query.exec(queryString,this.model);
		this.model.leaveCriticalSection();
		return results;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#printProperties(String)
	 */
	public void printProperties (String uri)
	{
		StmtIterator iter = this.getStatementIter(uri);
		while (iter.hasNext())
		{
			System.out.println(iter.next().toString());
		}
		iter.close();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#removeStatementElements(Quad)
	 */
	public void removeStatementElements (Quad elem)
			throws OntPropertyNotFoundException
	{
		Resource resource = this.getResource(elem.getSubject());
		Property property = this.getProperty(elem.getPredicate());

		Queue toRemove = new Queue();

		String strRange = this.getPropertyRange(property);
		switch (this.getTypeMarker(strRange).intValue())
		{
			case TypeXSD.BooleanType:
				this.prepareRemoveBooleanProperty(resource,property,elem
						.getObject(),elem.getProvider(),toRemove);
				break;
			case TypeXSD.CharType:
				this.prepareRemoveCharProperty(resource,property,elem.getObject(),
						elem.getProvider(),toRemove);
				break;
			case TypeXSD.DoubleType:
				this.prepareRemoveDoubleProperty(resource,property,
						elem.getObject(),elem.getProvider(),toRemove);
				break;
			case TypeXSD.FloatType:
				this.prepareRemoveFloatProperty(resource,property,elem.getObject(),
						elem.getProvider(),toRemove);
				break;
			case TypeXSD.LongType:
				this.prepareRemoveLongProperty(resource,property,elem.getObject(),
						elem.getProvider(),toRemove);
				break;
			case TypeXSD.StringType:
			case TypeXSD.AnyURIType:
				this.prepareRemoveStringProperty(resource,property,
						elem.getObject(),elem.getProvider(),toRemove);
				break;
			case TypeXSD.UnknownType:
			default:
				this.prepareRemoveNodeProperty(resource,property,elem.getObject(),
						elem.getProvider(),toRemove);
				break;
		}

		// phase 2: remove statements from model
		for (Iterator iter = toRemove.iterator(); iter.hasNext();)
		{
			this.model.remove((Statement)iter.next());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#setDbWrapper(logrus.ontmodel.DatabaseWrapper)
	 */
	public void setDbWrapper (DatabaseWrapper wrapper)
	{
		this.dbWrapper = wrapper;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#setModel(com.hp.hpl.jena.ontology.OntModel)
	 */
	public void setModel (OntModel model)
	{
		this.model = model;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#setPersistentModel(boolean)
	 */
	public void setPersistentModel (boolean isPersistent)
	{
		this.persistentModel = isPersistent;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#extractMatching(QuadsList)
	 */
	public QuadsList findMatchingObject (QuadsList objInfo)
			throws MatchingException
	{
		// extract onotological class name
		String ontClass = objInfo.getOntClass();

		if (ontClass == null)
		{
			// error: no ontological class specified
			throw new MatchingException(
					"No ontological class specified in object info");
		}

		// extract restaurant tag from object
		IObjectInfoTag interfTag = ObjectInfoTagFactory.getInstance()
				.createObjectInfoTag(ontClass);
		if (interfTag == null)
		{
			// error: no tag definition bound to this ontological class
			throw new MatchingException(
					"No tag definition bound to the ontological class " + ontClass);
		}

		IObjectInfoTag tag = interfTag.getTag(objInfo);
		if (!tag.isComplete())
		{
			// error: searching for incomplete tag is not supported
			throw new MatchingException(
					"Searching for incomplete tag is not supported");
		}

		// create searcher for this class
		IObjectInfoSearcherByTag searcherByTag = ObjectInfoSearcherByTagFactory
				.getInstance().createObjectSearcherByTag(ontClass);

		if (searcherByTag == null)
		{
			// error: no searcher bound to this ontological class
			throw new MatchingException(
					"No searcher bound to the ontological class " + ontClass);
		}

		// perform search on model for object matching tag
		QuadsList quadsList = null;
		try
		{
			quadsList = searcherByTag.findObject(this,tag);
		}
		catch (OntPropertyNotFoundException e)
		{
			// severe error: can't find property in model
			throw new MatchingException("Can't find property in model " + ontClass);
		}

		return quadsList;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#executeRDQLQuery(String)
	 */
	public QueryResults executeRDQLQuery (String queryString)
	{
		Query query = new Query(queryString);
		query.setSource(this.model);
		QueryExecution queryEngine = new QueryEngine(query);
		return queryEngine.exec();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#buildReport()
	 */
	public String buildReport (String ontClass)
			throws OntPropertyNotFoundException
	{
		StringBuffer strBuff = new StringBuffer();

		// get date and place report header
		Calendar cal = Calendar.getInstance(TimeZone.getDefault());
		String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
		sdf.setTimeZone(TimeZone.getDefault());
		strBuff.append("Report created ");
		strBuff.append(sdf.format(cal.getTime()));
		strBuff.append('\n');

		// prepare predicates
		Property rdfTypeProp = this.getProperty(Prefixes.RDF_TYPE);
		Property provLinkProp = this.getProperty(Prefixes.SourceLink);

		// total number of restaurants
		int countTotal = 0;

		// number of restaurants with multiple providers
		int countMultiProv = 0;

		// calculate number of objects in model with specified ontological class
		StmtIterator mainIterator = this.model.listStatements(null,rdfTypeProp,
				ontClass);
		while (mainIterator.hasNext())
		{
			// count object
			++countTotal;

			// check how many providers does it have
			Statement stmt = mainIterator.nextStatement();
			Resource resObj = stmt.getSubject();
			StmtIterator provIterator = resObj.listProperties(provLinkProp);
			int provCount = 0;
			while (provIterator.hasNext())
			{
				provIterator.next();
				provCount++;
			}
			if (provCount > 1)
			{
				++countMultiProv;
			}
		}

		strBuff.append("There are " + countTotal + " objects in model\n");
		strBuff.append("" + countMultiProv + " have more then 1 data provider.\n");
		
		// calculate total number of statments in model
		rdfTypeProp = null;
		String strNull = null;
		StmtIterator stmtIterator = this.model.listStatements(null,rdfTypeProp,
				strNull);
		int allStmts = 0;
		while (stmtIterator.hasNext())
		{
			++allStmts;
			stmtIterator.next();
		}
		strBuff.append("There are "+allStmts+" statements in total");
		return strBuff.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#enterReadCriticalSection()
	 */
	public void enterReadCriticalSection ()
	{
		this.model.enterCriticalSection(ModelLock.READ);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#enterWriteCriticalSection()
	 */
	public void enterWriteCriticalSection ()
	{
		this.model.enterCriticalSection(ModelLock.WRITE);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#leaveCriticalSection()
	 */
	public void leaveCriticalSection ()
	{
		this.model.leaveCriticalSection();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see logrus.ontmodel.IOntModelWrapper#findOutdatedObjects()
	 */
	public ArrayList findOutdatedObjects () throws OntPropertyNotFoundException,
			ParseException
	{
		ArrayList list = new ArrayList();
		Property rdfTypeProp = this.getProperty(Prefixes.RDF_TYPE);
		Property stampNextProp = this.getProperty(Prefixes.TIMESTAMP_NEXT);
		Property sourceLink = this.getProperty(Prefixes.SourceLink);
		String ontClass = Prefixes.CLASS_RES;

		StmtIterator mainIter = this.model.listStatements(null,rdfTypeProp,
				ontClass);
		while (mainIter.hasNext())
		{
			// get term of validity for the object
			Statement objStmt = mainIter.nextStatement();
			Resource subject = objStmt.getSubject();
			Statement nextStmt = subject.listProperties(stampNextProp)
					.nextStatement();
			String strNext = nextStmt.getObject().toString();
			Date dateNext = DateFormat.getInstance().parse(strNext);
			Date dateNow = Calendar.getInstance().getTime();

			// check if object is outdated
			if (dateNow.compareTo(dateNext) > 0)
			{
				// get links to it's data sources
				String strSourceLink = null;
				StmtIterator iterSources = this.model.listStatements(subject,
						sourceLink,strSourceLink);
				while (iterSources.hasNext())
				{
					Statement stmtSource = iterSources.nextStatement();
					list.add(stmtSource.getObject().toString());
				}
			}
		}
		if (list.size() > 0)
		{
			return list;
		}
		else
		{
			return null;
		}
	}
}
