package logrus.ontmodel;

import jade.content.Concept;
import jade.content.ContentElement;
import jade.util.leap.Serializable;

public class Quad implements ContentElement, Concept, Serializable
{
	// statement's subject
	private String subject;

	// statement's predicate
	private String predicate;

	// statement's object
	private String object;

	// data source of the statement
	private String provider;

	public Quad ()
	{
		this.subject = null;
		this.predicate = null;
		this.object = null;
		this.provider = null;
	}

	/**
	 * @param subject
	 *           Statement's subject
	 * @param predicate
	 *           Statement's predicate
	 * @param object
	 *           Statement's object
	 * @param provider
	 *           Data source of the statement
	 */
	public Quad (String subject, String predicate, String object, String provider)
	{
		super();
		this.setSubject(subject);
		this.setPredicate(predicate);
		this.setObject(object);
		this.setProvider(provider);
	}

	/**
	 * Check if at least one of quad elements is null
	 * 
	 * @return True if subject, predicate, object or provider is null, otherwise return false
	 */
	public boolean hasNull ()
	{
		return this.subject == null || this.predicate == null
				|| this.object == null || this.provider == null;
	}

	/**
	 * Calculate hash (concatenation of subject, predicate, object, provider)
	 * 
	 * @return hash
	 */
	public String calculateHash ()
	{
		return this.subject + this.predicate + this.object + this.provider;
	}

	/**
	 * Convert instance to string
	 */
	public String toString ()
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("<");
		buffer.append(this.subject);
		buffer.append("> ");
		buffer.append("<");
		buffer.append(this.predicate);
		buffer.append("> ");
		buffer.append("<");
		buffer.append(this.object);
		buffer.append("> ");
		buffer.append("<");
		buffer.append(this.provider);
		buffer.append(">");
		return buffer.toString();
	}

	/**
	 * Check if this Quad describes the statement with specified subject, predicate, object provided
	 * by provider
	 * 
	 * @param subject
	 *           Statement's subject
	 * @param predicate
	 *           Statement's predicate
	 * @param object
	 *           Statement's object
	 * @param provider
	 *           Data source of the statement
	 * @return Boolean value indicating statements equality
	 */
	public boolean isEqual (String subject, String predicate, String object,
			String provider)
	{
		if (this.subject.compareTo(subject) == 0
				&& this.predicate.compareTo(predicate) == 0
				&& this.object.compareTo(object) == 0
				&& this.provider.compareTo(provider) == 0)
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	public Object clone ()
	{
		return new Quad(this.subject,this.predicate,this.object,this.provider);
	}

	/**
	 * Check if this quad describes the same quad as <code>elem</code>
	 * 
	 * @param elem
	 *           Quad to be compared
	 * @return True if both quads have the same meaning, otherwise returns false
	 */
	public boolean isEqual (Quad elem)
	{
		return this.isEqual(elem.getSubject(),elem.getPredicate(),elem
				.getObject(),elem.getProvider());
	}

	/**
	 * Check if this quad describes the same RDF statement as <code>elem</code> (do not include
	 * provider ID in comparison)
	 * 
	 * @param elem
	 *           Quad to be compared
	 * @return True if both quads have the same meaning (describing the same RDF Statement),
	 *         otherwise returns false
	 */
	public boolean isRDFStatementEqual (Quad elem)
	{
		return this.subject.compareTo(elem.getSubject()) == 0
				&& this.predicate.compareTo(elem.getPredicate()) == 0
				&& this.object.compareTo(elem.getObject()) == 0;
	}

	public void setSubject (String subject)
	{
		this.subject = subject;
	}

	public String getSubject ()
	{
		return this.subject;
	}

	public void setPredicate (String predicate)
	{
		this.predicate = predicate;
	}

	public String getPredicate ()
	{
		return this.predicate;
	}

	public void setObject (String object)
	{
		this.object = object;
	}

	public String getObject ()
	{
		return this.object;
	}

	public void setProvider (String provider)
	{
		this.provider = provider;
	}

	public String getProvider ()
	{
		return this.provider;
	}

	/**
	 * Check if this quad describes statement meeting specified criterions, if criterion is null then
	 * it's omitted while comparing (such slot may be equal anything)
	 * 
	 * @param subjectPattern Subject criterion
	 * @param predicatePattern Predicate criterion
	 * @param objectPattern Object criterion
	 * @param providerPattern Provider criterion
	 * @return True if quad matches all criterions, otherwise return false
	 */
	public boolean isMatchingPattern (String subjectPattern,
			String predicatePattern, String objectPattern, String providerPattern)
	{
		// check subject
		if (subjectPattern != null && this.subject.compareTo(subjectPattern) != 0)
			return false;

		// check predicate
		if (predicatePattern != null
				&& this.predicate.compareTo(predicatePattern) != 0)
			return false;

		// check object
		if (objectPattern != null && this.object.compareTo(objectPattern) != 0)
			return false;

		// check provider
		if (providerPattern != null
				&& this.provider.compareTo(providerPattern) != 0)
			return false;

		return true;
	}
}
