/**
 * 
 */
package logrus.ontmodel;

import jade.content.Concept;
import jade.content.ContentElement;
import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;
import jade.util.leap.List;

import java.text.DateFormat;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;

import logrus.agents.datasources.DataProviderID;
import logrus.ontmodel.restaurant.RestaurantEssentials;
import logrus.tools.queue.Queue;
import logrus.tools.settings.SettingsManager;

/**
 * 
 */
public class QuadsList implements Concept, ContentElement
{
	private ArrayList list;

	/**
	 * 
	 */
	public QuadsList ()
	{
		super();
		this.list = new ArrayList();
	}

	/**
	 * Substitute subject in all statements to <code>newSubject</code>
	 * 
	 * @param newSubject
	 *           New subject for all statements
	 */
	public void substituteSubject (String newSubject)
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			((Quad)iter.next()).setSubject(newSubject);
		}
	}

	/**
	 * Assume that all statements describe the same resource, extract URI of this resource
	 * 
	 * @return URI of this resource or null if statements list is empty
	 */
	public String getObjectURI ()
	{
		if (this.list.size() > 0)
		{
			return ((Quad)this.list.get(0)).getSubject();
		}
		else
		{
			return null;
		}
	}

	/**
	 * Assume that all statements have the same provider, extract ID of this provider
	 * 
	 * @return Provider for this statements
	 */
	public String getProviderID ()
	{
		if (this.list.size() > 0)
		{
			return ((Quad)this.list.get(0)).getProvider();
		}
		else
		{
			return null;
		}
	}

	/**
	 * Remove all statements provided by data source identified by <code>providerID</code>
	 * 
	 * @param providerID
	 *           Data source identifier
	 */
	public void removeStatmentsWithProviderID (String providerID)
	{
		ArrayList toDelete = new ArrayList();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.getProvider().compareTo(providerID) == 0
					&& !RestaurantEssentials.getInstance().isOnList(
							quad.getPredicate()))
			{
				toDelete.add(quad);
			}
		}
		for (Iterator iter = toDelete.iterator(); iter.hasNext();)
		{
			this.list.remove(iter.next());
		}
	}

	/**
	 * Remove all statements mathcing specified pattern
	 * 
	 * @param pattern
	 *           Quad pattern used for matching quads
	 */
	public void removeStatementsMatchingPattern (Quad pattern)
	{
		ArrayList toDelete = new ArrayList();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.isMatchingPattern(pattern.getSubject(),
					pattern.getPredicate(),pattern.getObject(),pattern.getObject()))
			{
				toDelete.add(quad);
			}
		}
		for (Iterator iter = toDelete.iterator(); iter.hasNext();)
		{
			this.list.remove(iter.next());
		}
	}

	/**
	 * Get name of ontological class if specified
	 * 
	 * @return Name of ontological class if specified with rdf:type statement, if not specified then
	 *         returns null
	 */
	public String getOntClass ()
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.getPredicate().compareTo(Prefixes.RDF_TYPE) == 0)
			{
				return quad.getObject();
			}
		}
		return null;
	}

	/**
	 * Clean quads list (remove all quads that have at least one null value)
	 */
	public void cleanQuadsWithNull ()
	{
		ArrayList toRemove = new ArrayList();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.hasNull())
			{
				toRemove.add(quad);
			}
		}
		for (Iterator iter = toRemove.iterator(); iter.hasNext();)
		{
			this.list.remove(iter.next());
		}
	}

	public List getQuadsCollection ()
	{
		return this.list;
	}

	public void setQuadsCollection (List quads)
	{
		this.list = (ArrayList)quads;
	}

	/**
	 * Add Quad object
	 * 
	 * @param obj
	 *           Object to be added
	 */
	public void add (Quad obj)
	{
		this.getList().add(obj);
	}

	/**
	 * Add Quad object
	 * 
	 * @param subject
	 *           Statement's subject
	 * @param predicate
	 *           Statement's predicate
	 * @param object
	 *           Statement's object
	 * @param provider
	 *           Data source of the statement
	 */
	public void add (String subject, String predicate, String object,
			String provider)
	{
		this.getList().add(new Quad(subject,predicate,object,provider));
	}

	public Object clone ()
	{
		QuadsList clone = new QuadsList();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			clone.add((Quad)((Quad)iter.next()).clone());
		}
		return clone;
	}

	public boolean containsStatement (Quad stmt)
	{
		return this.containsStatement(stmt.getSubject(),stmt.getPredicate(),stmt
				.getObject(),stmt.getProvider());
	}

	/**
	 * Check if collection contains specified statement
	 * 
	 * @param subject
	 * @param predicate
	 * @param object
	 * @param provider
	 * @return
	 */
	public boolean containsStatement (String subject, String predicate,
			String object, String provider)
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			if (((Quad)iter.next()).isEqual(subject,predicate,object,provider))
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Check if collection contains statement matching specified pattern
	 * 
	 * @param subject
	 * @param predicate
	 * @param object
	 * @param provider
	 * @return
	 */
	public boolean containsPattern (String subject, String predicate,
			String object, String provider)
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			if (((Quad)iter.next()).isMatchingPattern(subject,predicate,object,
					provider))
			{
				return true;
			}
		}
		return false;
	}

	// public boolean hasStatementsMatching ()

	/**
	 * Get number of statement descriptions in this collection
	 * 
	 * @return number of statement descriptions in this collection
	 */
	public int count ()
	{
		return this.list.size();
	}

	/**
	 * Get number of statements with property slot equal <code>uriProperty</code>
	 * 
	 * @param uriProperty
	 *           Desired property
	 * @return number of statements with property slot equal <code>uriProperty</code>
	 */
	public int countProperties (String uriProperty)
	{
		int count = 0;
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad stmnt = (Quad)iter.next();
			if (stmnt.getPredicate().compareTo(uriProperty) == 0)
			{
				++count;
			}
		}
		return count;
	}

	/**
	 * Get quad from this object description that has the sam meaning as <code>pattern</code>
	 * 
	 * @param pattern
	 *           Pattern for quad
	 * @return Quad matching <code>pattern</code> or null if such quad does not exist
	 */
	public Quad getQuadMatching (Quad pattern)
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (pattern.isEqual(quad))
			{
				return quad;
			}
		}
		return null;
	}

	public Hashtable getHashtable ()
	{
		Hashtable ht = new Hashtable(this.list.size());
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad elem = (Quad)iter.next();
			String key = elem.getSubject() + elem.getPredicate()
					+ elem.getObject() + elem.getProvider();
			ht.put(key,elem);
		}
		return ht;
	}

	public ArrayList getList ()
	{
		return this.list;
	}

	/**
	 * Extract statements with property slot equal <code>uriProperty</code>
	 * 
	 * @param uriProperty
	 *           Desired property
	 * @return Array of statements with property slot equal <code>uriProperty</code>
	 */
	public Quad[] getStatements (String uriProperty)
	{
		ArrayList auxList = new ArrayList();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad stmnt = (Quad)iter.next();
			if (stmnt.getPredicate().compareTo(uriProperty) == 0)
			{
				auxList.add(stmnt);
			}
		}
		Quad[] stmnts = null;
		if (auxList.size() > 0)
		{
			stmnts = new Quad[auxList.size()];
			int i = 0;
			for (Iterator iter = auxList.iterator(); iter.hasNext();)
			{
				stmnts[i++] = (Quad)iter.next();
			}
		}
		return stmnts;
	}

	/**
	 * Check if <code>info</code> describes the same object as info
	 * 
	 * @param info
	 * @return
	 */
	public boolean isEqual (QuadsList info)
	{
		if (this.count() != info.count())
		{
			return false;
		}

		// old implementation
		// for (Iterator iter = this.list.iterator(); iter.hasNext();)
		// {
		// Quad stmt = (Quad)iter.next();
		// if (!info.containsStatement(stmt))
		// {
		// System.out.println("Niet stejtmenta "+Utils.getInstance().filterNames(stmt.toString()));
		// return false;
		// }
		// }
		// return true;

		// after performance tweaks
		Hashtable ht = info.getHashtable();
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad stmt = (Quad)iter.next();
			Object obj = ht.get(stmt.calculateHash());
			if (obj == null)
			{
				return false;
			}
			Quad stmt2 = (Quad)obj;
			if (!stmt.isEqual(stmt2))
			{
				return false;
			}
		}
		return true;
	}

	public String toString ()
	{
		StringBuffer buffer = new StringBuffer();
		int i = 1;
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			buffer.append('[');
			buffer.append(i++);
			buffer.append("]: ");
			buffer.append(iter.next().toString());
			buffer.append('\n');
		}
		return buffer.toString();
	}

	/**
	 * Remove all quads included in this quads list from <code>queue</code>
	 * 
	 * @param queue
	 *           Queue to be cleaned
	 */
	public void removeSelfFromQueue (Queue queue)
	{
		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			queue.remove(iter.next());
		}
	}

	/**
	 * Extract quads from <code>queue</code> with slots matching specified parameters. If parameter
	 * is null then it's not considered as a siginificant criterion
	 * 
	 * @param queue
	 *           Queue from which quads should be extracted
	 * @param subjectPattern
	 *           URI of filtered subject
	 * @param predicatePattern
	 *           URI of filtered predicate
	 * @param objectPattern
	 *           Value (literal or URI) of filtered object
	 * @param providerPattern
	 *           Identifier of data provider
	 * @return List of matching quads (may be empty but never returns null)
	 */
	public static QuadsList filterStatements (Queue queue,
			String subjectPattern, String predicatePattern, String objectPattern,
			String providerPattern)
	{
		QuadsList stmts = new QuadsList();
		for (java.util.Iterator iter = queue.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.isMatchingPattern(subjectPattern,predicatePattern,
					objectPattern,providerPattern))
			{
				stmts.add(quad);
			}
		}
		return stmts;
	}

	/**
	 * Remove timestamps from this description
	 * 
	 * @return Array storing timestamps of last update at position 0 and next update on position 1
	 */
	public String[] removeTimestamps ()
	{
		Quad lastUpdate = null;
		Quad nextUpdate = null;

		for (Iterator iter = this.list.iterator(); iter.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (quad.getPredicate().compareTo(Prefixes.TIMESTAMP_LAST) == 0)
			{
				lastUpdate = quad;
			}
			if (quad.getPredicate().compareTo(Prefixes.TIMESTAMP_NEXT) == 0)
			{
				nextUpdate = quad;
			}
			if (lastUpdate != null && nextUpdate != null)
			{
				break;
			}
		}

		String[] timestamps = new String[2];
		timestamps[0] = lastUpdate.getObject();
		timestamps[1] = nextUpdate.getObject();

		this.list.remove(lastUpdate);
		this.list.remove(nextUpdate);

		return timestamps;
	}

	/**
	 * Add timestamps to new object. Last update timestampt is set as current date. Next update
	 * timestamp is calculated using current date and default initial term of vailidity taken from
	 * settings.
	 */
	public void stampNewObject ()
	{
		Date now = Calendar.getInstance().getTime();
		long nextMili = now.getTime()
				+ SettingsManager.getSettings().getInitialValidityTerm();
		Date next = new Date(nextMili);
		DateFormat dateFormat = DateFormat.getInstance();
		this.addTimestamp(dateFormat.format(now),dateFormat.format(next));
	}

	/**
	 * Add timestamps (last and next update).
	 * 
	 * @param lastStamp
	 *           Last update timestamp
	 * @param nextStamp
	 *           Next update timestamp
	 */
	private void addTimestamp (String lastStamp, String nextStamp)
	{
		String id = this.getObjectURI();
		this.add(new Quad(id,Prefixes.TIMESTAMP_LAST,lastStamp,
				DataProviderID.OS_INTERNAL));
		this.add(new Quad(id,Prefixes.TIMESTAMP_NEXT,nextStamp,
				DataProviderID.OS_INTERNAL));
	}

	/**
	 * Update timestamps
	 * 
	 * @param prevStamps
	 *           Old timestamps
	 * @param wasUpdate
	 *           Flag indicating if update has been performed
	 * @throws ParseException
	 *            Thrown when error during parsing date from string format has occured
	 */
	public void updateStamps (String[] prevStamps, boolean wasUpdate)
			throws ParseException
	{
		String[] newStamps = this.calculateNewStamps(prevStamps,wasUpdate);
		this.addTimestamp(newStamps[0],newStamps[1]);
	}

	/**
	 * Calculate timestamp for new update
	 * 
	 * @param prevStamps
	 *           Previous timestamps
	 * @param wasUpdate
	 *           Flag indicating if update has been performed
	 * @return Timestamp for new update
	 * @throws ParseException
	 *            Thrown when error during parsing date from string format has occured
	 */
	private String[] calculateNewStamps (String[] prevStamps, boolean wasUpdate)
			throws ParseException
	{
		DateFormat dateFormat = DateFormat.getInstance();
		Date prevUpdate = dateFormat.parse(prevStamps[0]);
		long prevUpdateMili = prevUpdate.getTime();
		Date thisUpdate = dateFormat.parse(prevStamps[1]);
		long thisUpdateMili = thisUpdate.getTime();
		long diff = thisUpdateMili - prevUpdateMili;
		if (wasUpdate)
		{
			diff /= 2;
		}
		else
		{
			diff *= 2;
		}
		long nextUpdateMili = thisUpdateMili + diff;
		String nextUpdate = dateFormat.format(new Date(nextUpdateMili));
		String[] result = { dateFormat.format(thisUpdate), nextUpdate };
		return result;
	}
}
