package logrus.ontmodel.merging;

import java.util.Hashtable;
import java.util.Iterator;

import logrus.ontmodel.Quad;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.restaurant.tools.IPropertyMatcher;
import logrus.ontmodel.validation.Plurality;
import logrus.ontmodel.validation.Significance;
import logrus.tools.ObjectInitializationException;
import logrus.tools.Utils;
import logrus.tools.queue.Queue;
import logrus.tools.queue.QueueNode;

public class ObjectMerger
{
	private static ObjectMerger instance;

	private ObjectMerger ()
	{
		super();
	}

	public static ObjectMerger getInstance ()
	{
		if (ObjectMerger.instance == null)
		{
			ObjectMerger.instance = new ObjectMerger();
		}
		return ObjectMerger.instance;
	}

	/**
	 * Merge two statemet descriptions lists using set of rules included in <code>rules</code>
	 * collection
	 * 
	 * @param obj1
	 *           Statemet descriptions list
	 * @param obj2
	 *           Statemet descriptions list
	 * @param rules
	 *           Rules used for merging <code>obj1</code> and <code>obj2</code>
	 */
	public MergingReport merge (QuadsList obj1, QuadsList obj2,
			MergingRulesCollection rules)
	{
		// clean old statements
		String providerNew = obj2.getProviderID();
		obj1.removeStatmentsWithProviderID(providerNew);
		
		MergingReport report = new MergingReport();
		// map statements into groups with the same predicate
		Hashtable map1 = this.buildHashtable(obj1,rules);
		Hashtable map2 = this.buildHashtable(obj2,rules);
		QuadsList result = new QuadsList();

		for (Iterator iter = rules.getList().iterator(); iter.hasNext();)
		{
			// prepare data
			MergingRule rule = (MergingRule)iter.next();
			String property = rule.getProperty();
			Queue props1 = (Queue)map1.get(property);
			Queue props2 = (Queue)map2.get(property);
			Queue stmts = this.listFromHashtables(map1,map2,property);

			// check if whole process should be performed by outer class
			if (this.OnJavaClassRule(stmts,report,rule))
				continue;

			// check if essential data is present
			if (!this.checkEssential(props1,report,rule)
					|| !this.checkEssential(props1,report,rule))
			{
				continue;
			}

			// check if exclusive property should be applied and apply if appropriate
			if (this.OnExclusiveProvider(stmts,report,rule))
				continue;

			// check if matching property values is required
			if (this.OnRequireMatch(stmts,report,rule) == false)
				continue;

			// merge property lists if it's required
			if (this.OnMergeMultiple(stmts,report,rule))
				continue;

			this.OnLowSignificanceData(stmts,report,rule);

			this.OnProviderSignificance(stmts,report,rule);

			report.addStatements(stmts);
		}
		return report;
	}

	private boolean OnProviderSignificance (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getProviderSignificanceOrder() != null)
		{
			Queue toRemove = new Queue();
			QueueNode nodeProv = rule.getProviderSignificanceOrder().getTail();
			int iNodes = 0;
			// check providers starting form the least significant
			while (nodeProv != null)
			{
				iNodes = 0;
				toRemove.clear();
				String currProv = (String)nodeProv.getKey();
				// extract statements that potentially may be removed
				for (Iterator iter = stmts.iterator(); iter.hasNext();)
				{
					Quad elem = (Quad)iter.next();
					if (elem.getProvider().compareTo(currProv) == 0)
					{
						++iNodes;
						toRemove.inject(elem);
					}
				}
				// check if after removing statements with lower priority
				// we will still have some statemnts left in the queue
				if (stmts.count() - iNodes > 0)
				{
					for (Iterator delIter = toRemove.iterator(); delIter.hasNext();)
					{
						stmts.remove((Quad)delIter.next());
					}
				}
				nodeProv = nodeProv.getPrev();
			}
			return true;
		}
		return false;
	}

	private boolean OnLowSignificanceData (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getLowSignificanceData() != null)
		{
			Queue lsDataQueue = rule.getLowSignificanceData();
			Queue toRemove = new Queue();
			for (Iterator iter = lsDataQueue.iterator(); iter.hasNext();)
			{
				String lsData = (String)iter.next();
				for (Iterator iterStmt = stmts.iterator(); iterStmt.hasNext();)
				{
					Quad elem = (Quad)iterStmt.next();
					if (elem.getObject().compareTo(lsData) == 0)
					{
						toRemove.inject(elem);
					}
				}
				if (stmts.count() > toRemove.count())
				{
					for (Iterator delIter = toRemove.iterator(); delIter.hasNext();)
					{
						stmts.remove((Quad)delIter.next());
					}
				}
				toRemove.clear();
			}
			return true;
		}
		return false;
	}

	private boolean OnMergeMultiple (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.isMergeMultiple())
		{
			Queue clean = new Queue();
			Queue order = rule.getProviderSignificanceOrder();
			for (Iterator iter = stmts.iterator(); iter.hasNext();)
			{
				Quad elem = (Quad)iter.next();
				Quad inQueue = this.getStmtWithValue(clean,elem.getObject());
				if (inQueue != null)
				{
					// provider ignificance decides which statement should left in queue
					int elemProvRank = rule.getProviderRank(elem.getProvider());
					int inQueueProvRank = rule
							.getProviderRank(inQueue.getProvider());

					// now check if inQueue should be replaced by elem
					if (elemProvRank != -1 && elemProvRank < inQueueProvRank)
					{
						clean.replace(inQueue,elem);
					}
				}
				else
				{
					clean.inject(elem);
				}
			}

			report.addStatements(clean);
			return true;
		}
		return false;
	}

	private boolean OnRequireMatch (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.isRequireMatch())
		{
			boolean useMatcher = false;
			IPropertyMatcher matcher = null;
			// if matching should be performed by specialized object then int
			if (rule.getMatcherClassName() != null)
			{
				try
				{
					matcher = (IPropertyMatcher)Utils.getInstance().constructObject(
							rule.getMatcherClassName(),new Class[] { Boolean.class },
							new Object[] { new Boolean(true) });
				}
				catch (ObjectInitializationException e)
				{
					report.addError(rule.getProperty(),
							MergingError.JavaClassInitializationFailed);
					return false;
				}
				useMatcher = true;
			}
			// now check elements matching
			QueueNode node = stmts.getHead();
			QueueNode nextNode = null;
			Quad quad1 = null, quad2 = null;
			while (node.getNext() != null)
			{
				nextNode = node.getNext();
				quad1 = (Quad)node.getKey();
				quad2 = (Quad)nextNode.getKey();
				if (useMatcher)
				{
					// advanced matching - performed by specialized object
					if (!matcher.match(quad1.getObject(),quad2.getObject()))
					{
						report.addError(rule.getProperty(),
								MergingError.RequiredMatchFailed);
						return false;
					}
				}
				else
				{
					// simple mathcing - use direct comparison
					if (!quad1.isRDFStatementEqual(quad2))
					{
						report.addError(rule.getProperty(),
								MergingError.RequiredMatchFailed);
						return false;
					}
				}
				node = nextNode;
			}
			return true;
		}
		return true;
	}

	private boolean OnJavaClassRule (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getJavaClassRule() != null)
		{
			try
			{
				ICustomMerging merger = (ICustomMerging)Utils.getInstance()
						.constructObject(
								rule.getJavaClassRule(),
								new Class[] { stmts.getClass(), report.getClass(),
										rule.getClass() },
								new Object[] { stmts, report, rule });
				merger.merge(stmts,report,rule);
				// 
			}
			catch (ObjectInitializationException e)
			{
				report.addError(rule.getProperty(),
						MergingError.JavaClassInitializationFailed);
			}
			return true;
		}
		return false;
	}

	private boolean OnExclusiveProvider (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getExclusiveProvider() != null)
		{
			// get list of properties with only
			Queue cleaned = this.useExclusiveProvider(rule.getExclusiveProvider(),
					stmts);
			if (this.checkPlurality(stmts,report,rule))
			{
				report.addStatements(cleaned);
			}
			return true;
		}
		else
		{
			return false;
		}
	}

	// Check if essential property is present
	private boolean checkEssential (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getSignificance() == Significance.Essential
				&& stmts.count() == 0)
		{
			report.addError(rule.getProperty(),MergingError.LackingEssential);
			return false;
		}
		return true;
	}

	// Check if property singluarity is held
	private boolean checkPlurality (Queue stmts, MergingReport report,
			MergingRule rule)
	{
		if (rule.getPlurality() == Plurality.Singular && stmts.count() > 1)
		{
			report
					.addError(rule.getProperty(),MergingError.UnexpectedMultiplicity);
			return false;
		}
		return true;
	}

	/**
	 * Create new queue from <code>stmts</code> leaving only statements form datasource identified
	 * by <code>uriProvider</code>
	 * 
	 * @param uriProvider
	 *           Exclusive provider URI
	 * @param stmts
	 *           Collection of statement descriptions
	 * @return Queue containing only statements provided by <code>uriProvider</code>
	 */
	private Queue useExclusiveProvider (String uriProvider, Queue stmts)
	{
		Queue cleaned = new Queue();
		for (Iterator iter = stmts.iterator(); iter.hasNext();)
		{
			Quad elem = (Quad)iter.next();
			if (uriProvider.compareTo(elem.getProvider()) == 0)
			{
				cleaned.inject(elem);
			}
		}
		return stmts;
	}

	/**
	 * Create one list of statements containing only statements with <code>uriProperty</code> as
	 * predicate, merge lists obtained from hashmaps <code>map1</code> and <code>map2</code>
	 * 
	 * @param map1
	 *           Hashmap containing statements grouped by predicate
	 * @param map2
	 *           Hashmap containing statements grouped by predicate
	 * @param uriProperty
	 *           Predicate should match <code>uriProperty</code>
	 * @return One list of statements containing only statements with <code>uriProperty</code> as
	 *         predicate
	 */
	private Queue listFromHashtables (Hashtable map1, Hashtable map2,
			String uriProperty)
	{
		Queue stmts = new Queue();
		Object obj1 = map1.get(uriProperty);
		if (obj1 != null)
		{
			stmts.addAll((Queue)obj1);
		}

		try
		{
			Object obj2 = map2.get(uriProperty);
			if (obj2 != null)
			{
				Queue quadsList = (Queue)obj2;
				for (Iterator iter = quadsList.iterator(); iter.hasNext();)
				{
					Object obj = iter.next();
					if (obj != null && !this.checkElem((Quad)obj,stmts))
					{
						stmts.add((Quad)obj);
					}
				}
			}
		}
		catch (NullPointerException exc)
		{
			System.out.println("Property: " + uriProperty);
		}
		return stmts;
	}

	/**
	 * Check if statement representing the same information as <code>quad</code> is already
	 * included in <code>queue</code>
	 * 
	 * @param quad
	 * @param queue
	 * @return
	 */
	private boolean checkElem (Quad quad, Queue queue)
	{
		for (Iterator iter = queue.iterator(); iter.hasNext();)
		{
			Object obj = iter.next();

			if (obj != null && quad != null && quad.isEqual((Quad)obj))
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Check if statment describing the same quad as <code>stmt</code> is already included in
	 * <code>list</code>
	 * 
	 * @param list
	 *           List of statement descriptions
	 * @param stmt
	 *           Searched statement
	 * @return True if statment describing the same quad as <code>stmt</code> is already included
	 *         in <code>list</code>, otherwise returns false
	 */
	private boolean containsStmt (Queue list, Quad stmt)
	{
		for (Iterator iter = list.iterator(); iter.hasNext();)
		{
			if (stmt.isEqual((Quad)iter.next()))
			{
				return true;
			}
		}
		return false;
	}

	private Quad getStmtWithValue (Queue list, String value)
	{
		for (Iterator iter = list.iterator(); iter.hasNext();)
		{
			Quad stmt = (Quad)iter.next();
			if (stmt.getObject().compareTo(value) == 0)
			{
				return stmt;
			}
		}
		return null;
	}

	private Quad getStmt (Queue list, Quad stmt)
	{
		for (Iterator iter = list.iterator(); iter.hasNext();)
		{
			Quad elem = (Quad)iter.next();
			if (elem.isEqual(stmt))
			{
				return elem;
			}
		}
		return null;
	}

	/**
	 * Calculate auxiliary hashtable mapping statements into groups with the same predicate
	 * 
	 * @param list
	 *           List of statement descriptions
	 * @return Hashtable mapping statements into groups with the same predicate
	 */
	private Hashtable buildHashtable (QuadsList list,
			MergingRulesCollection rules)
	{
		// create one queue in hashtable for each rules (ontological property)
		Hashtable map = new Hashtable();
		for (Iterator iter = rules.getList().iterator(); iter.hasNext();)
		{
			MergingRule rule = (MergingRule)iter.next();
			map.put(rule.getProperty(),new Queue());
		}

		// browse all statements
		for (Iterator iter = list.getList().iterator(); iter.hasNext();)
		{
			Quad stmt = (Quad)iter.next();
			Object obj = map.get(stmt.getPredicate());
			if (obj == null)
			{
				Queue props = new Queue();
				props.inject(stmt);
				map.put(stmt.getPredicate(),props);
			}
			else
			{
				Queue props = (Queue)obj;
				props.inject(stmt);
			}
		}
		return map;
	}
}
