package logrus.ontmodel.merging;

import java.util.HashMap;

import logrus.ontmodel.IObjectInfoComparer;
import logrus.ontmodel.ObjectInfoComparerFactory;
import logrus.ontmodel.Prefixes;
import logrus.ontmodel.Quad;
import logrus.ontmodel.QuadsList;
import logrus.tools.Utils;
import logrus.tools.XMLAnalysisException;

/**
 * 
 */
public class Unificator
{
	private static Unificator instance;

	private ObjectMerger merger;

	private HashMap mapRulesClasses;

	/**
	 * @throws XMLAnalysisException
	 */
	private Unificator () throws XMLAnalysisException
	{
		super();
		this.merger = ObjectMerger.getInstance();
		String fileRestRules = Utils.getInstance().getCWD()
				+ "xml\\RestaurantMerging.xml";
		MergingRulesCollection restRules = new MergingRulesCollection(
				fileRestRules);
		this.mapRulesClasses = new HashMap();
		this.mapRulesClasses.put(Prefixes.CLASS_RES,restRules);
	}

	public static Unificator getInstance () throws XMLAnalysisException
	{
		if (Unificator.instance == null)
		{
			Unificator.instance = new Unificator();
		}
		return Unificator.instance;
	}

	/**
	 * Check if ontological classes of objects represented by <code>objRef</code> and
	 * <code>objIncl</code> are identical, if not throw exception
	 * 
	 * @param objRef
	 *           Statements describing some object
	 * @param objIncl
	 *           Statements describing some object
	 * @return Ontological class common for both objects
	 * @throws UnionException
	 *            Thrown when ontological classes of objects represented by <code>objRef</code> and
	 *            are not identical
	 */
	private String requireOntClassesCompatibility (QuadsList objRef,
			QuadsList objIncl) throws UnionException
	{
		String ontClassRef = objRef.getOntClass();
		String ontClassIncl = objIncl.getOntClass();

		if (ontClassRef == null || ontClassIncl == null)
		{
			throw new UnionException(
					"Can't determine at least one ontological class");
		}
		if (ontClassRef.compareTo(ontClassIncl)!=0)
		{
			throw new UnionException(
					"Can't merge objects of different ontological classes "
							+ ontClassRef + " and " + ontClassIncl);
		}
		return ontClassRef;
	}

	private boolean isRepresentingSame (QuadsList objRef, QuadsList objIncl,
			String ontClass) throws UnionException
	{
		IObjectInfoComparer comparer = ObjectInfoComparerFactory.getInstance()
				.createObjectInfoComparer(ontClass);
		if (comparer == null)
		{
			throw new UnionException("Comparer for class " + ontClass
					+ " not definied");
		}
		return comparer.isRepresentingSame(objRef,objIncl);
	}

	/**
	 * Perform unification process of <code>objRef</code> and <code>objIncl</code>
	 * 
	 * @param objRef
	 *           Object description
	 * @param objIncl
	 *           Object description
	 * @return Unified object description or null if
	 *         <ul>
	 *         <li>Object descriptions do not represent the same logical object</li>
	 *         </ul>
	 * @throws UnionException
	 */
	public MergingReport union (QuadsList objRef, QuadsList objIncl)
			throws UnionException
	{
		// prerequisite 1: both object must be represent instances of the same ontological class
		String ontClass = this.requireOntClassesCompatibility(objRef,objIncl);

		// if both objects do not describe the same object return null;
		if (!this.isRepresentingSame(objRef,objIncl,ontClass))
		{
			return null;
		}

		// substitute restaurant ID in objIncl for the one taken from objRef
		objIncl.substituteSubject(objRef.getObjectURI());
		
		// perform update of changed statements from provider in objIncl
		

		// get the set of merging rules
		Object objSetRules = this.mapRulesClasses.get(ontClass);
		if (objSetRules == null)
		{
			throw new UnionException(
					"No set of merging rules defined for ontological class"
							+ ontClass);
		}
		
		// we've got two object descriptions representing the same object in real world
		// so we have to merge them
		MergingReport report = this.merger.merge(objRef,objIncl,
				(MergingRulesCollection)objSetRules);
		
		//if (report.isCorrectMerge())
		//{
		//	System.out.println(report.getResult());			
		//}

		String objID = objIncl.getObjectURI();
		String objOntClass = objIncl.getOntClass();
		String objMainProv = ((Quad)objIncl.getList().get(0)).getProvider();
		report.getResult().add(objID,Prefixes.RDF_TYPE,objOntClass,objMainProv);

		return report;
	}
}
