package logrus.ontmodel.rdql;

import logrus.tools.Utils;

public class QueryBuilder
{
	// Utils singleton reference
	private Utils utils;

	// buffer used for building query
	private StringBuffer buff;

	public QueryBuilder ()
	{
		super();
		this.utils = Utils.getInstance();
		this.buff = new StringBuffer();
	}

	/**
	 * Append line to query buffer with tabulation indentention
	 * 
	 * @param str
	 */
	private void appendIndentLine (String str)
	{
		this.utils.appendIndentLine(this.buff,str);
	}

	/**
	 * Appen line describing criterion with literal value
	 * 
	 * @param subject
	 *           "Subject" of statement pattern
	 * @param predicate
	 *           "Predicate" of statement pattern
	 * @param object
	 *           "Object" of statement pattern
	 */
	private void appendCriterion (String subject, String predicate, String object)
	{
		this.appendIndentLine("(" + subject + " " + predicate + " " + object
				+ ")");
	}

	/**
	 * Appen line describing criterion with literal value
	 * 
	 * @param subject
	 *           "Subject" of statement pattern
	 * @param predicate
	 *           "Predicate" of statement pattern
	 * @param literal
	 *           "Literal" of statement pattern
	 */
	private void appendLiteralCriterion (String subject, String predicate,
			String literal)
	{
		this.appendCriterion(subject,predicate,"\"" + literal + "\"");
	}

	/**
	 * Append cuisine criterion
	 * 
	 * @param cuisine
	 *           Cuisines that restaurant should have
	 */
	private void appendCuisine (String[] cuisines)
	{
		if (cuisines != null)
		{
			for (int i = 0; i < cuisines.length; i++)
			{
				this.appendLiteralCriterion("?id","res:cuisine",cuisines[i]);
			}
		}
	}

	/**
	 * Append accepted methods of payment criterion
	 * 
	 * @param payments
	 *           Accepted methods of payment
	 */
	private void appendAcceptedPayment (String[] payments)
	{
		if (payments != null)
		{
			for (int i = 0; i < payments.length; i++)
			{
				this.appendLiteralCriterion("?id","res:accepts",payments[i]);
			}
		}
	}

	/**
	 * Append accessibility for handicapped criterion
	 * 
	 * @param accessibility
	 *           Accessibility for handicapped criterion
	 */
	private void appendAccessibility (String accessibility)
	{
		if (accessibility != null)
		{
			this.appendLiteralCriterion("?id","res:accessibility",accessibility);
		}
	}

	/**
	 * Append alcohol service criterion
	 * 
	 * @param alcohol
	 *           Alcohol service criterion
	 */
	private void appendAlcohol (String alcohol)
	{
		if (alcohol != null)
		{
			this.appendLiteralCriterion("?id","res:alcohol",alcohol);
		}
	}

	/**
	 * Append required dress criterion
	 * 
	 * @param dress
	 *           Required dress criterion
	 */
	private void appendDress (String dress)
	{
		if (dress != null)
		{
			this.appendLiteralCriterion("?id","res:dress",dress);
		}
	}

	/**
	 * Append restaurant's features criterion
	 * 
	 * @param features
	 *           Restaurant's features of payment
	 */
	private void appendFeatures (String[] features)
	{
		if (features != null)
		{
			for (int i = 0; i < features.length; i++)
			{
				this.appendLiteralCriterion("?id","res:features",features[i]);
			}
		}
	}

	/**
	 * Append parking criterion
	 * 
	 * @param parking
	 *           Parking criterion
	 */
	private void appendParking (String parking)
	{
		if (parking != null)
		{
			this.appendLiteralCriterion("?id","res:parking",parking);
		}
	}

	/**
	 * Build RDQL query with specifed criterions
	 * 
	 * @param cuisines
	 *           Cuisines that restaurant should have
	 * @param payments
	 *           Accepted methods of payment
	 * @param accessibility
	 *           Accessibility for handicapped criterion
	 * @param alcohol
	 *           Alcohol service criterion
	 * @param dress
	 *           Required dress criterion
	 * @param features
	 *           Restaurant's features criterion
	 * @param parking
	 *           Parking criterion
	 * @return RDQL query with specifed criterions
	 */
	public String buildQuery (String[] cuisines, String[] payments,
			String accessibility, String alcohol, String dress, String[] features,
			String parking)
	{
		// append variables
		this.utils.appendLine(this.buff,"SELECT ?name, ?address, ?zip, ?city");

		// append WHERE clause
		this.utils.appendLine(this.buff,"WHERE");

		this.appendCuisine(cuisines);
		this.appendAcceptedPayment(payments);
		this.appendAccessibility(accessibility);
		this.appendAlcohol(alcohol);
		this.appendDress(dress);
		this.appendFeatures(features);
		this.appendParking(parking);

		// append prefixes
		this.utils.appendLine(this.buff,
				"USING res FOR <http://www.agentlab.net/schemas/restaurant#>");

		return this.buff.toString();
	}
}
