package logrus.ontmodel.restaurant.tools;

import jade.util.leap.Iterator;
import logrus.ontmodel.IObjectInfoTag;
import logrus.ontmodel.Prefixes;
import logrus.ontmodel.Quad;
import logrus.ontmodel.QuadsList;

/**
 * Class representing restaurant's "tag" - basic information about restaurant used for identifing it
 * (name and location - city, zip code and street address)
 */
public class RestaurantTag implements IObjectInfoTag
{
	private Quad name = null;

	private Quad city = null;

	private Quad zipCode = null;

	private Quad streetAddress = null;

	public RestaurantTag ()
	{
		super();
	}

	public RestaurantTag (Quad name, Quad city, Quad zipCode, Quad streetAddress)
	{
		this.name = name;
		this.city = city;
		this.zipCode = zipCode;
		this.streetAddress = streetAddress;
	}
	
	public String toString ()
	{
		StringBuffer buffer = new StringBuffer();
		if (this.hasName())
		{
			buffer.append("Name: [");
			buffer.append(this.getName());
			buffer.append("]\n");
		}
		if (this.hasStreetAddress())
		{
			buffer.append("Street address: [");
			buffer.append(this.getStreetAddress());
			buffer.append("]\n");
		}
		if (this.hasZipCode())
		{
			buffer.append("Zip code: [");
			buffer.append(this.getZipCode());
			buffer.append("]\n");
		}				
		if (this.hasCity())
		{
			buffer.append("City: [");
			buffer.append(this.getCity());
			buffer.append("]\n");
		}		
		return buffer.toString();
	}

	/**
	 * Check if this tag is similar enough to <code>tag</code> to be reckoned as representing the
	 * same restaurant
	 * 
	 * @param tag
	 *           Restaurant's tag to be compared
	 * @return True if we assume that <code>tag</code> represents the same restaurant, otherwise
	 *         returns false
	 */
	public boolean isLike (RestaurantTag tag)
	{
		RestaurantComparer comparer = new RestaurantComparer();
		return comparer.isRepresentingSame(this,tag);
	}

	/**
	 * Assume that <code>quads</code> describes restaurant and attempt to extract restaurant tag
	 * from it
	 * 
	 * @param quads
	 *           Collection of quads describing some object (this method assumes that describes some
	 *           restaurant)
	 * @return Restaurant's tag extracted from <code>quads</code>
	 */
	public IObjectInfoTag getTag (QuadsList quads)
	{
		RestaurantTag tag = new RestaurantTag();
		int counter = 0;
		boolean wasName = false, wasCity = false, wasZipCode = false, wasStreetAddress = false;
		for (Iterator iter = quads.getQuadsCollection().iterator(); iter
				.hasNext();)
		{
			Quad quad = (Quad)iter.next();
			if (!wasName && quad.getPredicate().compareTo(Prefixes.RES_NAME) == 0)
			{
				tag.setName(quad);
				wasName = true;
				counter++;
				continue;
			}
			if (!wasCity && quad.getPredicate().compareTo(Prefixes.LOC_CITY) == 0)
			{
				tag.setCity(quad);
				wasCity = true;
				counter++;
				continue;
			}
			if (!wasStreetAddress
					&& quad.getPredicate().compareTo(Prefixes.LOC_STREET_ADDR) == 0)
			{
				tag.setStreetAddress(quad);
				wasStreetAddress = true;
				counter++;
				continue;
			}
			if (!wasZipCode
					&& quad.getPredicate().compareTo(Prefixes.LOC_ZIP) == 0)
			{
				tag.setZipCode(quad);
				wasZipCode = true;
				counter++;
				continue;
			}
			if (counter == 4)
			{
				break;
			}
		}
		return tag;
	}

	/**
	 * Check if this tag is complete (all fields set)
	 * 
	 * @return True if restaurant's tag is complete, otherwise returns false.
	 */
	public boolean isComplete ()
	{
		return this.name != null && this.city != null && this.zipCode != null
				&& this.streetAddress != null;
	}

	/**
	 * Check if this restaurant has some city specified
	 * 
	 * @return True if it has some city specifed, otherwise returns false
	 */
	public boolean hasCity ()
	{
		return this.city != null;
	}
	
	/**
	 * Check if this restaurant has some name specified
	 * 
	 * @return True if it has some name specifed, otherwise returns false
	 */
	public boolean hasName ()
	{
		return this.name != null;
	}

	/**
	 * Check if this restaurant has some street address specified
	 * 
	 * @return True if it has some street address specifed, otherwise returns false
	 */
	public boolean hasStreetAddress ()
	{
		return this.streetAddress != null;
	}
	
	/**
	 * Check if this restaurant has some zip code specified
	 * 
	 * @return True if it has some zip code specifed, otherwise returns false
	 */
	public boolean hasZipCode ()
	{
		return this.zipCode != null;
	}
	
	/**
	 * @param name
	 *           The name to set.
	 */
	public void setName (Quad name)
	{
		this.name = name;
	}
	
	/**
	 * @param name
	 *           The name to set.
	 */
	public void setNameValue (String name)
	{
		this.name.setObject(name);
	}	

	/**
	 * @return Returns the name.
	 */
	public String getName ()
	{
		return this.name.getObject();
	}

	/**
	 * @param city
	 *           The city to set.
	 */
	public void setCity (Quad city)
	{
		this.city = city;
	}
	
	/**
	 * @param city
	 *           The name to set.
	 */
	public void setCityValue (String city)
	{
		this.city.setObject(city);
	}		

	/**
	 * @return Returns the city.
	 */
	public String getCity ()
	{
		return this.city.getObject();
	}

	/**
	 * @param zipCode
	 *           The zipCode to set.
	 */
	public void setZipCode (Quad zipCode)
	{
		this.zipCode = zipCode;
	}
	
	/**
	 * @param zipCode
	 *           The zipCode to set.
	 */
	public void setZipCodeValue (String zipCode)
	{
		this.zipCode.setObject(zipCode);
	}
	

	/**
	 * @return Returns the zipCode.
	 */
	public String getZipCode ()
	{
		return this.zipCode.getObject();
	}

	/**
	 * @param streetAddress
	 *           The streetAddress to set.
	 */
	public void setStreetAddress (Quad streetAddress)
	{
		this.streetAddress = streetAddress;
	}
	
	/**
	 * @param streetAddress
	 *           The street address to set.
	 */
	public void setStreetAddressValue (String streetAddress)
	{
		this.streetAddress.setObject(streetAddress);
	}	

	/**
	 * @return Returns the streetAddress.
	 */
	public String getStreetAddress ()
	{
		return this.streetAddress.getObject();
	}
}
