package logrus.ontmodel.validation;

import java.util.HashMap;
import java.util.Iterator;

import logrus.ontmodel.Prefixes;
import logrus.ontmodel.Quad;
import logrus.ontmodel.QuadsList;
import logrus.tools.Utils;
import logrus.tools.XMLAnalysisException;

/**
 * Performs validation of ontological objects
 */
public class ObjectValidator
{
	// singleton instance
	private static ObjectValidator instance;

	// hashmap holding mappings ontological class -> set of validation rules
	private HashMap mapRules;

	/**
	 * Constructor kept private, following singleton design pattern
	 * 
	 * @throws XMLAnalysisException
	 *            Thrown when error during loading rules sets from files occures
	 */
	private ObjectValidator () throws XMLAnalysisException
	{
		super();
		this.mapRules = new HashMap();
		String path = Utils.getInstance().getCWD()
				+ "\\xml\\RestaurantValidation.xml";
		ValidationRulesCollection resRules = new ValidationRulesCollection(path);
		this.mapRules.put(Prefixes.CLASS_RES,resRules);
	}

	/**
	 * Get singleton instance of <code>ObjectValidator</code>
	 * 
	 * @return Singleton instance of <code>ObjectValidator</code>
	 * @throws XMLAnalysisException
	 *            Thrown when error during loading rules sets from files occures
	 */
	public static ObjectValidator getInstance () throws XMLAnalysisException
	{
		if (ObjectValidator.instance == null)
		{
			ObjectValidator.instance = new ObjectValidator();
		}
		return ObjectValidator.instance;
	}

	/**
	 * Get set of validation rules for ontological class <code>ontClass</code>
	 * 
	 * @param ontClass
	 *           Ontological class
	 * @return Set of validation rules for ontological class <code>ontClass</code>
	 */
	public ValidationRulesCollection getRulesForOntClass (String ontClass)
	{
		Object obj = this.mapRules.get(ontClass);
		if (obj == null)
		{
			return null;
		}
		else
		{
			return (ValidationRulesCollection)obj;
		}
	}

	/**
	 * Check if <code>description</code> is valid code
	 * 
	 * @param description
	 *           Object description that should be validated
	 * @return Validation status code (one of values defined in <code>ValidationStatus</code>)
	 */
	public int validate (QuadsList description)
	{
		return this.validate(description,this.getRulesForOntClass(description
				.getOntClass()));
	}

	/**
	 * Check if <code>description</code> meets validation rules included in <code>rules</code>
	 * collection
	 * 
	 * @param description
	 *           Object description that should be validated
	 * @param rules
	 *           Set of validation rules for ontological class ther <code>descriptions</code>
	 *           represents
	 * @return Validation status code (one of values defined in <code>ValidationStatus</code>)
	 */
	public int validate (QuadsList description, ValidationRulesCollection rules)
	{
		int valState = ValidationStatus.Complete;
		for (Iterator iter = rules.getList().iterator(); iter.hasNext();)
		{
			ValidationRule rule = (ValidationRule)iter.next();
			int count = description.countProperties(rule.getProperty());
			// check if not multiple property values if singular plurality required
			if (rule.getPlurality() == Plurality.Singular && count > 1)
			{
				return ValidationStatus.Incorrect;
			}
			// check if essential statement is not missing
			if (rule.getSignificance() == Significance.Essential && count == 0)
			{
				return ValidationStatus.Incorrect;
			}
			else
			{
				for (jade.util.leap.Iterator iterStmts = description.getList()
						.iterator(); iterStmts.hasNext();)
				{
					Quad quad = (Quad)iterStmts.next();
					if (quad.getObject()==null || quad.getObject().length()==0)
					{
						return ValidationStatus.Incorrect;
					}
				}
			}

			// check if required statement is not misssing
			if (rule.getSignificance() == Significance.Required && count == 0)
			{
				valState = ValidationStatus.Incomplete;
			}
			else
			{
				for (jade.util.leap.Iterator iterStmts = description.getList()
						.iterator(); iterStmts.hasNext();)
				{
					Quad quad = (Quad)iterStmts.next();
					if (quad.getObject()==null || quad.getObject().length()==0)
					{
						return ValidationStatus.Incomplete;
					}
				}
			}			
		}
		return valState;
	}
}
