package logrus.tools;

import jade.core.AID;
import jade.domain.FIPAException;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import logrus.agents.AgentNames;
import logrus.agents.OntoStorageAgent;
import logrus.agents.ServiceNames;
import logrus.agents.databaseAgent.DatabaseAgent;
import logrus.ontmodel.Prefixes;
import logrus.tools.queue.Queue;
import logrus.tools.settings.SettingsManager;
import oracle.xml.parser.v2.XMLDOMImplementation;
import oracle.xml.parser.v2.XMLLSParser;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSParser;
import org.w3c.dom.ls.LSSerializer;

/**
 * Various useful functions. This is a singleton.
 */
public class Utils
{
	private static Utils instance;

	/**
	 *  
	 */
	private Utils ()
	{
		super();
	}

	/**
	 * Get singleton instance of Utils
	 * 
	 * @return Singleton instance of Utils
	 */
	public static Utils getInstance ()
	{
		if (Utils.instance == null)
		{
			Utils.instance = new Utils();
		}
		return Utils.instance;
	}

	/**
	 * Construct service name of DBA agent performing service <code>mode</code>.
	 * 
	 * @param mode
	 *           Mode in which agent is running (<code>DatabaseAgent.MODE_WRITE</code> or
	 *           <code>DatabaseAgent.MODE_READ</code>)
	 * @return Service name of DBA agent
	 */
	public String getDBAServiceName (int mode)
	{
		String name = "";
		switch (mode)
		{
			case DatabaseAgent.MODE_WRITE:
				name = ServiceNames.ServiceDBA + "_WRITE";
				break;
			case DatabaseAgent.MODE_READ:
				name = ServiceNames.ServiceDBA + "_READ";
				break;
		}
		return name;
	}

	/**
	 * Construct name of DBA agent performing service <code>mode</code>.
	 * 
	 * @param mode
	 *           Mode in which agent is running (<code>DatabaseAgent.MODE_WRITE</code> or
	 *           <code>DatabaseAgent.MODE_READ</code>)
	 * @return Name of DBA agent
	 */
	public String getDBAName (int mode)
	{
		String name = "";
		switch (mode)
		{
			case DatabaseAgent.MODE_WRITE:
				name = AgentNames.DatabaseAgent + "_WRITE";
				break;
			case DatabaseAgent.MODE_READ:
				name = AgentNames.DatabaseAgent + "_READ_";
				break;
		}
		return name;
	}

	public AID tossDBARead (OntoStorageAgent osAgent) throws FIPAException
	{
		String servName = Utils.getInstance().getDBAServiceName(
				DatabaseAgent.MODE_READ);
		AID[] agents = osAgent.findAgentsPerformingService(servName);
		int num = (int)Math.round(Math.random() * (agents.length - 1));
		return agents[num];
	}

	/**
	 * Show exception information
	 * 
	 * @param exc
	 *           Exception to be reported
	 * @param parent
	 *           Shell used as parent by message box
	 */
	public void showExceptionInfo (Exception exc, Shell parent)
	{
		MessageBox box = new MessageBox(parent,SWT.ICON_ERROR | SWT.OK);
		box.setText("Error: exception thrown");
		box.setMessage("Error: " + exc.getLocalizedMessage() + "\n\n"
				+ "Stack trace: " + exc.getStackTrace());
		box.open();
	}

	/**
	 * Show message <code>msg</code> (informational message) in message box
	 * 
	 * @param msg
	 *           Message to be presented
	 * @param parent
	 *           Shell used as parent by message box
	 */
	public void showInfoMessage (String msg, Shell parent)
	{
		MessageBox box = new MessageBox(parent,SWT.OK | SWT.ICON_INFORMATION);
		box.setText("Information");
		box.setMessage(msg);
		box.open();
	}

	/**
	 * Show message <code>msg</code> (error message) in message box
	 * 
	 * @param msg
	 *           Message to be presented
	 * @param parent
	 *           Shell used as parent by message box
	 */
	public void showErrorMessage (String msg, Shell parent)
	{
		MessageBox box = new MessageBox(parent,SWT.OK | SWT.ICON_ERROR);
		box.setText("Error");
		box.setMessage(msg);
		box.open();
	}

	/**
	 * Create string "file://<code>dirpath</code>/<code>name</code>.<code>extension</code>"
	 * 
	 * @param dirpath
	 *           Path of directory in which file exists
	 * @param name
	 *           Name of file
	 * @param extension
	 *           Extension of file
	 * @return String <code>"file://dirpath/name.extension"<code>
	 */
	public String createFileURI (String dirpath, String name, String extension)
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("file://");
		buffer.append(dirpath);
		if (dirpath.charAt(dirpath.length() - 1) != '/')
		{
			buffer.append('/');
		}
		buffer.append(name);
		buffer.append('.');
		buffer.append(extension);
		return buffer.toString();
	}

	/**
	 * Create empty typed XML document
	 * 
	 * @param doctypeName
	 *           Document type name
	 * @param dtdURI
	 *           URI of DTD used for validating document
	 * @return Empty typed XML document
	 */
	public Document createTypedXMLDocument (String doctypeName, String dtdURI)
	{
		// XMLDOMImplementation class implements the DOMImplementationLS interface,
		// and has methods to create an XML document and doctype.
		XMLDOMImplementation implDOM = new XMLDOMImplementation();

		// create a doctype for a document
		DocumentType doctype = implDOM
				.createDocumentType(doctypeName,null,dtdURI);

		// create XML document with specified type
		return implDOM.createDocument(null,null,doctype);
	}

	/**
	 * Save XML document to a specified file
	 * 
	 * @param document
	 *           XML document to be saved
	 * @param filename
	 *           Name of file which should be used for saving XML document
	 * @throws FileNotFoundException
	 *            <code>filename</code> is not valid filesystem path
	 */
	public void saveXMLDocument (Document document, String filename)
			throws FileNotFoundException, IOException
	{
		XMLDOMImplementation implDOM = new XMLDOMImplementation();
		DOMImplementationLS implDOMls = implDOM;
		LSSerializer domWriter = implDOMls.createLSSerializer();
		LSOutput output = implDOMls.createLSOutput();
		OutputStream outputStream = new FileOutputStream(new File(filename));
		output.setByteStream(outputStream);
		output.setEncoding(SettingsManager.getSettings().getEncoding());
		domWriter.write(document,output);
		outputStream.close();
	}

	/**
	 * Load XML document from specified file
	 * 
	 * @param filename
	 *           Name of file from which XML document should be loaded
	 * @return Loaded XML document
	 */
	public Document loadXMLDocument (String filename)
	{
		DOMImplementationLS impl = new XMLDOMImplementation();
		LSParser parser = impl.createLSParser(
				DOMImplementationLS.MODE_SYNCHRONOUS,null);
		XMLLSParser lsParser = (XMLLSParser)parser;
		String uri = "file://" + filename.replace('\\','/');
		return lsParser.parseURI(uri);
	}

	/**
	 * Append string ".<code>extension<code>" to the string <code>filename</code> 
	 * if it doesn't end with it
	 * @param filename String to be (maybe) extended
	 * @param extension String that will be (maybe) added to <code>filename</code>
	 * @return String ending with <code>.extension<code>
	 */
	public String strAppendExtension (String filename, String extension)
	{
		if (!filename.endsWith(extension))
		{
			return filename + '.' + extension;
		}
		else
		{
			return filename;
		}
	}

	/**
	 * Normalize string (trim spaces and convert all characters to lower case)
	 * 
	 * @param str
	 *           String to be normalized
	 * @return Normalized string
	 */
	public static String strNormString (String str)
	{
		return str.trim().toLowerCase();
	}

	/**
	 * Normalize string <code>str</code> and check if is equal to <code>pattern</code>
	 * 
	 * @param str
	 * @param pattern
	 * @return
	 */
	public static boolean strIsEqual (String str, String pattern)
	{
		return Utils.strNormString(str).compareTo(Utils.strNormString(pattern)) == 0;
	}

	/**
	 * Convert char <code>c</code> to upper case
	 * 
	 * @param c
	 *           Character to be converted
	 * @return Converted character
	 */
	public static char charToUpperCase (char c)
	{
		return Character.toUpperCase(c);
	}

	/**
	 * Convert char <code>c</code> to lower case
	 * 
	 * @param c
	 *           Character to be converted
	 * @return Converted character
	 */
	public static char charToLowerCase (char c)
	{
		return Character.toLowerCase(c);
	}

	/**
	 * Change character in string <code>str</code> at position <code>index</code> to upper case
	 * 
	 * @param str
	 * @param index
	 * @return
	 */
	public static String strToUpperAt (String str, int index)
	{
		char[] upper = new char[1];
		upper[0] = Utils.charToUpperCase(str.charAt(0));
		if (Character.isLetter(upper[0]))
		{
			String strUpper = new String(upper);
			String subStr = str.substring(0,1);
			return str.replaceFirst(subStr,strUpper);
		}
		else
		{
			return str;
		}
	}

	/**
	 * Trim characters <code>c</code> at the beginning and at the end of the string
	 * 
	 * @param str
	 *           String to be trimmed
	 * @param c
	 *           Character to be removed form ends
	 * @return Trimmed string
	 */
	public String strTrim (String str, char c)
	{
		int idx1, idx2;
		int len = str.length();
		for (idx1 = 0; idx1 < len; idx1++)
		{
			if (str.charAt(idx1) != c)
			{
				break;
			}
		}
		for (idx2 = len - 1; idx2 > -1; idx2--)
		{
			if (str.charAt(idx2) != c)
			{
				break;
			}
		}
		if (idx1 >= idx2)
		{
			return "";
		}
		else
		{
			return str.substring(idx1,++idx2);
		}
	}

	/**
	 * Remove substring from <code>str</code> between positions <code>idxLeft</code> and
	 * <code>idxRight</code>
	 * 
	 * @param str
	 *           String from which substring will be removed
	 * @param idxLeft
	 *           Left index of substring to be removed
	 * @param idxRight
	 *           Right index of substring to be removed
	 * @return STring with removed substring
	 */
	public String removeSubstring (String str, int idxLeft, int idxRight)
	{
		if (str.length() == 0 || idxLeft < 0 || idxRight < 0
				|| idxLeft > idxRight)
		{
			return str;
		}
		else
		{
			int idxEnd = str.length() - 1;
			if (idxLeft == 0 && idxRight == idxEnd)
			{
				return "";
			}
			else if (idxRight == idxEnd)
			{
				return str.substring(0,idxLeft).trim();
			}
			else if (idxLeft == 0)
			{
				return str.substring(idxRight + 1,idxEnd + 1).trim();
			}
			else
			{
				return (str.substring(0,idxLeft) + " " + str.substring(
						idxRight + 1,idxEnd)).trim();
			}
		}
	}

	/**
	 * Checks if string describes Polish zip code
	 * 
	 * @param str
	 *           String to be checked
	 * @return String
	 */
	public boolean isPolishZipCode (String str)
	{
		if (str.length() == 6 && Character.isDigit(str.charAt(0))
				&& Character.isDigit(str.charAt(1)) && str.charAt(2) == '-'
				&& Character.isDigit(str.charAt(3))
				&& Character.isDigit(str.charAt(4))
				&& Character.isDigit(str.charAt(5)))
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	public boolean isSimpleName (String str)
	{
		return str.indexOf("/") < 0 && str.indexOf(";") < 0
				&& str.indexOf(",") < 0 && str.indexOf(".") < 0;
	}

	/**
	 * Extract value of attribute <code>attr</code> from XML node <code>node</code>
	 * 
	 * @param node
	 *           XML node from which attribute value should be extracted
	 * @param attr
	 *           Attribute name
	 * @return Attribute value
	 * @throws XMLAnalysisException
	 *            Thrown when node has no attribute named <code>attr</code>
	 */
	public String extractStrictXMLAttrValue (Node node, String attr)
			throws XMLAnalysisException
	{
		NamedNodeMap attrMap = node.getAttributes();
		if (attrMap == null)
		{
			throw new XMLAnalysisException("Node has no attributes");
		}
		Node attrNode = node.getAttributes().getNamedItem(attr);
		if (attrNode == null)
		{
			throw new XMLAnalysisException("Node has no attribute " + attr);
		}
		return attrNode.getNodeValue();
	}

	/**
	 * Extract value of attribute <code>attr</code> from XML node <code>node</code>
	 * 
	 * @param node
	 *           XML node from which attribute value should be extracted
	 * @param attr
	 *           Attribute name
	 * @return Attribute value or null if node has no such attribute
	 */
	public String extractXMLAttrValue (Node node, String attr)
	{
		NamedNodeMap attrMap = node.getAttributes();
		if (attrMap == null)
		{
			return null;
		}
		Node attrNode = node.getAttributes().getNamedItem(attr);
		if (attrNode == null)
		{
			return null;
		}
		else
		{
			return attrNode.getNodeValue();
		}
	}

	/**
	 * Get child of XML node <code>node</code> named <code>name</code>
	 * 
	 * @param node
	 * @param name
	 *           Name of desired node
	 * @return Chiled named <code>name</code> or null if <code>node</code> doesn't have such
	 *         child
	 */
	public Node getNamedChild (Node node, String name)
	{
		NodeList children = node.getChildNodes();
		for (int i = 0; i < children.getLength(); i++)
		{
			if (children.item(i).getNodeName().compareTo(name) == 0)
			{
				return children.item(i);
			}
		}
		return null;
	}

	/**
	 * Get only named child nodes (ignore pure text nodes and comments)
	 * 
	 * @param node
	 *           Get children of this node
	 * @return Queue of nodes (in order they've been defined in XML file)
	 */
	public Queue getOnlyNamedChildren (Node node)
	{
		Queue children = new Queue();
		NodeList nodeList = node.getChildNodes();
		for (int i = 0; i < nodeList.getLength(); i++)
		{
			String strName = nodeList.item(i).getNodeName();
			if (strName.compareTo("#text") != 0
					&& strName.compareTo("#comment") != 0)
			{
				children.inject(nodeList.item(i));
			}
		}
		return children;
	}

	/**
	 * Create new instance of <code>className</code> using contructor with arguments list matching
	 * <code>paramTypes</code> and arguments values taken from <code>paramValues</code>
	 * 
	 * @param className
	 *           Object of type <code>className</code> should be constructed
	 * @param paramTypes
	 *           Constructor with arguments list matching <code>paramTypes</code> should be used
	 * @param paramValues
	 *           Array of arguments passed to constructor
	 * @return Instance of <code>className</code>
	 * @throws ObjectInitializationException
	 *            Object initialization using reflection techniques has failed
	 */
	public Object constructObject (String className, String[] paramTypeNames,
			Object[] paramValues) throws ObjectInitializationException
	{
		Class[] paramTypes = new Class[paramTypeNames.length];
		for (int i = 0; i < paramTypeNames.length; i++)
		{
			try
			{
				paramTypes[i] = Class.forName(paramTypeNames[i]);
			}
			catch (ClassNotFoundException e)
			{
				throw new ObjectInitializationException(e);
			}
		}
		return this.constructObject(className,paramTypes,paramValues);
	}

	/**
	 * Create new instance of <code>className</code> using contructor with arguments list matching
	 * <code>paramTypes</code> and arguments values taken from <code>paramValues</code>
	 * 
	 * @param className
	 *           Object of type <code>className</code> should be constructed
	 * @param paramTypes
	 *           Constructor with arguments list matching <code>paramTypes</code> should be used
	 * @param paramValues
	 *           Array of arguments passed to constructor
	 * @return Instance of <code>className</code>
	 * @throws ObjectInitializationException
	 *            Object initialization using reflection techniques has failed
	 */
	public Object constructObject (String className, Class[] paramTypes,
			Object[] paramValues) throws ObjectInitializationException
	{
		try
		{
			Class classRep = Class.forName(className);
			Constructor cnstr = classRep.getConstructor(paramTypes);
			return cnstr.newInstance(paramValues);
		}
		catch (ClassNotFoundException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (SecurityException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (NoSuchMethodException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (IllegalArgumentException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (InstantiationException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (IllegalAccessException e)
		{
			throw new ObjectInitializationException(e);
		}
		catch (InvocationTargetException e)
		{
			throw new ObjectInitializationException(e);
		}
	}

	public String filterNames (String str)
	{
		String result = str;
		result = result.replaceAll(Prefixes.BaseLocationURI,"loc");
		result = result.replaceAll(Prefixes.BaseMoneyURI,"mon");
		result = result.replaceAll(Prefixes.BaseOntoStorageURI,"onts");
		result = result.replaceAll(Prefixes.BaseRestaurantURI,"res");
		result = result.replaceAll("http://www.w3.org/1999/02/22-rdf-syntax-ns",
				"rdf");
		result = result.replaceAll("http://www.w3.org/2000/01/rdf-schema","rdfs");
		return result;
	}

	/**
	 * Get current working directory path
	 * 
	 * @return Current working directory path
	 */
	public String getCWD ()
	{
		File tmpFile = new File(".");
		String cwd = tmpFile.getAbsolutePath();
		return cwd.substring(0,cwd.length() - 1);
	}

	/**
	 * Replace sequences of multiple spaces in <code>str</code> with single one
	 * 
	 * @param str
	 *           String to be cleaned
	 * @return String without multiple spaces
	 */
	public String strCleanMultipleSpaces (String str)
	{
		String aux = str;
		aux = aux.replaceAll("      "," ");
		aux = aux.replaceAll("     "," ");
		aux = aux.replaceAll("    "," ");
		aux = aux.replaceAll("   "," ");
		aux = aux.replaceAll("  "," ");
		return aux.trim();
	}

	/**
	 * Replace all characters in string <code>str</code> from <code>chars</code> with spaces and
	 * after that clean the result of replacement from multiple spaces
	 * 
	 * @param str
	 *           String to be cleaned
	 * @param chars
	 *           Charceters to be replaced by spaces
	 * @return Cleaned string
	 */
	public String strCharsToSpaces (String str, char[] chars)
	{
		String aux = str;
		for (int i = 0; i < chars.length; i++)
		{
			aux = aux.replace(chars[i],' ');
		}
		return this.strCleanMultipleSpaces(aux);
	}

	/**
	 * Transform string <code>str</code> to the format in which all words have first letter in
	 * upper case and all other in lower case
	 * 
	 * @param str
	 *           String to be transformed
	 * @return Transformed string
	 */
	public String strToXxxx (String str)
	{
		StringBuffer buffer = new StringBuffer();
		String strOper = this.strCleanMultipleSpaces(str);

		String[] parts = strOper.split(" ");
		for (int i = 0; i < parts.length; i++)
		{
			if (parts[i].length() > 0)
			{
				buffer.append(Utils.strToUpperAt(parts[i].toLowerCase(),0));
				buffer.append(" ");
			}
		}
		return buffer.toString();
	}

	/**
	 * Append <code>line</code> and character '\n' after it to buffer
	 * 
	 * @param buffer
	 *           String buffer to append characters
	 * @param line
	 *           String to appended
	 */
	public void appendLine (StringBuffer buffer, String line)
	{
		buffer.append(line);
		buffer.append('\n');
	}

	/**
	 * Append <code>line</code>, character '\t' before it and character '\n' after it to buffer
	 * 
	 * @param buffer
	 *           String buffer to append characters
	 * @param line
	 *           String to appended
	 */
	public void appendIndentLine (StringBuffer buffer, String line)
	{
		buffer.append('\t');
		buffer.append(line);
		buffer.append('\n');
	}
}
