package logrus.tools.queue;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * Implementation of queue, basic data structure with some additional capabilities
 */
public class Queue
{
	// number of nodes in queue
	private int count;

	// head of this queue
	private QueueNode head;

	// hashtable containing mappings element -> node (for fast checking element existance in queue)
	private Hashtable mappings;

	// tail of this queue
	private QueueNode tail;

	/**
	 * Creates empty queue
	 */
	public Queue ()
	{
		super();
		this.head = null;
		this.tail = null;
		this.count = 0;
		this.mappings = new Hashtable();
	}
	
	public String toString ()
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("There are ");
		buffer.append(this.count());
		buffer.append(" objects in queue:\n");
		int i=1;
		for (Iterator iter = this.iterator(); iter.hasNext();)
		{
			Object obj = iter.next();
			if (obj==null)
			{
				buffer.append("[");
				buffer.append(i);
				buffer.append("]: [null]\n");
			}
			else
			{
				buffer.append("[");
				buffer.append(i);
				buffer.append("]: [");
				buffer.append(obj);
				buffer.append("]\n");
			}
			++i;
		}
		return buffer.toString();
	}

	/**
	 * Clears whole queue
	 */
	public void clear ()
	{
		this.count = 0;
		this.mappings.clear();
		this.head = null;
		this.tail = null;
	}

	/**
	 * Check if queue contains specified element
	 * 
	 * @param key
	 *           Element to be checked
	 * @return Boolean value indicating if queue contains specified element
	 */
	public boolean contains (Object key)
	{
		return this.mappings.get(key) != null;
	}

	/**
	 * @return Returns the count.
	 */
	public int count ()
	{
		return this.count;
	}

	/**
	 * Remove LAst node in queue
	 * 
	 * @return Content of removed last node in queue
	 */
	public Object eject ()
	{
		if (this.tail == null)
		{
			return null;
		}
		else
		{
			QueueNode newTail = this.tail.getPrev();
			if (newTail == null)
			{
				this.head = null;
			}
			Object key = this.tail.remove();
			this.tail = newTail;
			this.mappings.remove(key);
			this.count--;
			return key;
		}
	}

	/**
	 * Check if queue is empty
	 * 
	 * @return
	 */
	public boolean empty ()
	{
		return this.head == null;
	}

	/**
	 * @return Returns the head.
	 */
	public QueueNode getHead ()
	{
		return this.head;
	}

	/**
	 * Get Node storing specified key
	 * 
	 * @param key
	 *           Desired content
	 * @return Node storing <code>key</code> or null if such node does not exist
	 */
	public QueueNode getNode (Object key)
	{
		return (QueueNode)this.mappings.get(key);
	}

	/**
	 * @return Returns the tail.
	 */
	public QueueNode getTail ()
	{
		return this.tail;
	}
	
	/**
	 * Equivalent to <code>inject</code> method
	 * @param key Object to be inserted into queue
	 */
	public void add (Object key)
	{
		this.inject(key);
	}
	
	/**
	 * Equivalent to <code>injectUnique</code> method
	 * @param key Object to be inserted into queue
	 */
	public void addUnique (Object key)
	{
		this.injectUnique(key);
	}
	

	/**
	 * Insert new node storing <code>key</code> as last node in queue
	 * 
	 * @param key Object to be inserted into queue
	 */
	public void inject (Object key)
	{
		QueueNode newNode = new QueueNode(key);

		newNode.setPrev(this.tail);
		if (this.tail != null)
		{
			this.tail.setNext(newNode);
		}
		else
		{
			this.head = newNode;
		}
		this.tail = newNode;
		this.mappings.put(key,newNode);
		this.count++;
	}

	/**
	 * Insert new node storing <code>key</code> as last node in queue
	 * 
	 * @param key
	 */
	public void injectUnique (Object key)
	{
		if (!this.contains(key))
		{
			this.inject(key);
		}
	}

	/**
	 * Get iterator for this queue
	 * 
	 * @return Iterator for this queue
	 */
	public Iterator iterator ()
	{
		return new QueueIterator(this);
	}

	/**
	 * Remove First node in queue
	 * 
	 * @return Content of removed first node in queue
	 */
	public Object pop ()
	{
		if (this.head == null)
		{
			return null;
		}
		else
		{
			QueueNode newHead = this.head.getNext();
			if (newHead == null)
			{
				this.tail = null;
			}
			Object key = this.head.remove();
			this.head = newHead;
			this.mappings.remove(key);
			this.count--;
			return key;
		}
	}

	/**
	 * Insert new node storing <code>content</code> as first node in queue
	 * 
	 * @param key
	 */
	public void push (Object key)
	{
		QueueNode newNode = new QueueNode(key);
		newNode.setNext(this.head);
		if (this.head != null)
		{
			this.head.setPrev(newNode);
		}
		else
		{
			this.tail = newNode;
		}
		this.head = newNode;
		this.mappings.put(key,newNode);
		this.count++;
	}

	/**
	 * Insert new node storing <code>content</code> as first node in queue only if it's not already
	 * included in queue
	 * 
	 * @param key
	 */
	public void pushUnique (Object key)
	{
		if (!this.contains(key))
		{
			this.push(key);
		}
	}

	/**
	 * Remove object storing specified key
	 * 
	 * @param content
	 */
	public void remove (Object key)
	{
		QueueNode toRemove = this.getNode(key);
		if (toRemove != null)
		{
			if (toRemove == this.head)
			{
				this.pop();
				return;
			}
			if (toRemove == this.tail)
			{
				this.eject();
				return;
			}
			toRemove.remove();
		}
	}

	/**
	 * Convert queue to array of objects
	 * 
	 * @return Array of objects stored in queue, in queue order
	 */
	public Object[] toArray ()
	{
		if (this.count > 0)
		{
			Object[] arr = new Object[this.count];
			QueueNode tmp = this.head;
			for (int i = 0; i < this.count; i++)
			{
				arr[i] = tmp.getKey();
				tmp = tmp.getNext();
			}
			return arr;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Add all elements from <code>queue</code>, preserve order of objects
	 * 
	 * @param queue
	 *           Queue from which objects are taken
	 */
	public void addAll (Queue queue)
	{
		for (Iterator iter = queue.iterator(); iter.hasNext();)
		{
			this.inject(iter.next());
		}
	}

	/**
	 * Add all elements from <code>queue</code>, preserve order of objects, but insert object to
	 * the queue only if it hasn't been already included
	 * 
	 * @param queue
	 *           Queue from which objects are taken
	 */
	public void addAllUnique (Queue queue)
	{
		for (Iterator iter = queue.iterator(); iter.hasNext();)
		{
			this.injectUnique(iter.next());
		}
	}

	/**
	 * Replace first occurence of <code>oldKey</code> to <code>newKey</code>
	 * @param oldKey Value to be replaced
	 * @param newKey New value
	 */
	public void replace (Object oldKey, Object newKey)
	{
		QueueNode node = this.getNode(oldKey);
		if (node != null)
		{
			node.setKey(newKey);
		}
	}
}
