package logrus.tools.settings;

import java.util.Iterator;

import logrus.tools.Utils;
import logrus.tools.XMLAnalysisException;
import logrus.tools.queue.Queue;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Singleton storing current application settings
 */
public class SettingsManager
{
	// singleton instance
	private static SettingsManager instance;

	// current application settings
	private Settings settings;

	/**
	 * Private constructor, following singleton design pattern
	 */
	private SettingsManager ()
	{
		super();
		if (this.settings == null)
		{
			this.settings = this.getDefaultSettings();
		}
	}

	/**
	 * @return Returns the instance.
	 */
	public static SettingsManager getInstance ()
	{
		if (SettingsManager.instance == null)
		{
			SettingsManager.instance = new SettingsManager();
		}
		return SettingsManager.instance;
	}

	/**
	 * @param settings
	 *           The settings to set.
	 */
	public void setSettings (Settings settings)
	{
		this.settings = settings;
	}

	/**
	 * @return Returns the settings.
	 */
	public static Settings getSettings ()
	{
		return SettingsManager.getInstance().settings;
	}

	/**
	 * Load settings from specified file
	 * 
	 * @param filename
	 * @throws XMLAnalysisException
	 */
	public void loadSettings (String filename) throws XMLAnalysisException
	{
		this.settings = new Settings();
		Utils utils = Utils.getInstance();
		Document doc = utils.loadXMLDocument(filename);
		Node root = doc.getFirstChild();
		if (root.getNodeName() != "Settings")
		{
			throw new XMLAnalysisException("Incorrect root element");
		}

		// get child nodes of "<Settings>" node
		Queue children = utils.getOnlyNamedChildren(root);
		for (Iterator iter = children.iterator(); iter.hasNext();)
		{
			Node node = (Node)iter.next();
			String nodeName = node.getNodeName();
			if (nodeName.compareTo("NumberReadDBA") == 0)
			{
				String value = utils.extractStrictXMLAttrValue(node,"value");
				int iValue = Integer.parseInt(value);
				this.settings.setNumberReadDBA(iValue);
				continue;
			}
			if (nodeName.compareTo("NumberWA") == 0)
			{
				String value = utils.extractStrictXMLAttrValue(node,"value");
				int iValue = Integer.parseInt(value);
				this.settings.setNumberWA(iValue);
				continue;
			}
			if (nodeName.compareTo("NumberMaxIndexingBehaviours") == 0)
			{
				String value = utils.extractStrictXMLAttrValue(node,"value");
				int iValue = Integer.parseInt(value);
				this.settings.setNumberMaxIndexingBehaviours(iValue);
				continue;
			}
			if (nodeName.compareTo("InitialTermOfValidity") == 0)
			{
				String value = utils.extractStrictXMLAttrValue(node,"value");
				int iValue = Integer.parseInt(value);
				// convert from minutes to miliseconds
				iValue *= (60 * 1000);
				this.settings.setInitialValidityTerm(iValue);
				continue;
			}
			if (nodeName.compareTo("UseDataRandomization") == 0)
			{
				String value = utils.extractStrictXMLAttrValue(node,"value");
				if (value.compareTo("true") == 0)
				{
					this.settings.setUseDataRandomization(true);
				}
				else if (value.compareTo("false") == 0)
				{
					this.settings.setUseDataRandomization(false);
				}
				continue;
			}
			if (nodeName.compareTo("Database") == 0)
			{
				String server = utils.extractStrictXMLAttrValue(node,"server");
				this.settings.setDbServer(server);
				String provider = utils.extractStrictXMLAttrValue(node,"provider");
				this.settings.setDbProvider(provider);
				String schemata = utils.extractStrictXMLAttrValue(node,"schemata");
				this.settings.setDbSchemata(schemata);
				String user = utils.extractStrictXMLAttrValue(node,"user");
				this.settings.setDbUser(user);
				String password = utils.extractStrictXMLAttrValue(node,"password");
				this.settings.setDbPassword(password);
				continue;
			}
		}
	}

	/**
	 * Create <code>Settings</code> object with default values
	 * 
	 * @return Default settings
	 */
	public Settings getDefaultSettings ()
	{
		Settings sett = new Settings();
		sett.setEncoding("windows-1250");
		sett.setNumberMaxIndexingBehaviours(2);
		sett.setNumberReadDBA(2);
		sett.setNumberWA(2);
		sett.setInitialValidityTerm(60 * 1000);
		sett.setUseDataRandomization(false);
		sett.setDbProvider("MySQL");
		sett.setDbSchemata("ontostorage");
		sett.setDbUser("root");
		sett.setDbPassword("aldebaran");
		sett.setDbServer("puter");
		return sett;
	}

	/**
	 * Load default settings
	 */
	public void loadSettings ()
	{
		this.settings = this.getDefaultSettings();
	}
}
