package logrus.tools.web;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.Charset;

import logrus.tools.ContentNotAvilableException;
import logrus.tools.Utils;
import logrus.tools.web.visitors.NodeVisitor;
import logrus.tools.web.visitors.NodeVisitorEx;
import logrus.tools.web.visitors.NthTagSearcherNodeVisitor;
import logrus.tools.web.visitors.PrefixSearcherVisitor;
import logrus.tools.web.visitors.PrintVisitor;
import logrus.tools.web.visitors.TagSearcherNodeVisitor;

import org.cyberneko.html.parsers.DOMParser;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * Class containing various web-related functions. This class is a singleton.
 */
public class WebUtils
{
	// singleton instance
	private static WebUtils instance = null;

	/**
	 * Constructor kept private (following singleton design pattern).
	 */
	private WebUtils ()
	{
		super();
	}

	/**
	 * Get instance
	 * 
	 * @return
	 */
	public static WebUtils getInstance ()
	{
		if (WebUtils.instance == null)
		{
			WebUtils.instance = new WebUtils();
		}
		return WebUtils.instance;
	}

	/**
	 * Get content of page/file located at specified URL
	 * 
	 * @param url
	 * @return
	 * @throws ContentNotAvilableException
	 *            Remote file is not accessible or is empty
	 */
	public String getContentByURL (String url, String encoding)
			throws ContentNotAvilableException
	{
		try
		{
			URL urlAddress = new URL(url);
			Charset charset = Charset.forName(encoding);
			
			DataInputStream inStream = new DataInputStream((InputStream)urlAddress
					.getContent());
			InputStreamReader input = new InputStreamReader(inStream,charset);
			
			StringBuffer buffer = new StringBuffer();
			Reader reader = new BufferedReader(input);
			int ch;
			while ((ch = reader.read()) > -1) {
				buffer.append((char)ch);
			}
			reader.close();
			return buffer.toString();
			/*
			StringBuffer contentBuff = new StringBuffer();
			String line;
			while ((line = inStream.readLine()) != null)
			{
				contentBuff.append(line);
			}
			String content = contentBuff.toString();
			if (content == null || content.length() == 0)
			{
				throw new ContentNotAvilableException("File has zero length");
			}
			return content;
			*/			
		}
		catch (MalformedURLException e)
		{
			throw new ContentNotAvilableException(e);
		}
		catch (IOException e)
		{
			throw new ContentNotAvilableException(e);
		}
	}

	/**
	 * Traverse HTML document using specified <code>NodeVisitor</code>
	 * 
	 * @param url
	 *           URL of HTML document
	 * @param visitor
	 *           Visitor used for visiting HTML nodes
	 * @return Boolean flag determining result of traversing HTML document
	 * @throws IOException
	 * @throws SAXException
	 */
	public boolean traverseHTMLDocument (String url, NodeVisitor visitor)
			throws SAXException, IOException
	{
		DOMParser parser = new DOMParser();
		parser.parse(url);
		Node root = parser.getDocument();
		return this.visitSubTree(root,visitor);
	}

	/**
	 * Traverse HTML subtree starting in <code>root</code> node using specified
	 * <code>NodeVisitor</code>
	 * 
	 * @param root
	 *           Root of HTML subtree to be parsed
	 * @param visitor
	 *           Visitor used for visiting HTML nodes
	 * @return Boolean flag determining result of subtree of HTML document
	 */
	public boolean traverseHTMLDocument (Node root, NodeVisitor visitor)
	{
		return this.visitSubTreeAndFirstSibling(root,visitor);
	}

	/**
	 * Traverse HTML document using specified <code>NodeVisitor</code>
	 * 
	 * @return
	 * @throws IOException
	 * @throws SAXException
	 */
	public boolean printHTML (String url) throws SAXException, IOException
	{
		DOMParser parser = new DOMParser();
		parser.parse(url);
		Node root = parser.getDocument();
		PrintVisitor visitor = new PrintVisitor(true);
		return this.visitNodePrint(root,visitor,"");
	}

	/**
	 * Process selected node using specified <code>NodeVisitor</code> and if visitor is requesting
	 * further traversing of HTML document (by returning false) then process its children or next
	 * sibling
	 * 
	 * @param node
	 *           Node from which traversing should be started
	 * @param visitor
	 *           Visitor used for processing nodes
	 * @return True if visitor succeeded, otherwise returns false
	 */
	private boolean visitNodePrint (Node node, NodeVisitorEx visitor,
			String indentation)
	{
		if (visitor.processNode(node,indentation))
		{
			return true;
		}
		else
		{
			Node child = node.getFirstChild();
			while (child != null)
			{
				if (this.visitNodePrint(child,visitor,indentation + " "))
				{
					return true;
				}
				child = child.getNextSibling();
			}
		}
		return false;
	}

	/**
	 * Process selected node using specified <code>NodeVisitor</code> and if visitor is requesting
	 * further traversing of HTML document (by returning false) then process its children
	 * 
	 * @param node
	 *           Node from which traversing should be started
	 * @param visitor
	 *           Visitor used for processing nodes
	 * @return True if visitor succeeded, otherwise returns false
	 */
	private boolean visitSubTree (Node node, NodeVisitor visitor)
	{
		if (visitor.processNode(node))
		{
			return true;
		}
		else
		{
			Node child = node.getFirstChild();
			while (child != null)
			{
				if (this.visitSubTree(child,visitor))
				{
					return true;
				}
				child = child.getNextSibling();
			}
		}
		return false;
	}

	/**
	 * Process selected node using specified <code>NodeVisitor</code> and if visitor is requesting
	 * further traversing of HTML document (by returning false) then process its children
	 * 
	 * @param node
	 *           Node from which traversing should be started
	 * @param visitor
	 *           Visitor used for processing nodes
	 * @param data
	 *           Additional data required by visitor
	 * @return True if visitor succeeded, otherwise returns false
	 */
	private boolean visitSubTree (Node node, NodeVisitorEx visitor, Object data)
	{
		if (visitor.processNode(node,data))
		{
			return true;
		}
		else
		{
			Node child = node.getFirstChild();
			while (child != null)
			{
				if (this.visitSubTree(child,visitor,data))
				{
					return true;
				}
				child = child.getNextSibling();
			}
		}
		return false;
	}

	/**
	 * Process selected node using specified <code>NodeVisitor</code> and if visitor is requesting
	 * further traversing of HTML document (by returning false) then process its children or next
	 * sibling
	 * 
	 * @param node
	 *           Node from which traversing should be started
	 * @param visitor
	 *           Visitor used for processing nodes
	 * @return True if visitor succeeded, otherwise returns false
	 */
	private boolean visitSubTreeAndFirstSibling (Node node, NodeVisitor visitor)
	{
		if (visitor.processNode(node))
		{
			return true;
		}
		else
		{
			Node child = node.getFirstChild();
			while (child != null)
			{
				if (this.visitSubTreeAndFirstSibling(child,visitor))
				{
					return true;
				}
				child = child.getNextSibling();
			}
			Node sibling = node.getNextSibling();
			if (sibling != null)
			{
				return this.visitSubTreeAndFirstSibling(sibling,visitor);
			}
			else
			{
				return false;
			}
		}
	}

	/**
	 * Process selected node using specified <code>NodeVisitor</code> and if visitor is requesting
	 * further traversing of HTML document (by returning false) then process its children or next
	 * sibling
	 * 
	 * @param node
	 *           Node from which traversing should be started
	 * @param visitor
	 *           Visitor used for processing nodes
	 * @param data
	 *           Additional data required by visitor
	 * @return True if visitor succeeded, otherwise returns false
	 */
	private boolean visitSubTreeAndFirstSibling (Node node,
			NodeVisitorEx visitor, Object data)
	{
		if (visitor.processNode(node,data))
		{
			return true;
		}
		else
		{
			Node child = node.getFirstChild();
			while (child != null)
			{
				if (this.visitSubTreeAndFirstSibling(child,visitor,data))
				{
					return true;
				}
				child = child.getNextSibling();
			}
			Node sibling = node.getNextSibling();
			if (sibling != null)
			{
				return this.visitSubTreeAndFirstSibling(sibling,visitor,data);
			}
			else
			{
				return false;
			}
		}
	}

	/**
	 * Find next node representing HTML tag <code>tagName</code>. Start search from
	 * <code>node</code>
	 * 
	 * @param node
	 *           Search starts from this node
	 * @param tagName
	 *           Name of tag to be found
	 * @param tagEnd
	 *           Stop searching for node named <code>tagName</code> after founding node
	 *           <code>tagName</code>
	 * @return First node tagged <code>tagEnd</code>
	 */
	public Node findNextTag (Node node, String tagName, String tagEnd)
	{
		TagSearcherNodeVisitor visitor = new TagSearcherNodeVisitor(tagName,
				tagEnd);
		if (this.visitSubTreeAndFirstSibling(node,visitor))
		{
			if (visitor.isFoundLimitingTag())
			{
				return null;
			}
			else
			{
				return (Node)visitor.getResult();
			}
		}
		else
		{
			return null;
		}
	}

	/**
	 * Find next node representing HTML tag <code>tagName</code>. Start search from
	 * <code>node</code>
	 * 
	 * @param node
	 *           Search starts from this node
	 * @param tagName
	 *           Name of tag to be found
	 * @return First node tagged <code>tagEnd</code>
	 */
	public Node findNextTag (Node node, String tagName)
	{
		return this.findNextTag(node,tagName,null);
	}

	/**
	 * Find <code>n</code>-th node representing HTML tag <code>tagName</code>. Start search
	 * from <code>node</code>
	 * 
	 * @param node
	 *           Search starts from this node
	 * @param tagName
	 *           Name of tag to be found
	 * @param tagEnd
	 *           Stop searching for node named <code>tagName</code> after founding node
	 *           <code>tagName</code>
	 * @param n
	 * @return First node tagged <code>tagEnd</code>
	 */
	public Node findTagN (Node node, String tagName, String tagEnd, int n)
	{
		NthTagSearcherNodeVisitor visitor = new NthTagSearcherNodeVisitor(
				tagName,tagEnd,n);
		if (this.visitSubTreeAndFirstSibling(node,visitor))
		{
			if (visitor.isFoundLimitingTag())
			{
				return null;
			}
			else
			{
				return (Node)visitor.getResult();
			}
		}
		else
		{
			return null;
		}
	}

	/**
	 * Find <code>n</code>-th node representing HTML text tag starting with <code>strStart</code>.
	 * Start search from <code>node</code> and stop search if tag <code>tagEnd</code> encountered
	 * 
	 * @param node
	 *           Search starts from this node
	 * @param prefix
	 *           Prefix of text node
	 * @param tagEnd
	 *           Stop searching for node named <code>tagName</code> after founding node
	 *           <code>tagName</code>
	 * @param n
	 *           Number of occurrence to be found
	 * @return First node tagged <code>tagEnd</code>
	 */
	public Node findTextTagPrefixN (Node node, String prefix, String tagEnd,
			int n)
	{
		PrefixSearcherVisitor visitor = new PrefixSearcherVisitor(prefix,tagEnd,n);
		if (this.visitSubTreeAndFirstSibling(node,visitor))
		{
			if (visitor.isFoundLimitingTag())
			{
				return null;
			}
			else
			{
				return (Node)visitor.getResult();
			}
		}
		else
		{
			return null;
		}
	}

	/**
	 * Find <code>n</code>-th node representing HTML tag <code>tagName</code>. Start search
	 * from <code>node</code>
	 * 
	 * @param node
	 *           Search starts from this node
	 * @param tagName
	 *           Name of tag to be found
	 * @param n
	 * @return First node tagged <code>tagEnd</code>
	 */
	public Node findTagN (Node node, String tagName, int n)
	{
		return this.findTagN(node,tagName,null,n);
	}

	/**
	 * Check if node <code>node</code> is a HTML tag with type <code>tagName</code> or is a text
	 * node with string <code>tagName</code>
	 * 
	 * @param node
	 *           HTML node to be checked
	 * @param tagName
	 *           Desired tag name or text node content
	 * @return Boolean value indicating if
	 */
	public boolean compareNode (Node node, String tagName)
	{
		if (node.getNodeName().compareTo("#text") == 0)
		{
			String strNode = node.toString();
			return Utils.strIsEqual(this.extractTextNodeContent(node),tagName);
		}
		else
		{
			return Utils.strIsEqual(node.getNodeName(),tagName);
		}
	}

	/**
	 * Extract content from HTML text node
	 * 
	 * @param node
	 *           HTML text node
	 * @return Text content from <code>node</code>
	 */
	public String extractTextNodeContent (Node node)
	{
		String strNode = node.toString();
		return strNode.substring(7,strNode.length() - 1).trim();
	}

	/**
	 * Extract link from href
	 * 
	 * @param node
	 *           Expceted node is A
	 * @return Link extracted from href attribute or null if it's not A node
	 */
	public String extractLink (Node node)
	{
		return this.getAttributeValue(node,"href");
	}

	/**
	 * Extract value of attribute <code>attr</code> from HTML node <code>node</code>
	 * 
	 * @param node
	 *           HTML node from which value should be extracted
	 * @param attr
	 *           Desired attribute name
	 * @return Value of attribute <code>attr</code> of <code>null</code> if <code>node</code>
	 *         doesn't have such attribute
	 */
	public String getAttributeValue (Node node, String attr)
	{
		Node attrNode = node.getAttributes().getNamedItem(attr);
		if (attrNode == null)
		{
			return null;
		}
		else
		{
			String nodeString = attrNode.toString();
			return nodeString.substring(attr.length() + 2,nodeString.length() - 1);
		}
	}

	/**
	 * Check if <code>node</code> has attribute <code>attr</code> with value equal
	 * <code>expValue</code>
	 * 
	 * @param node
	 *           HTML node
	 * @param attr
	 *           Attribute name
	 * @param expValue
	 *           Expected value of <code>attr</code>
	 * @return True if <code>node</code> has attribute <code>attr</code> with value equal
	 *         <code>expValue</code>, otherwise returns false
	 */
	public boolean matchesValue (Node node, String attr, String expValue)
	{
		String value = this.getAttributeValue(node,attr);
		if (value == null)
		{
			return expValue == null;
		}
		else
		{
			return value.trim().toLowerCase().compareTo(
					expValue.trim().toLowerCase()) == 0;
		}
	}
}
