package tests;

import junit.framework.TestCase;
import logrus.agents.datasources.DataProviderID;
import logrus.agents.datasources.IPropertiesExtractor;
import logrus.agents.datasources.WebpageParseException;
import logrus.agents.datasources.chefmoz.ChefMozPropertiesExtractor;
import logrus.agents.datasources.restauracjecom.RestComPropertiesExtractor;
import logrus.ontmodel.IOntModelWrapper;
import logrus.ontmodel.OntModelWrapperFactory;
import logrus.ontmodel.OntUtils;
import logrus.ontmodel.QuadsList;
import logrus.ontmodel.exceptions.MatchingException;
import logrus.ontmodel.exceptions.OntPropertyNotFoundException;
import logrus.ontmodel.merging.MergingReport;
import logrus.ontmodel.merging.Unificator;
import logrus.ontmodel.merging.UnionException;
import logrus.ontmodel.validation.ObjectValidator;
import logrus.ontmodel.validation.ValidationStatus;
import logrus.tools.Utils;
import logrus.tools.XMLAnalysisException;
import logrus.tools.queue.Queue;

import com.hp.hpl.jena.ontology.OntModelSpec;

public class BugHuntTest extends TestCase
{
	private IOntModelWrapper wrapper;

	private String id;

	private String resID;

	private String urlRC;

	private String urlCM;

	protected void setUp () throws Exception
	{
		super.setUp();
		this.id = "da62b3949bba4477eed8ea00950218cd529bf2b7";
		this.resID = "<http://www.agentlab.net/schemas/restaurant/" + this.id
				+ ">";
		this.urlCM = "http://chefmoz.org/Poland/MP/Cracow/U_Szkota%2C_Restauracja997476304.html";
		this.urlRC = "http://www.restauracje.com/proste_szczegoly.php3?id=1787";

		// create wrapper
		this.wrapper = OntModelWrapperFactory.getInstance().createModelFromFile(
				"file:///" + Utils.getInstance().getCWD()
						+ "databases\\default.rdf",OntModelSpec.RDFS_MEM);
		System.out.println("Model with " + this.wrapper.countStatements()
				+ " statements loaded");
	}
	
	public void testProcess ()
	{
		for (int i = 0; i < 2; i++)
		{
			System.out.println("START OF ITERATION "+i);
			System.out.println("Processing CM object");
			QuadsList quadsCM = this.parsePage(this.urlCM,DataProviderID.CHEFMOZ);
			this.simulateProcess(quadsCM);

			System.out.println("Processing RC object");
			QuadsList quadsRC = this.parsePage(this.urlRC,DataProviderID.RESTCOM);
			this.simulateProcess(quadsRC);
		}
	}

	private void printQuads (QuadsList quads)
	{
		System.out.println(quads.toString().replaceAll(this.resID,""));
	}

	private void simulateProcess_validate (QuadsList quads)
	{
		if (quads == null || quads.count() == 0)
		{
			fail("Validation: quads == null || quads.count() == 0");
		}
		try
		{
			switch (ObjectValidator.getInstance().validate(quads))
			{
				case ValidationStatus.Complete:
				case ValidationStatus.Incomplete:
					// OK
					break;
				case ValidationStatus.Incorrect:
					fail("Validation: object incorrect");
					break;
			}
		}
		catch (XMLAnalysisException e)
		{
			fail("Validation: " + e.getMessage());
		}
	}

	private QuadsList simulateProcess_matching (QuadsList quads)
	{
		try
		{
			return this.wrapper.findMatchingObject(quads);
		}
		catch (MatchingException e)
		{
			fail("Matching: " + e.getMessage());
		}
		return null;
	}

	private void simulateProcess_inject (QuadsList quads)
	{
		String objectURI = quads.getObjectURI();
		QuadsList quadsDB = null;
		try
		{
			quadsDB = this.wrapper.getObjectInfo(objectURI);
			if (quadsDB == null)
			{
				quadsDB = new QuadsList();
			}
		}
		catch (OntPropertyNotFoundException exc)
		{
			fail("Injecting: searching by URI " + exc.getMessage());
		}

		try
		{
			Queue operations = OntUtils.getInstance().calculateOperations(quadsDB,
					quads);
			System.out.println("Calculated operations:");
			System.out.println(operations.toString().replaceAll(this.resID,""));
			this.wrapper.performOperations(operations);
		}
		catch (OntPropertyNotFoundException exc)
		{
			fail("Injecting: performing operations " + exc.getMessage());
		}
	}

	private QuadsList simulateProcess_unification (QuadsList objRef,
			QuadsList objIncl)
	{
		QuadsList quadsResult = null;
		objRef.cleanQuadsWithNull();
		objIncl.cleanQuadsWithNull();
		try
		{
			MergingReport report = Unificator.getInstance().union(objRef,objIncl);
			if (!report.isCorrectMerge())
			{
				fail("Unification: Merge failed " + report);
			}
			else
			{
				quadsResult = report.getResult();
			}
		}
		catch (UnionException exc)
		{
			fail("Unification: " + exc.getMessage());
		}
		catch (XMLAnalysisException exc)
		{
			fail("Unification: " + exc.getMessage());
		}
		return quadsResult;
	}

	private void simulateProcess (QuadsList quads)
	{
		// Step 1: validate
		this.simulateProcess_validate(quads);

		// Step 2: ask wrapper about existance of this object
		QuadsList quadsMatch = this.simulateProcess_matching(quads);

		// here merged description will be stored
		QuadsList quadsMerged = null;

		// Fork: add new object or match with existing
		if (quadsMatch == null)
		{
			// Step 3a: no match in model so add object and finish process
			System.out.println("Adding new:");
			this.printQuads(quads);
			this.simulateProcess_inject(quads);
			return;
		} 
		else 
		{
			// Step 3b: matching object found in model so unification is required
			System.out.println("Matching object:");
			this.printQuads(quadsMatch);
			quadsMerged = this.simulateProcess_unification(quadsMatch,quads);

			// Step 4: add object and finish process
			this.simulateProcess_inject(quadsMerged);
			System.out.println("Adding merged:");
			this.printQuads(quads);
		}
	}

	private QuadsList parsePage (String url, String prov)
	{
		IPropertiesExtractor extractor = null;
		if (prov.compareTo(DataProviderID.CHEFMOZ) == 0)
		{
			extractor = new ChefMozPropertiesExtractor(this.id);
		} 
		else if (prov.compareTo(DataProviderID.RESTCOM) == 0)
		{
			extractor = new RestComPropertiesExtractor(this.id);
		}
		else
		{
			fail("Unknown provider" + prov);
		}

		try
		{
			return extractor.extract(url);
		}
		catch (WebpageParseException exc)
		{
			fail(exc.getMessage());
		}
		catch (XMLAnalysisException exc)
		{
			fail(exc.getMessage());
		}
		return null;
	}
}
